/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include "gtc_main.h"
#include <sstmac/common/messages/vectorpayload.h>

namespace gtc
{

  using namespace std;
  using namespace sstmac;
  using namespace sstmac::sw;

  /**
   * ---------------------------------------------------------------------------------------
   * pushi
   * ---------------------------------------------------------------------------------------
   */
  void
  gtc_main::pushi(int istep)
  {
    //short names

    field_array::ptr fa = field_array_;
    particle_decomp::ptr pd = part_decomp_;
    config_parameters::ptr p = gtcparams_;

    SSTMAC_DEBUG << "gtc(" << p->mype_ << "):: starting pushi(" << istep << ") \n";

    p->limit_vpara_ = 0; // limit_vpara=1 :  parallel velocity kept <= abs(umax)
    int conserve_particles = 0;

    if (p->irk_ == 1)
      {
        // 1st step of Runge-Kutta method

        //$omp parallel do private(m)
        comp_->compute_loop2(1, fa->mi_, 1, 5, 1);


        // 2nd step of Runge-Kutta method
      }
    else
      {

        if (p->nonlinear_ > 0.5)
          {
            comp_->compute_loop(1, p->mflux_, 1);
          }
      }

    // gather e_field using 4-point gyro-averaging, sorting in poloidal angle
    // The following line is an OpenMP directive for loop-level parallelism
    // on shared memory machines (see http://www.openmp.org).
    //$omp parallel do private(m,e1,e2,e3,kk,wz1,wz0,larmor,ij,wp0,wt00,&
    //$omp& wt10,wp1,wt01,wt11)
    comp_->compute_loop2(1, fa->mi_, 1, 4, 1);


    // primary ion marker temperature and parallel flow velocity
    comp_->compute_loop(1, p->mpsi_, 1);


    //!********** test gradual kappan **********
    //!     kappati=min(6.9_wp,(kappan+(6.9_wp-kappan)*real(istep,wp)/4000._wp))
    //!     if(mype==0.and.irk==2)write(36,*)istep,kappan,kappati
    //! ****************************************

    //! update GC position
    //! Another OpenMP parallel loop...
    //!$omp parallel do private(m,r,rinv,ii,ip,wp0,wp1,tem,q,qinv,cost,sint,cost0,&
    //!$omp& sint0,b,g,gp,ri,rip,dbdp,dbdt,dedb,deni,upara,energy,rfac,kappa,dptdp,&
    //!$omp& dptdt,dptdz,epara,vdr,wdrive,wpara,wdrift,wdot,pdot,tdot,zdot,rdot)
    comp_->compute_loop(1, fa->mi_, 60);

    if (p->irk_ == 2)
      {

        //! Avoid runaway particles by limiting absolute value of parallel velocity
        //! if parameter limit_vpara=1
        if (p->limit_vpara_ == 1)
          {

            comp_->compute_loop(1, fa->mi_, 5);

          }

        //! out of boundary particle
        //!$omp parallel do private(m)
        comp_->compute_loop(1, fa->mi_, 12);

        double sum_of_weights = 0.0;
        double total_sum_of_weights = 0.0;
        int mi_total;
        double residual_weight;
        if (conserve_particles == 1)
          {
            comp_->compute_loop(1, fa->mi_, 1);

            MPI_Allreduce(NULL, NULL, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);

            MPI_Allreduce(NULL, NULL, 1, MPI_INT, MPI_SUM, MPI_COMM_WORLD);

            comp_->compute_loop(1, fa->mi_, 1);

          }

        //! Restore temperature profile when running a nonlinear calculation
        //! (nonlinear=1.0) and parameter fixed_Tprofile > 0.

        if (p->nonlinear_ > 0.5 && p->fixed_Tprofile > 0)
          {
            if ((istep % p->ndiag_) == 0)
              {

                //!$omp parallel do private(m,cost,b,upara)
                comp_->compute_loop(1, fa->mi_, 5);

                comp_->compute_loop(1, fa->mi_, 3);

                // ! S.Ethier 06/04/03  According to the MPI standard, the send and receive
                // ! buffers cannot be the same for MPI_Reduce or MPI_Allreduce. It generates
                // ! an error on the Linux platform. We thus make sure that the 2 buffers
                // ! are different.

                MPI_Allreduce(NULL, NULL, p->mflux_, MPI_DOUBLE, MPI_SUM,
                    MPI_COMM_WORLD);

                MPI_Allreduce(NULL, NULL, p->mflux_, MPI_DOUBLE, MPI_SUM,
                    MPI_COMM_WORLD);

                comp_->compute_loop(1, p->mflux_, 3);

                // !$omp parallel do private(m,ip)
                comp_->compute_loop(1, fa->mi_, 2);

              }
          }
      }

    if (p->idiag_ == 0)
      {
        //! fluxes diagnose at irk=1

        comp_->compute_loop(1, fa->mi_, 15);

        MPI_Allreduce(NULL, NULL, p->mpsi_ + 1, MPI_DOUBLE, MPI_SUM,
            MPI_COMM_WORLD);

        comp_->compute_loop(1, p->mpsi_, 1);

        MPI_Allreduce(NULL, NULL, p->mflux_, MPI_DOUBLE, MPI_SUM,
            MPI_COMM_WORLD);

        MPI_Allreduce(NULL, NULL, p->mflux_, MPI_DOUBLE, MPI_SUM,
            MPI_COMM_WORLD);

        comp_->compute_loop(1, p->mpsi_, 3);

        //! Field energy: 1/(4*pi)sum_ijk{|Grad_phi|^2*Jacobian*r*deltar*deltat*deltaz}
        comp_->compute_loop2(1, pd->mzeta_, 1, p->mpsi_, 17);

        //  ! Sum up the values from all the toroidal domains. We need the value of
        //  ! only one processor per domain.

        MPI_Reduce(NULL, NULL, 3, MPI_DOUBLE, MPI_SUM, 0, pd->toroidal_comm_);

      }

  }

  /**
   * ---------------------------------------------------------------------------------------
   * pushe
   * @param icycle
   * @param irke
   * @param istep
   * ---------------------------------------------------------------------------------------
   */
  void
  gtc_main::pushe(int icycle, int irke, int istep)
  {
    //short names

    field_array::ptr fa = field_array_;
    particle_decomp::ptr pd = part_decomp_;
    config_parameters::ptr p = gtcparams_;

    SSTMAC_DEBUG << "gtc(" << p->mype_ << "):: starting pushe(" << icycle << ", " << irke << ", " << istep << ") \n";

    if (irke == 1)
      {
        // 1st step of Runge-Kutta method

        if (icycle == 1)
          {
            // save electron info
            if (p->irk_ * p->ihybrid_ == 1)
              {
                p->ntracer1_ = p->ntracer_;
                fa->me1_ = fa->me_;
                //$omp parallel do private(m)
                comp_->compute_loop2(1, fa->me_, 1, 6, 1);

              }
            else
              {
                // electron start from initial position
                p->ntracer_ = p->ntracer1_;
                fa->me_ = fa->me1_;
                //$omp parallel do private(m)
                comp_->compute_loop2(1, fa->me_, 1, 6, 1);

              }
          }

        //$omp parallel do private(m)
        comp_->compute_loop2(1, fa->me_, 1, 5, 1);

        // 2nd step of Runge-Kutta method
      }
    else
      {

      }

    //$omp parallel do private(m,psitmp,thetatmp,zetatmp,r,ip,jt,ipjt,wz1,&
    //$omp& kk,rdum,ii,wp1,tflr,im,tdum,j00,j01)
    comp_->compute_loop(1, fa->me_, 28);

    //  ! gather e_field using 4-point gyro-averaging, sorting in poloidal angle
    //  ! The following line is an OpenMP directive for loop-level parallelism
    //  ! on shared memory machines (see http://www.openmp.org).
    //  !$omp parallel do private(m,e1,e2,e3,e4,kk,wz1,wz0,ij,wp0,wt00,&
    //  !$omp& wt10,wp1,wt01,wt11)
    comp_->compute_loop(1, fa->me_, 38);

    //! marker temperature
    comp_->compute_loop(1, p->mpsi_, 1);

    // ! update GC position
    // ! Another OpenMP parallel loop...
    // !$omp parallel do private(m,r,rinv,ii,wp0,wp1,tem,dzonal,q,qinv,&
    // !$omp& cost,sint,cost0,sint0,&
    // !$omp& b,g,gp,ri,rip,dbdp,dbdt,dedb,deni,upara,energy,rfac,kappa,dptdp,&
    // !$omp& dptdt,dptdz,pptpt,wpara,wdrift,vdr,wdrive,wdot,pdot,tdot,zdot,rdot)
    comp_->compute_loop(1, fa->me_, 54);

    if (irke == 2)
      {
        //  ! out of boundary particle
        //  !$omp parallel do private(m)
        comp_->compute_loop(1, fa->me_, 12);

        // ! restore temperature profile

        if (p->nonlinear_ > 0.5)
          {
            if ((istep % p->ndiag_) == 0 && p->ihybrid_ == p->nhybrid_
                && icycle == 2 * p->ncycle_)
              {
                comp_->compute_loop(1, fa->me_, 4);

                MPI_Allreduce(NULL, NULL, p->mflux_, MPI_DOUBLE, MPI_SUM,
                    MPI_COMM_WORLD);

                MPI_Allreduce(NULL, NULL, p->mflux_, MPI_DOUBLE, MPI_SUM,
                    MPI_COMM_WORLD);

                // !          dtem=dtem*tem_inv/max(1.0d+00,dden) !perturbed thermal energy
                comp_->compute_loop(1, p->mflux_, 3);

                // !$omp parallel do private(m,r,ip)
                comp_->compute_loop(1, fa->me_, 3);

              }
          }

      }

    if (p->idiag_ == 0 && p->ihybrid_ * icycle * irke == 1)
      {
        //! fluxes diagnose

        comp_->compute_loop(1, fa->me_, 5);

      }

  }

}
