/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009-2011 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

/*
 * The major contents of this file was taken from http://sushpa.wordpress.com/2008/05/20/simple-matrix-multiplication-on-mpi/
 * and formatted to fit your screen (made into SST/macro skeleton)
 */

#include "simplemxm.h"
#include <iomanip>
#include <math.h>  // ceil



using boost::shared_ptr;
using namespace std;
using namespace sstmac;
using namespace sstmac::sw;

#define ERR_BADORDER    255
#define TAG_INIT      31337
#define TAG_RESULT       42
#define DISP_MAXORDER    12

namespace matrix {

    SSTRegisterApp("simplemxm", simplemxm);

    simplemxm::simplemxm(softwareid id) :
        mpiapp(id, to_string())
    {

      compute_matrix_ = lib_compute_matrix::construct(id);
      register_lib(compute_matrix_);

    }

    simplemxm::ptr
    simplemxm::construct(softwareid id)
    {
      ptr r(new simplemxm(id));
      return r;
    }

    simplemxm::~simplemxm() throw ()
    {
    }

    app::ptr
    simplemxm::clone(softwareid id)
    {
      app::ptr cl = construct(id);
      return cl;
    }

    long
    simplemxm::getRowCount(long rowsTotal, long mpiRank, long mpiSize)
    {
      /* Adjust slack of rows in case rowsTotal is not exactly divisible */
      return (rowsTotal / mpiSize) + (rowsTotal % mpiSize > mpiRank);
    }

    void
    simplemxm::consume_params(const sim_parameters::ptr &params)
    {
      matrix_order_ = params->get_optional_long_param("simplemxm_matrix_order",
          100);
    }

    void
    simplemxm::skeleton_main()
    {
      long n_ubound, n_local, n_sq, i;

      double t;
      int sizeSent, sizeToBeSent;

      /* Hello, world */
      mpi()->init();

      sstmac::sw::mpicomm world = mpi()->comm_world();

      long mpiRank = world.rank().id;
      long mpiSize = world.size().id;

      /* Get n and broadcast it to all processes */

      mpi()->bcast(1, mpitype::mpi_int, mpiid(0), world);

      n_local = getRowCount(matrix_order_, mpiRank, mpiSize);

      n_ubound = matrix_order_ * n_local; /* slave array's upper bound (partial matrix) */
      n_sq = matrix_order_ * matrix_order_; /* master array's upper bound (total matrix) */

      /* Send A by splitting it in row-wise parts */
      mpitag itag(TAG_INIT);
      const_mpistatus_t stat;
      mpirequest_t req;
      if (!mpiRank)
        {
          sizeSent = n_ubound;

          for (i = 1; i < mpiSize; i++)
            {
              sizeToBeSent = matrix_order_
                  * getRowCount(matrix_order_, i, mpiSize);
              mpi()->isend(sizeToBeSent, mpitype::mpi_double, mpiid(i), itag,
                  world, req);
              sizeSent += sizeToBeSent;
            }
        }
      else
        { /* Receive parts of A */

          mpi()->irecv(n_ubound, mpitype::mpi_double, mpiid(0), itag, world,
              req);
        }

      /* Send B completely to each process */
      mpi()->bcast(matrix_order_ * matrix_order_, mpitype::mpi_double, mpiid(0),
          world);

      mpi()->wait(req, stat);

      /* Let each process initialize C to zero */

      /* Let each process perform its own multiplications */
      SSTMAC_DEBUG << mpiRank << ": entering matrix computation \n";
      compute_matrix_->double_mxm(n_local, 1, matrix_order_, 1);
      SSTMAC_DEBUG << mpiRank << ": done with matrix computation \n";

      /* Receive partial results from each slave */
      mpitag rtag(TAG_RESULT);
      if (!mpiRank)
        {
          sizeSent = n_ubound;
          for (i = 1; i < mpiSize; i++)
            {
              sizeToBeSent = matrix_order_
                  * getRowCount(matrix_order_, i, mpiSize);
              const_mpistatus_t stat;
              mpi()->recv(sizeToBeSent, mpitype::mpi_double, mpiid(i), rtag,
                  world, stat);
              sizeSent += sizeToBeSent;
            }
        }
      else
        { /* Send partial results to master */
          mpi()->send(n_ubound, mpitype::mpi_double, mpiid(0), rtag, world);
        }

      mpi()->finalize();

    }

} //end namespace matrix

