/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/backends/native/managerdegraded.h>
#include <sstmac/hardware/node/degradednode.h>
#include <sstmac/hardware/node/simplenode.h>
#include <sstmac/hardware/nic/simplenic.h>

#include <sstmac/common/driver_util.h>

namespace sstmac
{
  namespace native
  {

    long
    manager_degraded::add_nodes(sim_parameters::ptr params)
    {
      sstmac::hw::node::ptr type;
      if (params->get_param("node_name") == "simple")
        {
          type = simplenode::construct(params);
        }
      else
        {
          std::cout << "node type: " << params->get_param("node_name");
          throw ssterror(
              "manager_degraded: don't know how to construct node type");
        }

      sstmac::hw::networkinterface::ptr nictype;
      if (params->get_param("nic_name") == "simple")
        {
          nictype = simplenic::construct(type, interconnect_);
        }
      else
        {
          std::cout << "nic type: " << params->get_param("nic_name");
          throw ssterror(
              "manager_degraded: don't know how to construct nic type");
        }

      if (interconnect_ == 0)
        {
          throw nullerror(
              "manager_degraded::add_nodes:  No interconnect defined yet.");
        }
      if (type == 0)
        {
          throw nullerror("manager_degraded::add_nodes:  Null node pointer\n");
        }

      long count = params->get_long_param("node_num");
      if (count != interconnect_->slots()){
          throw illformederror("node_num parameter does not match interconnect size. "
                               "node_num should never be specified in the input file.");
      }

      SSTMAC_DEBUG << "manager_degraded::add_nodes(" << type << ", " << count
          << ")\n";

      /** DEPRECATED node_num parameter
      if ((interconnect_->slots() > 0)
          && (count > int(interconnect_->slots() - this->nodes())))
        {
          throw valueerror(
              "manager_degraded::add_nodes:  Insufficient slots available.");
        }
      */
      if (count < 0)
        {
          std::cerr << "manager::add_nodes:  Simulator has unlimited size. "
              << " Unlimited interconnect size is no longer valid." << endl;
          throw illformederror("manager_degraded    ::add_nodes:  Infinite interconnect size.");
          /** DEPRECATED node_num parameter
          if (interconnect_->slots() <= 0)
            {
              std::cerr
              << "manager_degraded::add_nodes:  Simulator has unlimited size, "
              << "so an unlimited node count is not possible\n";
              throw illformederror(
                  "manager_degraded::add_nodes:  Need infinite memory.");
            }
          count = check_cast<int>(interconnect_->slots()) - this->nodes();
          */
        }

      std::vector<int> degraded;

      if (params->has_param("manager_degraded_nodes"))
        {
          std::string deg = params->get_param("manager_degraded_nodes");

          if (deg == "evens")
            {
              for (int i = 0; i < count; ++i)
                {
                  if (i % 2 == 0)
                    {
                      degraded.push_back(i);
                    }
                }
            }
          else if (deg == "odds")
            {
              for (int i = 0; i < count; ++i)
                {
                  if (i % 2 == 1)
                    {
                      degraded.push_back(i);
                    }
                }
            }
          else
            {
              bool error = false;
              degraded = get_intvec(deg.c_str(), error);
            }
        }

      double factor = 2.0;
      if(params->has_param("node_degraded_factor")){
          factor = params->get_double_param("node_degraded_factor");
      }


      int degi = 0;
      for (int i = 0; i < count; ++i)
        {
          node::ptr newnode;

          if(degraded.size() > degi && degraded[degi] == i){
              SSTMAC_DEBUG << "creating degraded node " << i << "\n";
              newnode = degradednode::construct(params, factor);
              degi++;
          }else{
              SSTMAC_DEBUG << "creating regular node " << i << "\n";
              newnode = type->clone();
          }

          //shared_ptr<has_stats> sget = boost::dynamic_pointer_cast<has_stats>(
          //    newnode);
          // stats->add_stats(sget->get_stats());
          sstmac::hw::networkinterface::ptr newnic = nictype->clone(newnode, interconnect_);
          newnode->set_nic(newnic);

          nodeid::ptr id = nodes_->add(newnode);
          SSTMAC_DEBUG << "manager_degraded: adding nodeid " << id->to_string()
                      << "\n";
          //available_.insert(id);
        }
      return this->open_slots();
    }

  }
}
