/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include "flownic.h"

#include <network_m.h>

namespace sstmac { namespace omnetpp {


  Define_Module(flownic)
    ;

  void
  flownic::initialize()
  {
    fromProc = gate("toProcessor$i", 0);
    toProc = gate("toProcessor$o", 0);
    fromNet = gate("toNetwork$i", 0);
    toNet = gate("toNetwork$o", 0);

    networkFlowBW = par("networkFlowBandwidth");
    memoryFlowBW = par("memoryFlowBandwidth");

    effectiveBW = std::min(networkFlowBW, memoryFlowBW);

    currentlyTransmitting = false;
    flowid = 0;

    guardMsg = new cMessage("NICguardMsg");
  }

  void
  flownic::finish()
  {
    delete guardMsg;
  }

  void
  flownic::handleMessage(cMessage *cmsg)
  {
    cGate* arrival = cmsg->getArrivalGate();

    if (arrival == fromProc)
      {
        CoreMsg* c = check_and_cast<CoreMsg*> (cmsg);

        netQ.insert(c);
        try_net_Q();

      }
    else if (arrival == fromNet)
      {

        NetworkFlowMsg* net = check_and_cast<NetworkFlowMsg*> (cmsg);

        if (!net->getIsHead())
          {
            SSTMAC_DEBUG<< "NIC: tail arrived at " << simTime().dbl() << "\n";

            CoreMsg* m = (CoreMsg*) net->decapsulate();
            send(m, toProc);
          }
        delete net;

      }
    else if (cmsg == guardMsg)
      {
        currentlyTransmitting = false;
        try_net_Q();
      }
    else
      { //it's the tail that's scheduled to go out
        send(cmsg, toNet);

        scheduleAt(simTime() + GUARD_TIME, guardMsg);

      }
  }

  void
  flownic::try_net_Q()
  {
    if (!netQ.empty() && !currentlyTransmitting)
      {

        //std::cout << "NIC: sending message " << endl;
        CoreMsg* msg = check_and_cast<CoreMsg*> (netQ.pop());

        //ack on start of transmission
        if (msg->getMsgType() == APP)
          {
            ApplicationMsg* appmsg = (ApplicationMsg*) msg;
            send_proc_ack(appmsg->dup(), toProc);

          }

        NetworkFlowMsg* head = new NetworkFlowMsg();
        head->setToAddr(msg->getToAddr());
        head->setFromAddr(msg->getFromAddr());
        head->setIsHead(true);
        head->setByteLength(0);
        head->setFlowID(flowid);
        head->setSourceBW(effectiveBW);

        send(head, toNet);

        NetworkFlowMsg* tail = new NetworkFlowMsg();
        tail->setToAddr(msg->getToAddr());
        tail->setFromAddr(msg->getFromAddr());
        tail->setIsHead(false);
        tail->encapsulate(msg);
        tail->setByteLength(0);
        tail->setFlowID(flowid);
        tail->setSourceBW(effectiveBW);

        flowid++;

        currentlyTransmitting = true;

        scheduleAt(simTime() + msg->getByteLength() / effectiveBW, tail);
      }

  }


} }
