/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include "packetnic.h"

#include <network_m.h>

namespace sstmac { namespace omnetpp {


  Define_Module(packetnic)
    ;

  void
  packetnic::initialize()
  {
    fromProc = gate("toProcessor$i", 0);
    toProc = gate("toProcessor$o", 0);
    fromNet = gate("toNetwork$i", 0);
    toNet = gate("toNetwork$o", 0);

    mtu = par("packetMTU");

    sent = 0;
    currentMsg = NULL;

    nextPacketMsg = new cMessage("nextPacketMsg");
  }

  void
  packetnic::finish()
  {

    delete nextPacketMsg;
  }

  void
  packetnic::handleMessage(cMessage *cmsg)
  {
    cGate* arrival = cmsg->getArrivalGate();

    if (cmsg == nextPacketMsg)
      {
        if (currentMsg == NULL)
          try_net_Q();
        else
          next_packet();
      }
    else if (arrival == fromProc)
      {
        CoreMsg* c = check_and_cast<CoreMsg*> (cmsg);

        netQ.insert(c);
        if (currentMsg == NULL)
          try_net_Q();

      }
    else if (arrival == fromNet)
      {
        NetworkPacketMsg* net = check_and_cast<NetworkPacketMsg*> (cmsg);

        if (net->getIsLast())
          {
            net->setByteLength(0x7FFFFFFF); //do this so omnet won't complain about decapsulating
            CoreMsg* m = (CoreMsg*) net->decapsulate();
            send(m, toProc);
          }
        delete net;
      }
  }

  void
  packetnic::next_packet()
  {
    if (currentMsg == NULL)
      {
        opp_error(
            "packetnic: in next_packet(), but currentMsg is null. shouldn't be here.");
      }

    if (!nextPacketMsg->isScheduled())
      {

        bool last = (mtu < 0) || ((sent + mtu) > currentMsg->getByteLength());

        NetworkPacketMsg* pack = new NetworkPacketMsg();
        pack->setToAddr(currentMsg->getToAddr());
        pack->setFromAddr(currentMsg->getFromAddr());
        pack->setIsLast(last);
        pack->setKind(currentMsg->getKind());

        if (last)
          {
            int diff = currentMsg->getByteLength() - sent;
            pack->encapsulate(currentMsg);
            pack->setByteLength(diff);
            sent = 0;
          }
        else
          {
            pack->setByteLength(mtu);
            sent += mtu;
          }
        SSTMAC_DEBUG << "NIC: Sending a packet of length "
            << (int) pack->getByteLength() << "\n";

        send(pack, toNet);

        if (last)
          {
            currentMsg = NULL;
          }

        simtime_t txFinishTime =
            toNet->getTransmissionChannel()->getTransmissionFinishTime();
        scheduleAt(txFinishTime, nextPacketMsg);

      }

  }

  void
  packetnic::try_net_Q()
  {
    if (!netQ.empty())
      {
        CoreMsg* msg = check_and_cast<CoreMsg*> (netQ.pop());
        currentMsg = msg;

        //ack on start of transmission
        if (msg->getMsgType() == APP)
          {
            ApplicationMsg* appmsg = (ApplicationMsg*) msg;
            send_proc_ack(appmsg->dup(), toProc);

          }

        next_packet();

      }

  }


} }
