/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_COMMON_STATS_STAT_TIME_WINDOW_H_INCLUDED
#define SSTMAC_COMMON_STATS_STAT_TIME_WINDOW_H_INCLUDED

#include <sstmac/common/errors.h>
#include <sstmac/common/stats/stat_logger.h>
#include <iostream>
#include <fstream>
#include <map>
#include <list>

#include <sstmac/common/timestamp.h>


namespace sstmac {


  /**
   * Base class for stats that collect in time windows,
   * mainly just to hold the time interval parameter.
   *
   * In general, never give a future time for the time
   * parameter.  Each stat tries to keep track of what
   * the current simulation time is, and fills in time
   * windows as needed.  So putting in a future time
   * could trigger a time window update, when there
   * could be more records that come in for the current
   * window.
   */
  class stat_time_window : public stat_logger {
  public:
     typedef stat_time_window* ptr;

     static timestamp TIME_INTERVAL;
     static bool interval_is_set_;

     static void set_time_interval( timestamp interval ) {
       TIME_INTERVAL = interval; 
       interval_is_set_ = true;
     }

  protected:
     stat_time_window(std::string namesig, std::string filename) :
      stat_logger(namesig, filename) {
        if( !interval_is_set_ ) TIME_INTERVAL = timestamp(1e-4);  
      }
  };

  /**
   * this stat_logger class just adds up all the records given to it for a window,
   * and outputs every window
   */
  class stat_time_window_sum : public stat_time_window
  {
  public:
    typedef stat_time_window_sum* ptr;

  protected:



    stat_time_window_sum(std::string namesig, std::string filename) :
      stat_time_window(namesig, filename)
    {

      current_win_time_ = timestamp(0);
      current_win_ = 0;

      val_ = 0;
    }

  public:
     static ptr
     construct(std::string namesig, std::string filename){
       return new stat_time_window_sum(namesig, filename);
     }

    virtual
    ~stat_time_window_sum()
    {
    }

    virtual void simulation_finished(timestamp end){
      record(end, 0);
    }

    void
    record(timestamp t, double val);

  protected:

    double val_;
    timestamp current_win_time_;
    int current_win_;
  };

  /**
   * this stat_logger class averages the values given to it in a window,
   * and outputs every window
   */
  class stat_time_window_avg : public stat_time_window
  {
  public:
    typedef stat_time_window_avg* ptr;

  protected:


    stat_time_window_avg(std::string namesig, std::string filename) :
      stat_time_window(namesig, filename)
    {

      current_win_time_ = timestamp(0);

      val_ = 0;
      numvals_ = 0;

    }

  public:
     static ptr
     construct(std::string namesig, std::string filename){
       return new stat_time_window_avg(namesig, filename);
     }

    virtual
    ~stat_time_window_avg()
    {
    }

    virtual void simulation_finished(timestamp end){

    }

    void
    record(timestamp t, double val);

  protected:

    double val_;

    int numvals_;
    timestamp current_win_time_;

  };

  /**
   * this class averages the values, weighting them by the time they persisted.
   * obviously, this assumes a persistent state that is being tracked, not individual
   * events (where calling record(...) is the change of this state)
   */
  class stat_time_window_time_avg : public stat_time_window
  {
    public:
    typedef stat_time_window_time_avg* ptr;

  protected:


    stat_time_window_time_avg(std::string namesig, std::string filename) :
      stat_time_window(namesig, filename)
    {

      current_win_time_ = timestamp(0);

      val_ = 0;
      pending_val_ = 0;

      last_update_ = timestamp(0);

    }

  public:
     static ptr
     construct(std::string namesig, std::string filename){
       return new stat_time_window_time_avg(namesig, filename);
     }

    virtual
    ~stat_time_window_time_avg()
    {
    }

    virtual void simulation_finished(timestamp end){

    }

    void
    record(timestamp t, double val);

  protected:

    double val_;
    double pending_val_;

    timestamp current_win_time_;

    timestamp last_update_;
  };

  /**
   * this class persists a value given to it for a given amount of time, summing it
   * into any windows that it crosses multiplied by the time that it took in that window.
   *  this is obviously useful for power (record with a power, not an energy)
   */
  class stat_time_window_sum_duration : public stat_time_window
  {
  public:
    typedef stat_time_window_sum_duration* ptr;

  protected:


    stat_time_window_sum_duration(std::string namesig, std::string filename) :
      stat_time_window(namesig, filename)
    {

      current_win_time_ = timestamp(0);

    }

  public:
     static ptr
     construct(std::string namesig, std::string filename){
       return new stat_time_window_sum_duration(namesig, filename);
     }

    virtual
    ~stat_time_window_sum_duration()
    {
    }

    virtual void simulation_finished(timestamp end){

    }

    void
    record(timestamp t, double val, timestamp dur);

  protected:

    timestamp current_win_time_;

    struct duration_record
    {
      duration_record() :
        val_(-1), start_(-1), end_(-1)
      {
        id_ = -1;
      }

      duration_record(double val, timestamp st, timestamp dur) :
        val_(val), start_(st), end_(dur)
      {
        id_ = ID_CNT++;
      }

      double val_;
      timestamp start_;
      timestamp end_;
      int id_;

      static int ID_CNT;
    };

    std::map<int, duration_record> records_;
  };
} // end of namespace sstmac
#endif
