/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */
#ifndef SSTMAC_HARDWARE_FAILURE_FAILURE_RATE_MODEL_H_INCLUDED
#define SSTMAC_HARDWARE_FAILURE_FAILURE_RATE_MODEL_H_INCLUDED

#include <sstmac/hardware/failure/failure_model.h>
#include <sstmac/common/rng.h>
#include <math.h>

namespace sstmac
{
  namespace hw
  {
    class failable;

    class failure_rate_model : public failure_model
    {
    protected:
      failure_rate_model(double l)
      {
        lambda_ = (1.0 / 31556926.0) * l; // change from years to seconds

        SSTMAC_DEBUG << to_string() << ": lambda is " << lambda_ << "\n";

        if (!rng_)
          rng_ = RNG::SHR3::construct();
      }

    public:
      typedef boost::intrusive_ptr<failure_rate_model> ptr;

      static ptr
      construct(double l)
      {
        return ptr(new failure_rate_model(l));
      }

      virtual
      ~failure_rate_model()
      {
      }

      virtual std::string
      to_string() const
      {
        return "failure_rate_model";
      }

      void
      add_failable_child(boost::intrusive_ptr<failable> ch, double prop)
      {
        children_[ch] = prop;
      }

      virtual void
      next_failure(boost::intrusive_ptr<eventmodule> parentmodel);

      /**
       * propagate - this function is called when a parent must signal that its
       * children should possibly fail directly because of it (like a power supply)
       */
      virtual void
      propagate();

    protected:
      double lambda_;
      static long rand_max;
      static RNG::UniformInteger::ptr rng_;

      // double, propagation probability
      std::map<boost::intrusive_ptr<failable>, double> children_;

      double
      genexp()
      {
        double u, x;
        u = urand();
        SSTMAC_DEBUG << "u is " << u << "\n";
        x = (-1 * log(u) / lambda_);
        return (x);
      }

      double
      urand()
      {
        return ((double) rng_->value_in_range(rand_max) + 1.0)
            / ((double) rand_max + 1.0);
      }

    };

  }
}

#endif
