#include <sstmac/hardware/memory/mc_memorymodel.h>
#include <sstmac/software/libraries/compute/mem_access.h>
namespace sstmac
{
  namespace hw
  {

    using namespace std;
    using namespace sstmac::sw;

    mc_memorymodel::mc_memorymodel(double mem_bw, timestamp mem_lat,
        double proclink_bw, timestamp proclink_lat, int num_mc, int num_socket,
        eventhandler::ptr whendone, long nodeid, computescheduler::ptr proc) :
      memorymodel(nodeid), num_mc_(num_mc), num_socket_(num_socket),
          done_(whendone), proc_(proc)
    {
      self_ = ptr(this);

      for (int i = 0; i < num_socket_ * num_mc_; i++)
        {
          mc_cnt_[i] = 0;
          memlinks_[i] = link::construct(mem_bw, mem_lat);
        }

      //assumes point-to-point between sockets (common)
      for (int i = 0; i < num_socket_; i++)
        {

          for (int j = 0; j < num_socket_; j++)
            {
              if (i != j)
                proclinks_[i + j * num_socket_] = link::construct(proclink_bw,
                    proclink_lat);
            }
        }
    }

    void
    mc_memorymodel::handle(const sst_message::ptr &msg)
    {
      if (msg->get_type() == sst_message::FAILURE)
        {
          failure_occurred(msg, self_);
          return;
        }
      mem_access::ptr acc = boost::dynamic_pointer_cast<mem_access>(msg);

      if (!acc)
        {
          throw ssterror(
              "mc_memorymodel:: retiring an access with a non mem_access message");
        }

      int sock, mc;
      get_location(acc->dstthread_, sock, mc);
      memlinks_[sock * num_mc_ + mc]->access_done();
      int src = proc_->get_socket(acc->srcthread_);
      if (src != sock)
        {
          proclinks_[src + sock * num_socket_]->access_done();
        }

      done_->handle(msg);
    }

    void
    mc_memorymodel::get_location(long threadid, int &sock, int &mc)
    {
      if (threadid >= 0)
        {
          sock = proc_->get_socket(threadid);

          if (threadmcs_.find(threadid) == threadmcs_.end())
            {
              threadmcs_[threadid] = mc_cnt_[sock];
              mc_cnt_[sock] = (mc_cnt_[sock] + 1) % num_mc_;
            }
          mc = threadmcs_[threadid];
        }
      else
        {
          sock = 0;
          mc = 0;
        }
    }

    void
    mc_memorymodel::do_access(sst_message::ptr msg)
    {
      use_me();

      mem_access::ptr acc = boost::dynamic_pointer_cast<mem_access>(msg);

      if (!acc)
        {
          throw ssterror(
              "mc_memorymodel:: trying to do an access with a non mem_access message");
        }

      int sourcesock, destsock;
      int mc, temp;

      get_location(acc->srcthread_, sourcesock, temp);



      get_location(acc->dstthread_, destsock, mc);

      timestamp done(0);

      if (destsock != sourcesock)
        {
          SSTMAC_DEBUG << "mc_memorymodel : proc link " << (sourcesock
                + destsock * num_socket_) << "\n";
          done += proclinks_[sourcesock + destsock * num_socket_]->new_access(
              msg->get_byte_length());
        }

      SSTMAC_DEBUG << "mc_memorymodel: doing access of "
            << msg->get_byte_length() << " at socket " << destsock << " at mc "
            << mc << "\n";

      done += memlinks_[destsock * num_mc_ + mc]->new_access(
          msg->get_byte_length());

      schedule(now() + done, self_, msg);
    }

  }
} /* namespace sstmac */
