/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/network/congestion/circuit-transfer.h>
#include <sstmac/hardware/network/congestion/circuit-transfer-finish.h>
#include <sstmac/hardware/network/congestion/circuit-edge.h>

#include <sstmac/hardware/network/congestion/transfer_time.h>
#include <sstmac/hardware/network/topology/topology.h>
#include <sstmac/hardware/network/topology/walk.h>
#include <sstmac/common/errors.h>

namespace sstmac
{
  namespace hw
  {

    // Types.

    //
    // Hi.
    //
    circuit::transfer::transfer(const circuit::ptr &parent,
        const sst_message::ptr &msg, const walk &the_walk,
        const timestamp &send_time, const timestamp &recv_time,
        const timestamp &tran_time, const eventhandler::ptr &sendcomplete,
        const eventhandler::ptr &recvcomplete) :
        parent_(parent), msg_(msg), send_time_(send_time), recv_time_(
            recv_time), tran_time_(tran_time), sendcomplete_(sendcomplete), recvcomplete_(
            recvcomplete)
    {

      if (parent_ == 0)
        {
          throw nullerror("circuit::transfer(...):  null parent pointer.");
        }
      pending_edges_ = the_walk.length();
      walk_length_ = pending_edges_;
      edges_.reserve(pending_edges_);
      for (int index = 0; index < pending_edges_; ++index)
        {
          int edgenum =
              parent_->topology_->g().edge_to_index(the_walk.e(index)).value();
          edges_.push_back(edgenum);
        }
    }

    //
    /// Hi.
    //
    circuit::transfer::ptr
    circuit::transfer::construct(const circuit::ptr &parent,
        const sst_message::ptr &msg, const walk &the_walk,
        const timestamp &send_time, const timestamp &recv_time,
        const timestamp &tran_time, const eventhandler::ptr &sendcomplete,
        const eventhandler::ptr &recvcomplete)
    {
      transfer::ptr retval(
          new transfer(parent, msg, the_walk, send_time, recv_time, tran_time,
              sendcomplete, recvcomplete));

      return retval;
    }

    //
    // Goodbye.
    //
    circuit::transfer::~transfer() throw ()
    {

    }

    //
    // Eventhandler completion.
    //
    void
    circuit::transfer::handle(const sst_message::ptr &msg)
    {
      for (size_t index = 0; index < edges_.size(); ++index)
        {
          parent_->edge_.at(edges_[index]).reserve(ptr(this));
        }
    }

    /// A callback used by an edge to indicate that it is now available.
    void
    circuit::transfer::edge_available(edge *)
    {
      --pending_edges_;
      if (pending_edges_ == 0)
        {
          const timestamp &now = parent_->now();
          timestamp t_send_done = now + send_time_;
          timestamp t_recv_done = now + recv_time_;
          timestamp t_tran_done = now + tran_time_;
          // We have the full circuit.
          if (msg_->get_needs_ack())
            {
              sst_message::ptr sendack = msg_->clone(sst_message::SEND_ACK);
              parent_->schedule(t_send_done, sendcomplete_, sendack);
            }
          parent_->schedule(t_recv_done, recvcomplete_, msg_);
          parent_->schedule(t_tran_done, finish::construct(parent_, edges_),
              msg_);

          if (parent_->energystat_->is_active())
            {

              parent_->energystat_->record(
                  now,
                  walk_length_ * parent_->pwr_config_->get_rtr_hop_per_byte()
                      * msg_->get_byte_length(), t_tran_done - now);

              parent_->bandwidthstat_->record(now,
                  (double) msg_->get_byte_length());

            }

          // Release any reference counted pointers, just to be sure.
          parent_ = circuit::ptr();
          edges_.clear();
          sendcomplete_ = eventhandler::ptr();
          recvcomplete_ = sendcomplete_;
        }
    }

  }
} // end of namespace sstmac.
