/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_HARDWARE_NETWORK_CONGESTION_CIRCUIT_TRANSFER_H_INCLUDED
#define SSTMAC_HARDWARE_NETWORK_CONGESTION_CIRCUIT_TRANSFER_H_INCLUDED

#include <sstmac/hardware/network/congestion/circuit.h>
#include <sstmac/common/eventhandler.h>

namespace sstmac { namespace hw  {

  // Important types.
  class walk;

  /**
   * An individual data transfer request on a circuit network.
   */
  class circuit::transfer: public eventhandler {
    /// Handler to release the edges when transmission is done.
    friend class finish;
    class finish;


    /// The parent circuit.
    circuit::ptr parent_;

    /// The number of bytes we're sending.
    sst_message::ptr msg_;

    /// The time we are spending on send, recv, and transfer.
    timestamp send_time_, recv_time_, tran_time_;

    /// The edges we need to visit.
    std::vector<int> edges_;

    /// The signal we will be sending when the sender is done.
    eventhandler::ptr sendcomplete_;

    /// The signal we will be sending when the receiver is done.
    eventhandler::ptr recvcomplete_;

#ifdef CIRCUIT_WORM
    /// The next edge (local indexing) we need to grab to complete the circuit.
    size_t next_edge_;
#else
    /// The number of edges we need to wait for before progressing.
    int64_t pending_edges_;
    int64_t walk_length_;
#endif

  protected:
    /// The constructor has to operate in two phases (unfortunately).
    transfer(const circuit::ptr &parent,
        const sst_message::ptr &msg, const walk &the_walk,
        const timestamp &send_time, const timestamp &recv_time,
        const timestamp &tran_time,
        const eventhandler::ptr &sendcomplete,
        const eventhandler::ptr &recvcomplete);



  public:
    typedef boost::intrusive_ptr<transfer> ptr;

    /// Hi.
    static transfer::ptr
    construct(const circuit::ptr &parent,
        const sst_message::ptr &msg, const walk &the_walk,
        const timestamp &send_time, const timestamp &recv_time,
        const timestamp &tran_time,
        const eventhandler::ptr &sendcomplete,
        const eventhandler::ptr &recvcomplete);

    /// Goodbye.
    virtual ~transfer() throw ();

    /// Eventhandler completion.
    virtual void handle(const sst_message::ptr &msg);

    virtual std::string to_string() const {
      return "circuit::transfer";
    }

    /// A callback used by an edge to indicate that it is now available.
    void edge_available(edge *the_edge);
  };

} } // end of namespace sstmac

#endif
