/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_HARDWARE_NETWORK_CONGESTION_ORDERED_FLOWER_H_INCLUDED
#define SSTMAC_HARDWARE_NETWORK_CONGESTION_ORDERED_FLOWER_H_INCLUDED

#include <sstmac/hardware/network/congestion/typical_flower.h>
#include <sstmac/common/timestamp.h>
#include <sstmac/hardware/network/topology/topology.h>
#include <vector>
#include <map>
#include <set>
#include <string>
#include <sstmac/common/ptr_type.h>

namespace sstmac { namespace hw  {
  
  /**
   * Allocate bandwidth on a first-come first-serve basis.
   *
   * This allocates bandwidth between flows. A priority is assigned to
   * each flow giving earlier flows higher priority. The bandwidth
   * available to a flow is the smallest maximum available bandwidth
   * on any network link in the flow. The maximum available bandwidth
   * of a link is the maximum bandwidth minus the bandwidth of all
   * higher priority flows. When a flow terminates, the liberated
   * bandwidth is reallocated among the remaining flows using the same
   * priority scheme.
   * 
   */
  class ordered_flower: public typical_flower {
  public:
    /// This specifies the number of flows considered when
    /// rebalancing the remaining flows after one flow
    /// terminates. A value of -1 causes all flows to be
    /// considered. Use of the default is recommended.
    int n_rebalance_max;

  protected:
    ordered_flower(const topology::ptr &net,
           double node_router_bw, double router_router_bw);

    /// Override of typical_flower::rebalance_flow_start
    void rebalance_flow_start(flow_id_t f);

    /// Override of typical_flower::rebalance_flow_stop
    void rebalance_flow_stop(flow_id_t oldf);

  public:
    typedef boost::intrusive_ptr<ordered_flower> ptr;

    /// Construct an ordered_flower.
    /// \param node_router_bw maximum bandwidth on network links each
    ///        node and router
    /// \param router_router_bw maximum bandwidth on network links each
    ///        router
    /// \return a shared_ptr containing the new object
    static ordered_flower::ptr
    construct(const topology::ptr &net,
              double node_router_bw, double router_router_bw);

    virtual ~ordered_flower() throw();
  };

} } // end of namespace sstmac

#endif
