/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/network/congestion/sharedbuffer.h>
#include <sstmac/hardware/network/congestion/sharedbuffer-bandwidthchange.h>
#include <sstmac/hardware/network/congestion/sharedbuffer-bufferchange.h>
#include <sstmac/hardware/network/congestion/sharedbuffer-flowregistration.h>
#include <sstmac/hardware/network/congestion/sharedbuffer-departure.h>
#include <sstmac/hardware/network/congestion/sharedbuffer-arrival.h>

#include <sstmac/hardware/network/topology/topology.h>
#include <sstmac/common/errors.h>
#include <sstmac/common/debug.h>

#include <math.h>

namespace sstmac
{
  namespace hw
  {

    

    // Make
    sharedbuffer::bandwidthchange::bandwidthchange(
        sharedbuffer::ptr parent, vertex_index v, edge_index e) :
        parent_(parent), v_(v), e_(e)
    {
    }

    // Construct
    sharedbuffer::bandwidthchange::ptr
    sharedbuffer::bandwidthchange::construct(sharedbuffer::ptr parent,
        vertex_index v, edge_index e)
    {
      return bandwidthchange::ptr (new bandwidthchange(parent, v, e));
    }

    // Destruct
    sharedbuffer::bandwidthchange::~bandwidthchange() throw ()
    {
    }

    // Handle
    void
    sharedbuffer::bandwidthchange::handle(
        const sst_message::ptr &msg)
    {

      if (parent_->dbg_.is_active())
        parent_->dbg_ << "bandwidthchange::handle: bw change on vertex "
            << v_.value() << " edge " << e_.value() << "\n";

      /// walk through flows affected by this buffer change and determine their bandwidth demands

      // this multimap allows a flow to be assiciated with its bw demand (as a double), where the flow is a pair
      // of its flow id and a pointer to its registration.
      // the multimap sorts the demands by increasing demand (used in the max-min bw allocation)
      std::multimap<double, std::pair<flow_id_t, flowregistration::ptr > > demands;

      // remember the flows waiting on a VC. they don't count in
      // the max-min fair share bw allocation
      int num_blocked_on_vc = 0;

      // remember link max bw
      double max_bw;

      // walk all flows at this vertex, looking for ones requesting this output edge
      for (std::map<flow_id_t, flowregistration::ptr >::iterator fr =
          parent_->flow_registrations_[v_.value()].begin();
          fr != parent_->flow_registrations_[v_.value()].end(); fr++)
            {

        // skip bw allocation to any flows not requesting this flow
        if (fr->second->output_edge_ != e_)
          continue;

        // maintain count of flows blocked on vc
        if (!fr->second->has_vc_)
          num_blocked_on_vc++;

        // perform buffer update
        parent_->update_buffer_usage(fr->second, parent_->now());

        // calculate new bandwidth demand
        max_bw = parent_->edge_max_bw_[fr->second->output_edge_.value()];

        double supply_rate = (
            fr->second->buffer_occupancy_ > 0.0 ?
                max_bw : fr->second->input_bw_);

        // determing if we're being held up by full buffer downstream
        double restricting_bw = std::max(parent_->node_router_bw_,
            parent_->router_router_bw_);

        vertex_index next_vertex = parent_->topology_->g().index_to_edge(
            fr->second->output_edge_).to();

        // if the flow has arrived downstream and is not ejection
        if (parent_->flow_registrations_[next_vertex.value()].find(fr->first)
            != parent_->flow_registrations_[next_vertex.value()].end()
            && !parent_->flow_registrations_[next_vertex.value()][fr->first]->ejection_)
          {

            parent_->update_buffer_usage(
                parent_->flow_registrations_[next_vertex.value()][fr->first],
                parent_->now());

            // if downstream buffer is full, can only push as hard as that buffer is draining
            // otherwise we can push as hard as we want (will just start filling buffer)
            if (parent_->flow_registrations_[next_vertex.value()][fr->first]->buffer_occupancy_
                > parent_->buffer_capacity_
                || parent_->are_equal(
                    parent_->flow_registrations_[next_vertex.value()][fr->first]->buffer_occupancy_,
                    parent_->buffer_capacity_))
              {

                parent_->flow_registrations_[next_vertex.value()][fr->first]->buffer_occupancy_ =
                    parent_->buffer_capacity_;
                restricting_bw =
                    parent_->flow_registrations_[next_vertex.value()][fr->first]->output_bw_;
              }

          }

        // keep track of this flow's demand
        double demand = std::min(restricting_bw, std::min(supply_rate, max_bw));
        std::pair<flow_id_t, flowregistration::ptr > this_flow_pair(
            fr->first, fr->second);
        demands.insert(
            std::pair<double,
                std::pair<flow_id_t, flowregistration::ptr > >(demand,
                this_flow_pair));

        if (parent_->dbg_.is_active())
          parent_->dbg_ << "bandwidthchange::handle: flow " << fr->first
              << " demanding " << demand << "\n";

      }

    // if no flows to rebalance, can quit here
    if (demands.size() == 0)
      return;

    // store bw yet to dole out
    parent_->edge_bw_[e_.value()] = max_bw;

    // keep track of which edges/vertices to issue bw change events to.
    // this eliminates redundant bw change events; since maps must have unique keys,
    // two bandwidth changes to the same edge (the map key) only result in one pair in the map
    std::map<edge_index, vertex_index> bw_changes_to_issue;

    // store how many demanding flows WITH VCs we've walked through
    int i = 0;

    // walk through demands, performing max-min fair share bw allocation
    for (std::multimap<double,
        std::pair<flow_id_t, flowregistration::ptr > >::iterator demand =
        demands.begin(); demand != demands.end(); demand++)
      {

      // demanding flow id
      flow_id_t fid = demand->second.first;
      // demanding flow registration
      flowregistration::ptr fr = demand->second.second;

      // max min allocation. either as much as we request, or our fair share.
      // ignore any flows blocked on a vc, they don't get a chance
      // at fair share bw allocation.
      // demands.size() - num_blocked_on_vc - i == num NOT blocked on vc left to be assigned bw
      double new_rate = 0.0;
      if (fr->has_vc_)
        {
          new_rate = std::min(
              demand->first,
              (parent_->edge_bw_[e_.value()]
                  / (demands.size() - num_blocked_on_vc - i)));

          // increment count of flows with a vc that have been allocated
          i++;
        }

      // we've given away this bandwidth
      parent_->edge_bw_[e_.value()] -= new_rate;

      if (parent_->dbg_.is_active())
        parent_->dbg_ << "bandwidthchange::handle: flow " << fid << " assigned "
            << new_rate << "\n";

      // alert a change in bandwidth assignment up/down stream ONLY if we've not just arrived at this router
      if (!fr->head_ && fabs(new_rate - fr->output_bw_) > parent_->bw_epsilon_)
        {

          // propagate bw change downstream
          vertex_index next_vertex = parent_->topology_->g().index_to_edge(
              fr->output_edge_).to();

          std::map<flow_id_t, flowregistration::ptr >::iterator it;
          it = parent_->flow_registrations_[next_vertex.value()].find(fid);

          if (it != parent_->flow_registrations_[next_vertex.value()].end()
              && !it->second->ejection_)
            {

              // set this output rate as downstream input rate
              it->second->input_bw_ = new_rate;
              parent_->update_buffer_usage(it->second, parent_->now());

              // add downstream vertex to list of ones to issue
              bw_changes_to_issue[it->second->output_edge_] = next_vertex;

            }

          // propagate bw change upstream
          // unless at injection node
          if (!fr->injection_)
            {

              vertex_index previous_vertex = parent_->topology_->g().index_to_edge(
                  fr->input_edge_).from();

              std::map<flow_id_t, flowregistration::ptr >::iterator it;
              it = parent_->flow_registrations_[previous_vertex.value()].find(
                  fid);

              if (it
                  != parent_->flow_registrations_[previous_vertex.value()].end())
                {

                  bw_changes_to_issue[it->second->output_edge_] =
                      previous_vertex;

                }
            }
        }

      // set output bw as new rate
      fr->output_bw_ = new_rate;

      /// determing buffer events

      // schedule buffer event if it could potentially fill (ie input bw > output bw)
      if (fr->input_bw_ - fr->output_bw_ > parent_->bw_epsilon_)
        {
          if (parent_->dbg_.is_active())
            parent_->dbg_
                << "bandwidthchange::handle: buffer fill being caused by input bw "
                << fr->input_bw_ << " output bw " << fr->output_bw_ << " cap "
                << parent_->buffer_capacity_ << " occ " << fr->buffer_occupancy_
                << "\n";

          // add 1e-12 (ie 1 tick) to ensure we go beyond full
          timestamp fill_time(
              parent_->now()
                  + timestamp(
                      (parent_->buffer_capacity_ - fr->buffer_occupancy_)
                          / (fr->input_bw_ - fr->output_bw_) + 1e-12));
          sharedbuffer::bufferchange::ptr bc =
              sharedbuffer::bufferchange::construct(parent_, v_, fid);
          parent_->schedule(fill_time, bc);

        }

      // schedule buffer event if buffer could potentially drain (ie output bw > input bw)
      // cannot drain from tail (will just depart)
      if (fr->output_bw_ - fr->input_bw_ > parent_->bw_epsilon_ && !fr->tail_)
        {
          if (parent_->dbg_.is_active())
            parent_->dbg_
                << "bandwidth::handle: buffer drain being caused by input bw "
                << fr->input_bw_ << " output bw " << fr->output_bw_ << "\n";

          // add 1e-12 (ie 1 tick) to ensure we go beyond empty
          timestamp drain_time(
              parent_->now()
                  + timestamp(
                      fr->buffer_occupancy_ / (fr->output_bw_ - fr->input_bw_)
                          + 1e-12));
          sharedbuffer::bufferchange::ptr bc =
              sharedbuffer::bufferchange::construct(parent_, v_, fid);
          parent_->schedule(drain_time, bc);

        }

      // if we're the tail, reschedule departure event
      if (fr->tail_ && fr->output_bw_ > parent_->bw_epsilon_)
        {

          // add 1e-12 (ie 1 tick) to ensure we go beyond empty
          timestamp departure_time(
              parent_->now()
                  + timestamp(fr->buffer_occupancy_ / fr->output_bw_ + 1e-12));
          departure::ptr dep = departure::construct(parent_, fid, v_);
          parent_->schedule(departure_time, dep);
          fr->expected_departure_time_ = departure_time;

        }

      // if we're the head and have been assigned bw, forward arrival event
      if (fr->head_ && fr->output_bw_ > parent_->bw_epsilon_)
        {
          fr->head_ = false;
          timestamp next_arrival(
              parent_->now() + parent_->link_lat_ + parent_->router_lat_);
          arrival::ptr arr = arrival::construct(parent_, fid,
              parent_->topology_->g().index_to_edge(fr->output_edge_).to(),
              fr->output_bw_, 0); // num bytes only matters at injection
          parent_->schedule(next_arrival, arr);

          if (parent_->dbg_.is_active())
            parent_->dbg_ << "bandwidthchange::handle: flow " << fid
                << " arriving at next vertex "
                << parent_->topology_->g().index_to_edge(fr->output_edge_).to().value()
                << " at " << next_arrival << "\n";

        }

    }

  // issue all bw change events we've built up
  for (std::map<edge_index, vertex_index>::iterator it =
      bw_changes_to_issue.begin(); it != bw_changes_to_issue.end(); it++)
    {

    bandwidthchange::ptr bwc = bandwidthchange::construct(parent_,
        it->second, it->first);
    parent_->schedule(parent_->now(), bwc);
  }

}

} } //end of namespace sstmac

