/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/network/congestion/sharedbuffer.h>
#include <sstmac/hardware/network/congestion/sharedbuffer-bufferchange.h>
#include <sstmac/hardware/network/congestion/sharedbuffer-bandwidthchange.h>
#include <sstmac/hardware/network/congestion/sharedbuffer-flowregistration.h>

#include <sstmac/hardware/network/topology/topology.h>
#include <sstmac/common/errors.h>
#include <sstmac/common/debug.h>

#include <math.h>

namespace sstmac
{
  namespace hw
  {

    

// Make
    sharedbuffer::bufferchange::bufferchange(sharedbuffer::ptr parent,
        vertex_index v, flow_id_t fid) :
        parent_(parent), v_(v), fid_(fid)
    {
    }

// Construct
    sharedbuffer::bufferchange::ptr
    sharedbuffer::bufferchange::construct(sharedbuffer::ptr parent,
        vertex_index v, flow_id_t fid)
    {
      return bufferchange::ptr(new bufferchange(parent, v, fid));
    }

// Destruct
    sharedbuffer::bufferchange::~bufferchange() throw ()
    {
    }

// Handle
    void
    sharedbuffer::bufferchange::handle(
        const sst_message::ptr &msg)
    {

      // find this flow's registration at this vertex.
      std::map<flow_id_t, flowregistration::ptr >::iterator fr =
          parent_->flow_registrations_[v_.value()].find(fid_);

      // see if we're actually here
      if (fr == parent_->flow_registrations_[v_.value()].end())
        {
          //throw illformederror("trying to make buffer change on vertex flow isn't registered at");
          return;
        }

      // First things first: update buffer usage to see if we've ACTUALLY filled/drained
      parent_->update_buffer_usage(fr->second, parent_->now());

      if (parent_->dbg_.is_active())
        parent_->dbg_
            << "bufferchange::handle: buffer status changing for flow " << fid_
            << " vertex " << v_.value() << " relative error "
            << (parent_->buffer_capacity_ - fr->second->buffer_occupancy_)
                / parent_->buffer_capacity_ << " input bw "
            << fr->second->input_bw_ << " output bw " << fr->second->output_bw_
            << " eps " << parent_->buffer_epsilon_ << " new occ "
            << fr->second->buffer_occupancy_ << " ibw var from max "
            << fabs(fr->second->input_bw_ - parent_->node_router_bw_)
                / parent_->node_router_bw_ << "\n";

      // handle fill effect
      // fills only occur when the input bw is greater than the output bw
      if (fr->second->input_bw_ - fr->second->output_bw_ > parent_->bw_epsilon_)
        {

          // if we've actually filled (ie buffer occupancy = buffer capacity)
          if (fr->second->buffer_occupancy_ > parent_->buffer_capacity_
              || parent_->are_equal(fr->second->buffer_occupancy_,
                  parent_->buffer_capacity_))
            {
              // reset occupancy. helps negate rounding effects.
              fr->second->buffer_occupancy_ = parent_->buffer_capacity_;

              if (parent_->dbg_.is_active())
                parent_->dbg_ << "bufferchange::handle: fill effect for flow "
                    << fid_ << " on vertex " << v_.value() << "\n";

              const std::vector<edge_index> &edges = parent_->flow_edges_[fid_];
              // only schedule upstream event if not at injection
              if (!fr->second->injection_)
                {
                  int i;
                  for (i = 0; i < edges.size(); i++)
                    if (parent_->topology_->g().index_to_edge(edges[i]).to() == v_)
                      break;

                  // send bandwidth change to the previous vertex, where it will know to stop sending.
                  vertex_index previous_vertex =
                      parent_->topology_->g().index_to_edge(edges[i]).from();
                  bandwidthchange::ptr bwc = bandwidthchange::construct(
                      parent_, previous_vertex, edges[i]);
                  parent_->schedule(parent_->now(), bwc);
                }

              fr->second->input_bw_ = fr->second->output_bw_;
            }
          else
            {
              // reschedule buffer event (it will fill eventually)
              // add 1e-12 (ie 1 tick) to ensure we go beyond full
              timestamp refill_time(
                  parent_->now()
                      + timestamp(
                          (parent_->buffer_capacity_
                              - fr->second->buffer_occupancy_)
                              / (fr->second->input_bw_ - fr->second->output_bw_)
                              + 1e-12));

              bufferchange::ptr refill = bufferchange::construct(parent_,
                  v_, fid_);
              parent_->schedule(refill_time, refill);

            }
        }

      // handle drain effect
      // drains only occur when output bw is greater than input bandwidth
      if (fr->second->output_bw_ - fr->second->input_bw_ > parent_->bw_epsilon_)
        {

          // if we've actually drained (ie buffer occupancy = 0)
          if (fr->second->buffer_occupancy_ < 0.0
              || parent_->are_equal(fr->second->buffer_occupancy_, 0.0))
            {
              // reset buffer occupancy. helps negate rounding errors.
              fr->second->buffer_occupancy_ = 0.0;

              if (parent_->dbg_.is_active())
                parent_->dbg_ << "bufferchange::handle: drain effect for flow "
                    << fid_ << " on vertex " << v_.value() << "\n";

              bandwidthchange::ptr bwc = bandwidthchange::construct(
                  parent_, v_, fr->second->output_edge_);
              parent_->schedule(parent_->now(), bwc);

            }
          else
            {
              // reschedule buffer event (it will drain eventually)
              // add 1e-12 (ie 1 tick) to ensure we go beyond empty
              timestamp redrain_time(
                  parent_->now()
                      + timestamp(
                          fr->second->buffer_occupancy_
                              / (fr->second->output_bw_ - fr->second->input_bw_)
                              + 1e-12));

              bufferchange::ptr redrain = bufferchange::construct(
                  parent_, v_, fid_);
              parent_->schedule(redrain_time, redrain);

            }

        }

    }

  }
} //end of namespace sstmac

