/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/network/congestion/sharedbuffer.h>
#include <sstmac/hardware/network/congestion/sharedbuffer-departure.h>
#include <sstmac/hardware/network/congestion/sharedbuffer-flowregistration.h>
#include <sstmac/hardware/network/congestion/sharedbuffer-bandwidthchange.h>

#include <sstmac/common/errors.h>
#include <sstmac/common/debug.h>
#include <map>
#include <vector>
#include <queue>

namespace sstmac
{
  namespace hw
  {

// Maker.
    sharedbuffer::departure::departure(sharedbuffer::ptr parent, flow_id_t fid,
        vertex_index v) :
        parent_(parent), fid_(fid), v_(v)
    {
    }

// Constructor.
    sharedbuffer::departure::ptr
    sharedbuffer::departure::construct(sharedbuffer::ptr parent, flow_id_t fid,
        vertex_index v)
    {
      return departure::ptr(new departure(parent, fid, v));
    }

// Destructor.
    sharedbuffer::departure::~departure() throw ()
    {
    }

// Handler.
    void
    sharedbuffer::departure::handle(const sst_message::ptr &msg)
    {

      // trying to leave a link we were never at
      std::map<flow_id_t, flowregistration::ptr>::iterator fr =
          parent_->flow_registrations_[v_.value()].find(fid_);
      if (fr == parent_->flow_registrations_[v_.value()].end())
        {
          //throw illformederror("trying to leave a link flow is not registered at");
          return;
        }

      // this is not when we expect to depart
      if (fr->second->expected_departure_time_ != timestamp(-1)
          && parent_->now() != fr->second->expected_departure_time_)
        return;

      // trying to depart twice
      if (parent_->vertices_departed_[v_.value()].find(fid_)
          != parent_->vertices_departed_[v_.value()].end())
        {
          if (parent_->dbg_.is_active())
            parent_->dbg_ << "sharedbuffer::departure::handle: "
                "trying to leave a vertex twice.\n";
          return;
          //throw illformederror("trying to leave a link twice");
        }

      // we're the tail if we know we're departing
      fr->second->tail_ = true;



      const std::vector<edge_index> &edges = parent_->flow_edges_[fid_];

      // delay departure until buffer has drained
      // perform buffer update only if not at ejection
      if (!fr->second->ejection_)
        {
          parent_->update_buffer_usage(fr->second, parent_->now());

          // we're no longer being pushed from upstream
          fr->second->input_bw_ = 0.0;

          // are there still bytes in the buffer that need to drain?
          if (fr->second->buffer_occupancy_ > 0.0)
            {
              if (fr->second->output_bw_ > parent_->bw_epsilon_)
                {
                  if (parent_->dbg_.is_active())
                    parent_->dbg_
                        << "departure::handle: departure is being rescheduled for flow "
                        << fid_ << " vertex " << v_.value() << " due to "
                        << fr->second->buffer_occupancy_
                        << " bytes in buffer. input bw: "
                        << fr->second->input_bw_ << " output bw: "
                        << fr->second->output_bw_ << "\n";
                  // add 1e-12 (ie 1 tick) to ensure we go beyond empty
                  timestamp redo_departure(
                      parent_->now()
                          + timestamp(
                              fr->second->buffer_occupancy_
                                  / fr->second->output_bw_ + 1e-12));
                  departure::ptr rdep = departure::construct(parent_, fid_, v_);
                  parent_->schedule(redo_departure, rdep);
                  fr->second->expected_departure_time_ = redo_departure;
                }

              // have rescheduled for the future. time to stop handling.
              return;
            }

        }

      if (parent_->dbg_.is_active())
        parent_->dbg_ << "departure::handle: flow " << fid_
            << " is departing from vertex " << v_.value() << "\n";

      // schedule bw change and update vc allocation for this flow's output edge
      // (all flows have output ports unless at ejection)
      if (!fr->second->ejection_)
        {

          // this departing flow frees a virtual channel for other requesting flows.
          // have to give up this vc to any flows blocking for one
          if (!parent_->flows_blocked_on_vc_[fr->second->output_edge_.value()].empty())
            {

              // pull the blocked flow from the queue
              flow_id_t blocked_flow =
                  parent_->flows_blocked_on_vc_[fr->second->output_edge_.value()].front();
              parent_->flows_blocked_on_vc_[fr->second->output_edge_.value()].pop();

              if (parent_->dbg_.is_active())
                parent_->dbg_ << "departure::handle: vc at vertex "
                    << v_.value() << " edge "
                    << fr->second->output_edge_.value()
                    << " is being reassigned to flow " << blocked_flow << "\n";

              // give it the vc
              parent_->flow_registrations_[v_.value()][blocked_flow]->has_vc_ =
                  true;
            }
          else
            {
              // the vc has become available
              parent_->available_vcs_[fr->second->output_edge_.value()]++;
            }

          // send bw event to this vertex for all flows using this flow's output edge
          bandwidthchange::ptr bwc = bandwidthchange::construct(parent_, v_,
              fr->second->output_edge_);
          parent_->schedule(parent_->now(), bwc);

        }

      // delete flow state
      bool ejecting = fr->second->ejection_;
      bool injecting = fr->second->injection_;
      parent_->flow_registrations_[v_.value()].erase(fid_);

      // injection: notify sender
      if (injecting)
        {
          if (parent_->recv_messages_[fid_]->get_needs_ack())
            {
              sst_message::ptr sendack = parent_->recv_messages_[fid_]->clone(
                  sst_message::SEND_ACK);
              parent_->schedule(parent_->now(), parent_->send_handlers_[fid_],
                  sendack);
            }
          if (parent_->dbg_.is_active())
            parent_->dbg_ << "sharedbuffer::departure::handle: "
                "departing injection port.\n";
          parent_->send_handlers_.erase(fid_);
        }

      // ejection: notify receiver
      if (ejecting)
        {
          timestamp recv_time = parent_->now() + parent_->node_lat_;
          parent_->schedule(recv_time, parent_->recv_handlers_[fid_],
              parent_->recv_messages_[fid_]);
          if (parent_->dbg_.is_active())
            parent_->dbg_ << "sharedbuffer::departure::handle: "
                "departing ejection port. flow " << fid_ << " recv complete!\n";
          parent_->recv_handlers_.erase(fid_);
          parent_->recv_messages_.erase(fid_);
          /** We no longer need the flow - delete it */
          parent_->flow_edges_.erase(fid_);
          return;
        }

      // find edge index
      int i;
      for (i = 0; i < edges.size(); i++)
        {
          if (parent_->topology_->g().index_to_edge(edges[i]).to() == v_)
            break;
        }

      // if at injection, we want the very first edge
      if (i == edges.size())
        i = -1;

      // everyone but ejection forwards departure event
      timestamp next_departure = parent_->now() + parent_->router_lat_
          + parent_->link_lat_;
      departure::ptr dep = departure::construct(parent_, fid_,
          parent_->topology_->g().index_to_edge(edges[i + 1]).to());
      parent_->schedule(next_departure, dep);

      vertex_index next_vertex = parent_->topology_->g().index_to_edge(
          edges[i + 1]).to();

      // record expected departure time downstream if the flow has arrived there
      if (parent_->flow_registrations_[next_vertex.value()].find(fid_)
          != parent_->flow_registrations_[next_vertex.value()].end())
        parent_->flow_registrations_[next_vertex.value()][fid_]->expected_departure_time_ =
            next_departure;

    }

  }
} //end of namespace sstmac

