/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/network/congestion/throttle_flower.h>
#include <sstmac/common/errors.h>

#include <boost/lexical_cast.hpp>
#include <iterator>
#include <iostream>
#include <iomanip>
#include <algorithm>
#include <stdlib.h>



namespace sstmac { namespace hw  {

  throttle_flower::ptr
  throttle_flower::construct(const topology::ptr &net,
                               double node_router_bw, double router_router_bw)
  {
    throttle_flower::ptr self(new throttle_flower(net,
                                       node_router_bw, router_router_bw));
    return self;
  }

  throttle_flower::throttle_flower(const topology::ptr &net,
                                       double node_router_bw, double router_router_bw):
    typical_flower(net, node_router_bw, router_router_bw)
  {
  }

  throttle_flower::~throttle_flower() throw()
  {
  }

  void
  throttle_flower::rebalance_flow_start(flow_id_t f)
  {
    double new_flow_bw = max_bw_;
    const std::vector<edge_id_t> &edges = flow_edges_[f];
    for (unsigned long i=0; i<edges.size(); i++) {
      edge_id_t edge = edges[i];
      double bw_req = new_flow_bw + edge_bw_utilization(edge);
      SSTMAC_DEBUG << "edge_bw_utilization = "
                          << edge_bw_utilization(edge) << "\n";
      if (edge_bw_[edge] < bw_req) {
        double scale_factor = edge_bw_[edge]/bw_req;
        new_flow_bw *= scale_factor;
        // Slow all of the flows through the overcommitted edge.  Note
        // that the flow rates satisfying the constraints are not
        // unique, and the result of this algorithm depends on the
        // order with which we iterate through the edges.
        for (flow_set_t::iterator iter=edge_flows_[edge].begin();
             iter!=edge_flows_[edge].end(); iter++) {
          SSTMAC_DEBUG << "Edge " << edge
                    << " adjusting flow " << *iter
                    << " old rate = " << get_flow_bw(*iter);
          if (*iter != f) scale_flow_bw(*iter, scale_factor);
          SSTMAC_DEBUG << " new rate = " << get_flow_bw(*iter)
                    << "\n";
          SSTMAC_DEBUG<< "edge_bw_utilization = "
                              << edge_bw_utilization(edge) << "\n";
        }
      }
    }
    set_flow_bw(f, new_flow_bw);
  }

  void
  throttle_flower::rebalance_flow_stop(flow_id_t oldf)
  {
  }

} } // end of namespace sstmac
