/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_HARDWARE_NETWORK_CONGESTION_THROTTLE_FLOWER_H_INCLUDED
#define SSTMAC_HARDWARE_NETWORK_CONGESTION_THROTTLE_FLOWER_H_INCLUDED

#include <sstmac/hardware/network/congestion/typical_flower.h>
#include <sstmac/common/timestamp.h>
#include <sstmac/hardware/network/topology/topology.h>
#include <vector>
#include <map>
#include <set>
#include <string>
#include <sstmac/common/ptr_type.h>

namespace sstmac { namespace hw  {
  
  /**
   * Allocate bandwidth by scaling back flows on overcomitted links.
   *
   * This allocates bandwidth between flows. When a new flow is added
   * it is given the smallest maximum bandwidth provided by any link
   * used by the flow. If, after consideration of all other flows in
   * the system, any link is overutilized, then all flows through that
   * link are throttled back using the the scale factor for all the
   * flows. While fast, this approach is experimental and gives
   * anomalous results in some cases.  The order_flower is recommended
   * over throttle_flower.
   */
  class throttle_flower: public typical_flower {

  protected:
    throttle_flower(const topology::ptr &net,
           double node_router_bw, double router_router_bw);

    /// Override of typical_flower::rebalance_flow_start
    void rebalance_flow_start(flow_id_t f);

    /// Override of typical_flower::rebalance_flow_stop
    void rebalance_flow_stop(flow_id_t oldf);

  public:
    typedef boost::intrusive_ptr<throttle_flower> ptr;

    /// Construct a throttle_flower.
    /// \param node_router_bw maximum bandwidth on network links each
    ///        node and router
    /// \param router_router_bw maximum bandwidth on network links each
    ///        router
    /// \return a shared_ptr containing the new object
    static throttle_flower::ptr
    construct(const topology::ptr &net,
              double node_router_bw, double router_router_bw);

    virtual ~throttle_flower() throw();
  };

} } // end of namespace sstmac

#endif
