/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */


// fattree.h: Interface for fat tree networks.
//
// Author: Curtis Janssen <cljanss@ca.sandia.gov>

#ifndef SSTMAC_HARDWARE_NETWORK_TOPOLOGY_FATTREE_H_INCLUDED
#define SSTMAC_HARDWARE_NETWORK_TOPOLOGY_FATTREE_H_INCLUDED

#include <sstmac/hardware/network/topology/topology.h>

namespace sstmac { namespace hw  {

  /** The fattree network generates a k-ary fat tree with l tiers.  In the fat
      tree, not all vertices are nodes. The inputs for this network
      can specify the following attributes:

      <table border="1">

      <tr><td>Keyword<td>Type<td>Default<td>Description

      <tr><td><tt>l</tt><td>int<td>2<td>The number of levels in the fat tree.

      <tr><td><tt>k</tt><td>int<td>2<td>Half the number of links a switch
      element has.

      <tr><td><tt>nndpersw</tt><td>int<td>0<td>The number of nodes per switch.

      <tr><td>width<td>int<td>0<td>The width of edges connecting two switches.

      </table>
  */

  class fattree: public topology {
    int width;
  private:
    int l_, k_;
    void init_fattree(int l, int k, int nndpersw, int widtharg);
  protected:
    //fattree(TiXmlHandle);
    fattree(int l, int k, int nndpersw, int widtharg);

  public:
    typedef boost::intrusive_ptr<fattree> ptr;

    static fattree::ptr construct(int levels=2, int half_links=2,
                                                int nndpersw=0, int widtharg=1);
    virtual ~fattree();

    static size_t size(int l, int k);
    void shortest_path(const vertex&v1, const vertex&v2, walk&w) const;
    void route(const vertex &v1, const vertex &v2, walk &w) const;
    int l() const;
    int k() const;
    int64_t diameter() const;
    double a() const;
    int64_t b() const;
    bool vertex_symmetric() const;
    std::string name() const;
    int edge_width(edge_index e) const;
    bool is_switch(vertex_index v) const;
  };

} } //end of namespace sstmac

#endif
