/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */


// line.cc: Implementation of line (1D mesh) networks.
//
// Author: (ring) Curtis Janssen <cljanss@ca.sandia.gov>
// Adaption to 1D mesh by David Evensky <evensky@sandia.gov>

#include <sstream>
#include <sstmac/hardware/network/topology/line.h>
#include <sstmac/hardware/network/topology/walk.h>

using std::logic_error;
using std::set;
using std::vector;
using std::string;

namespace sstmac { namespace hw  {

  static void
  compute_vertices(int nx, set<vertex> &vertices)
  {
    vertex v(1);
    for (int i=0; i<nx; i++) {
      v[0] = i;
      vertices.insert(v);
    }
  }

  static void
  compute_edges(int nx, set<edge> &edges)
  {
    vertex vi(1);
    vertex vo(1);
    for (int i=0; i<nx; i++) {
      vi[0] = i;
      vo[0] = i+1;
      if (vo[0] < nx)
        edges.insert(edge(vi,vo)); // fwd link.
      vo[0] = i-1;
      if (vo[0] >= 0)
        edges.insert(edge(vi,vo)); // back link.
    }
  }

  // static sstmac::xmlclass::register_class reg("sstmac::line",
  //                               sstmac::xmlclass::ctor_template<line>);

  // line::line(TiXmlHandle h)
  // {
  //   int nx = 2;
  //   h.ToElement()->QueryIntAttribute("nnode",&nx);
  //   init_line(nx);
  // }

  line::line(int nx)
  {
    mesh_routing =1;
    init_line(nx);
  }

  void
  line::init_line(int nx)
  {
    nx_ = nx;
    compute_vertices(nx_, nodes_);

    set<edge> edges;
    compute_edges(nx_, edges);

    g_.init(nodes_, edges);
    init();
  }

  line::ptr line::construct(int nodes) {
    return line::ptr(new line(nodes));
  }

  line::~line() {
  }

  void
  line::shortest_path(const vertex &v1, const vertex &v2, walk &w) const
  {
    vector<vertex_index> vertices;

    int a = v1[0];
    int b = v2[0];

    vertex v(1);
    int diff = b - a;
    // AP   update 
    if (mesh_routing) {
      if (diff > 0) {
        for (; a<=b; a++) {
          v[0] = a;
          vertices.push_back(g().vertex_to_index(v));
        }
      }
      else if (diff != 0) {
        for (; a>=b; a--) {
          v[0] = a;
          vertices.push_back(g().vertex_to_index(v));
        }
      } // vertices is empty if we're going nowhere.
    }
    else { // not using mesh routing on a mesh....
      throw logic_error("line::shortest_path() not using mesh routing on a 1D mesh aka line.");
    } 
    w.init(vertices);
  }

  int64_t
  line::diameter() const
  {
    return n(); // should be same as nx_, I think.
  }

  int64_t
  line::b() const
  {
    return (nx_ < 2 ? 0 : 2); // we are bidirectional so 2 edges.
  }

  bool
  line::vertex_symmetric() const
  {
    return (nx_ == 1 || nx_ == 2 ? true : false);
  }

  string
  line::name() const
  {
    std::ostringstream ostr;
    ostr << "Line" << "(" << nx_ << ")";
    return ostr.str();
  }

} } //end of namespace sstmac
