/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_BACKENDS_NATIVE_COMPONENTS_NIC_NETWORKINTERFACE_H_INCLUDED
#define SSTMAC_BACKENDS_NATIVE_COMPONENTS_NIC_NETWORKINTERFACE_H_INCLUDED

#include <sstmac/hardware/node/node.h>
#include <sstmac/common/timestamp.h>
//#include <sstmac/process/server.h>
//#include <sstmac/backends/native/nodeid.h>
#include <sstmac/hardware/network/congestion/interconnect.h>
#include <sstmac/common/eventmodule.h>
#include <sstmac/common/debug.h>
#include <sstmac/common/ptr_type.h>
#include <sstmac/hardware/failure/failable.h>
#ifdef SSTMAC_HAVE_BOOST_1_37
#include <boost/unordered_map.hpp>
#else
#include <map>
#endif

namespace sstmac
{
  namespace hw
  {

    // Important types.
    // class networkdata;
    //class eventhandler;
    class node;

    /**
     * A networkinterface is a delegate between a node and a server module.
     * This object helps ornament network operations with information about
     * the process (ppid) involved.
     */
    class networkinterface : public eventmodule, public failable
    {

      /// The network servers for parallel processes running on this node.
      //typedef std::map<ppid, boost::shared_ptr<server> > servermap_t;
      // servermap_t servers_;

    protected:
      networkinterface(const node::ptr &parent,
          const sstmac::hw::interconnect::ptr &interconn);

      static debug dbg_;

      sstmac::hw::interconnect::ptr interconn_;
      node::ptr parent_;

    public:
      typedef boost::intrusive_ptr<networkinterface> ptr;

      nodeaddress::ptr
      get_parent_id()
      {
        return parent_->get_id();
      }

      virtual void
      set_eventmanager(const eventmanager::ptr &m)
      {

        eventmodule::set_eventmanager(m);

        if (!spy_)

          {
            spy_ = stat_spyplot::construct("<stats> spyplot", "network_spyplot.csv");
            register_stat(spy_);

            spy_bytes_ = stat_spyplot::construct("<stats> spyplot", "network_spyplot_bytes.csv");
            register_stat(spy_bytes_);
          }
      }

      virtual long
      failure_id() const
      {
        return parent_->get_id()->unique_id();
      }

      /// Goodbye.
      virtual
      ~networkinterface();

      /// Registers a server for the given ppid.
      /// \param ppidarg is the ppid for the new server
      /// \param srvr is the new server
      /// \throws illformederror if a server already exists for the ppid
      /* virtual void
       register_server(ppid caller,
       const boost::shared_ptr<server> &srvr);

       /// Unregisters a server for the given ppid.
       /// \param ppidarg is the ppid for the new server
       /// \throws nullerror if a server does not already exist for the ppid
       virtual void
       unregister_server(ppid caller);

       /// Test whether the given ppid has defined a server on this node.
       /// \param caller is the ppid for the sought-after server
       /// \return true if and only if a server is registered for caller
       virtual bool
       has_server(ppid caller) const;

       /// Get the network server for the given ppid.
       /// \param caller is the desired ppid
       /// \throws nullerror no server is registered for the ppid
       virtual boost::shared_ptr<server>
       get_server(ppid caller);



       /// Incoming network transfer.
       /// Originates from a network simulator and is forwarded to the
       /// appropriate server.
       /// \param start_time is the time of message arrival
       /// \param caller is the ppid of the target server
       /// \param sender is the id of the sending node
       /// \param load is the network data
       /// \throws nullerror if the server is not registered
       virtual void
       incoming_message(const sstmac::sw::timestamp &start_time, ppid caller, nodeid sender,
       const boost::shared_ptr<networkdata> &load);*/

      /// Schedule a send operation to happen at the given time.
      /// \param start_time is the time when the send is started.
      /// \param ppidarg is the parallel process id of the source and target.
      /// \param recver is the global index of the receiver node.
      /// \param payload is the data sent over the network
      /// \param completion is the eventhandler signaled when the send is complete
      virtual void
      send(const nodeaddress::ptr &recver, const sst_message::ptr &payload,
          const eventhandler::ptr &completion) = 0;

      /// Get a new nic of the same type as the current one.
      virtual networkinterface::ptr
      clone(const boost::intrusive_ptr<node> &parent,
          const sstmac::hw::interconnect::ptr &interconn) const = 0;

      static void
      set_node_map(nodeaddress::ptr id, const networkinterface::ptr &n);

      virtual std::string
      to_string() const
      {
        return "nic";
      }

      virtual sstmac::hw::topology::ptr
      get_topology()
      {
        if (interconn_)
          return interconn_->get_topology();
        else
          return sstmac::hw::topology::ptr();
      }

    protected:
#ifdef SSTMAC_HAVE_BOOST_1_37
      typedef boost::unordered_map<nodeaddress::ptr, networkinterface::ptr>
          map_type;
#else
      typedef std::map<nodeaddress::ptr, networkinterface::ptr, nodeaddress::nodeaddresscomp> map_type;
#endif

      static map_type node_map_;

      static stat_spyplot::ptr spy_;
      static stat_spyplot::ptr spy_bytes_;

    };

  }
} // end of namespace sstmac.

#endif
