/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/processor/simpleprocessor-config.h>
#include <sstmac/hardware/processor/instructionprocessor-config.h>
#include <sstmac/hardware/memory/simple_memorymodel.h>
#include <sstmac/hardware/node/detailednode.h>
#include <sstmac/hardware/processor/pinnedcomputescheduler.h>
#include <sstmac/hardware/memory/mc_memorymodel.h>

#include <sstmac/hardware/nic/networkinterface.h>

#include <sstmac/software/libraries/filesystem/disk_message.h>
#include <sstmac/software/libraries/filesystem/nppfs_server.h>

#include <sstmac/common/errors.h>
#include <sstmac/software/services/launch/machine_descriptor.h>
#include <iostream>
#include <sstream>

namespace sstmac
{
  namespace hw
  {

    detailednode::detailednode(const sim_parameters::ptr &params) :
      simplenode(params)
    {
      os_->setAMI(sstmac::sw::AMI::AMI_COMM, this);
      os_->setAMI(sstmac::sw::AMI::AMI_COMPUTE, this);
      os_->setAMI(sstmac::sw::AMI::AMI_QUERY, this);

      cores_ = params->get_int_param("node_cores");
      memory_latency_ = params->get_time_param("node_mem_latency");
      bool errorflag = false;
      memory_bandwidth_ = get_bandwidth(
          params->get_param("node_mem_bandwidth").c_str(), errorflag);
      freq_ = params->get_double_param("node_frequency");

      proc_link_bw_ = get_bandwidth(
          params->get_param("proc_link_bandwidth").c_str(), errorflag);
      proc_link_lat_ = params->get_time_param("proc_link_latency");

      num_mc_ = params->get_int_param("proc_num_mc");
      num_sock_ = params->get_int_param("node_num_procs");

      timestamp pre(-1);
      if (params->has_param("node_preemption"))
        {
          pre = params->get_time_param("node_preemption");
        }
      preempt_ = pre;

      SSTMAC_DEBUG << "constructed a detailednode with parameters mem_lat = "
            << memory_latency_ << ", memory_bandwidth = " << memory_bandwidth_
            << ", freq = " << freq_ << ", cores = " << cores_ << "\n";
    }

    void
    detailednode::initialize(nodeaddress::ptr my_id, eventmanager::ptr man)
    {

      node::initialize(my_id, man);

      SSTMAC_DEBUG << "starting detailednode initialization \n";

      double tcycle = 1.0 / (freq_);

      timestamp tflop(tcycle); // 10 clock cycles for a floating point
      //timestamp tmemop(memory_latency_.sec() + 1.0 / (memory_bandwidth_));
      timestamp tmemop = tflop;

      sstmac::hw::simpleprocessor_config::ptr cfg =
          sstmac::hw::simpleprocessor_config::construct(cores_, freq_, 10, 30);

      compute_sched_ = sstmac::hw::pinnedcomputescheduler::construct(
          my_id->unique_id(), cores_, preempt_, num_sock_);
      compute_sched_->set_eventmanager(man);
      proc_model_ = sstmac::hw::simpleprocessor::construct(cfg, msg_handler_,
          compute_sched_);

      sstmac::hw::instructionprocessor_config::ptr pcfg =
          sstmac::hw::instructionprocessor_config::construct(cores_,
              INST_FLOP_LABEL, tflop, INST_MEMOP_LABEL, tmemop,
              INST_MEMOP_LOCAL_LABEL, tmemop, INST_MEMOP_RAND_LABEL, tmemop);

      proc_inst_ = sstmac::hw::instructionprocessor::construct(pcfg,
          msg_handler_, compute_sched_);

      mem_model_ = sstmac::hw::mc_memorymodel::construct(memory_bandwidth_,
          memory_latency_, proc_link_bw_, proc_link_lat_, num_mc_, num_sock_,
          msg_handler_, my_id->unique_id(), compute_sched_);
      mem_model_->set_eventmanager(man);

      if (params_ && params_->has_param("disk_type"))
        {
          if (params_->get_param("disk_type") == "vesper")
            {
              disk_ = sstmac::hw::disk_controller::construct(
                  my_id->unique_id(), params_, msg_handler_, man);
            }
        }

      if (params_ && params_->has_param("node_nppfs_server"))
        {
          long nid = params_->get_long_param("node_nppfs_server");
          nodeaddress::ptr addr = nodeaddress::convert_from_uniqueid(nid);
          nppfs_server_ = addr;

          std::string fname = params_->get_param("node_nppfs_diskfile");
          sstmac::sw::nppfs_server::ptr serv =
              sstmac::sw::nppfs_server::construct(fname, addr->equals(nodeid_));
          os_->register_ser(serv);
          serv->start();

        }

      SSTMAC_DEBUG << "detailednode initialization done, starting things \n";
      os_->start_apps();

      launcher_->start();
    }

    node::ptr
    detailednode::clone() const
    {
      node::ptr ret = construct(params_);

      return ret;
    }



  }
} // end of namespace sstmac.
