/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_BACKENDS_NATIVE_COMPONENTS_NODE_NODE_H_INCLUDED
#define SSTMAC_BACKENDS_NATIVE_COMPONENTS_NODE_NODE_H_INCLUDED

#include <sstmac/common/timestamp.h>
#include <sstmac/software/process/taskid.h>
//#include <sstmac/backends/native/nodeid.h>

#include <sstmac/common/ptr_type.h>

#include <sstmac/common/eventmodule.h>
#include <sstmac/software/ami/ami.h>
#include <sstmac/software/process/operatingsystem.h>
#include <sstmac/common/debug.h>

#include <sstmac/software/services/launch/launcher.h>
#include <sstmac/software/services/launch/launchinfo.h>
#include <sstmac/software/services/launch/allocation/allocationstrategy.h>

#include <sstmac/hardware/failure/failable.h>

#include <list>
#include <string>

namespace sstmac
{
  namespace hw
  {

    // Forward declarations of relevant types.

    class networkinterface;

    /**
     * An interface for "physical" nodes -- may eventually be decorated
     * with additional functionality (OS/cpu modeling or whatever).
     */
    class node : public eventmodule, public sstmac::sw::AMI
    {

      friend class message_handler;

    public:
      typedef boost::intrusive_ptr<node> ptr;

    protected:
      class message_handler : public eventhandler
      {
      protected:
        node::ptr parent_;

        message_handler(const node::ptr &p) :
            parent_(p)
        {

        }
          
         

      public:
        virtual void
        handle(const sst_message::ptr &msg)
        {
          if (parent_ == 0)
            {
              throw ssterror("node::message_handler - my parent is null!");
            }
          if (parent_->os_ == 0)
            {
              throw ssterror("node::message_handler - my parent's os is null!");
            }
          SSTMAC_DEBUG << "node[" << parent_->nodeid_->to_string()
              << "]: handling message \n";
        parent_->os_->handle_message(msg);

            
        
            if(parent_->has_failed()){
                //reset hardware to be not-failed (self-healing)
                parent_->reset_failed();
                              
                parent_->launcher_->apps_failed();
                
            }

            parent_->os_->cleanup();
      }

      static boost::intrusive_ptr<message_handler>
      construct(const node::ptr &p)
      {
        boost::intrusive_ptr<message_handler> ret(new message_handler(p));
        return ret;
      }

      virtual std::string
      to_string() const
      {
        return "node::message_handler";
      }
      };
        
        virtual bool has_failed();
        virtual void reset_failed();

      boost::intrusive_ptr<message_handler> msg_handler_;

      sstmac::sw::environment::ptr env_;
      sstmac::sw::operatingsystem::ptr os_;

      /// The id for this node.
      nodeaddress::ptr nodeid_;

      sstmac::sw::launcher::ptr launcher_;


      static debug dbg_;

      sim_parameters::ptr params_;

      node(const sim_parameters::ptr &params);

      static sstmac::sw::allocationstrategy::ptr allocation_;


      boost::intrusive_ptr<networkinterface> nic_;
    public:

      /// Get a new node of the same type as the current one.
      virtual node::ptr
      clone() const = 0;

      /// Assign a node index and event call interface for this node.
      /// \throw nullerror if ev is null.
      virtual void
      initialize(nodeaddress::ptr my_id,
          eventmanager::ptr man);

      virtual void
      set_nic(const boost::intrusive_ptr<networkinterface> &n);

      virtual boost::intrusive_ptr<networkinterface>
      get_nic() const;

      virtual void
      handle(const sst_message::ptr &msg);

      virtual std::string
      to_string() const
      {
        std::stringstream ss;
        ss << "node(" << nodeid_->to_string() << ")";
        return ss.str();
      }

      nodeaddress::ptr
      get_id()
      {
        return nodeid_;
      }

      virtual std::list<failable::ptr>
      get_failure_components() = 0;


      void
      add_service(sstmac::sw::service::ptr serv);

      //virtual void
      //init_failure_model(double l_proc, double l_mem, double l_disk, double l_nic, failable::ptr ps, double p_ps_node) = 0;

    };

  }
} // end of namespace sstmac

#endif
