/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.


#ifndef SSTMAC_BACKENDS_NATIVE_COMPONENTS_NODE_NODEINTERFACE_H_INCLUDED
#define SSTMAC_BACKENDS_NATIVE_COMPONENTS_NODE_NODEINTERFACE_H_INCLUDED

#include <sstmac/simulator/timestamp.h>
#include <sstmac/simulator/ppid.h>
#include <sstmac/simulator/taskid.h>
#include <sstmac/simulator/eventmanager.h>
#include <sstmac/util/nocopy.h>
#include <sstmac/common/ptr_type.h>


namespace sstmac {

  // Important types.
  class node;
  class server;
  class kernel;
  class eventhandler;
  class instructions;

  /**
   * Interface type to give a kernel object access to the node that it runs on.
   * This permits kernels to use the zero-based indexing of nodes used by
   * a process (as exposed by an environment object) and map it onto the
   * index nodes used by a manager and the nodes themselves.
   *
   * Each nodeinterface object corresponds to exactly one node object.
   * The nodeinterface maintains a pointer to the node and probably a pointer
   * to the environment mapping object as well.  The nodeinterface should do
   * no short-circuiting outside of those three object types.

  class nodeinterface : virtual public nocopy {
    /// This object.
    boost::weak_ptr<nodeinterface> self_;

    /// My node.
    boost::weak_ptr<node> node_;

    /// The eventmanager (intended to make now() take less time).
    eventmanager::ptr mgr_;

    /// The ppid of my process.
    ppid ppid_;

    /// The taskid for which I am working.
    taskid taskid_;

  protected:
    /// Hi there.
    nodeinterface(const node::ptr &my_node,
                  const ppid &my_ppid,
                  const taskid &my_taskid);

    /// Who am I?
    void set_self(const boost::shared_ptr<nodeinterface> &self);

    /// This is who I am.
    boost::shared_ptr<nodeinterface> get_self();

  public:
    /// Hello.
    static boost::shared_ptr<nodeinterface>
    construct(const node::ptr &my_node,
              const ppid &my_ppid,
              const taskid& my_taskid);

    /// Goodbye.
    virtual ~nodeinterface() throw();
    
    /// Get the ppid associated with this parallel simulation.
    virtual const ppid& get_ppid() const;

    /// Get the taskid associated with this peer instance.
    virtual const taskid& get_taskid() const;

    /// Get current time from the node.
    const timestamp& now() const {
      return mgr_->now();
    }

    node::ptr get_node() const { return node_.lock(); }

    /// Set a simple trigger (wake-up call) for this kernel.
    virtual void schedule(const timestamp &start_time,
                          const eventhandler::ptr &event);

    /// Start a new kernel at the given time.
    /// The kernel will be started using kernel::start and wil be given
    /// access to this (or an equivalent) node interface object.
    virtual void add(const timestamp &start_time,
                     const boost::shared_ptr<kernel> &task);

    /// Pretend to compute for the given number of cpu seconds
    /// in a single thread.
    virtual void compute(const timestamp &start_time,
                         const timestamp &cputime_per_thread,
                         const eventhandler::ptr &completion);

    /// Pretend to compute for the given number of cpu seconds
    /// on each of the given number of threads.
    /// Wakeup-call is received when the last thread completes its work.
    virtual void compute(const timestamp &start_time,
                         const timestamp &cputime_per_thread,
                         int threads,
                         const eventhandler::ptr &completion);

    /// Pretend to compute for the given number of cpu seconds.
    //virtual void compute(const timestamp &start_time,
    //                     ppid caller,
    //                     const timestamp &cputime,
    //                     eventhandler::ptr completion);

    /// I am not sure yet how best to handle this one :)
    /// Eventually, this is where we would bridge over to SST/microscale
    /// or other processor models.
    virtual void compute(const timestamp &start_time,
                         const boost::shared_ptr<instructions> &instr,
                         const eventhandler::ptr& completion);

    /// Request a copy of the given number of bytes between two address spaces.
    /// Used by messaging models to assess cost of non-network 
    /// data transfers.
    /// \param start_time time when copy is initiated.
    /// \param bytes number of bytes to copy.
    /// \return time of completion.
    virtual void copy(const timestamp &start_time,
                      int64_t bytes,
                      const eventhandler::ptr &completion);
  };
    
} } // end of namespace sstmac.

#endif

*/
