/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/processor/simpleprocessor-config.h>
#include <sstmac/hardware/processor/instructionprocessor-config.h>
#include <sstmac/hardware/memory/simple_memorymodel.h>
#include <sstmac/hardware/node/simplenode.h>
#include <sstmac/hardware/processor/migratingcomputescheduler.h>
#include <sstmac/hardware/nic/networkinterface.h>

#include <sstmac/software/libraries/filesystem/disk_message.h>
#include <sstmac/software/libraries/filesystem/nppfs_server.h>

#include <sstmac/common/errors.h>
#include <sstmac/software/services/launch/machine_descriptor.h>
#include <iostream>
#include <sstream>

namespace sstmac
{
  namespace hw
  {

    simplenode::simplenode(const sim_parameters::ptr &params) :
      node(params)
    {
      os_->setAMI(sstmac::sw::AMI::AMI_COMM, this);
      os_->setAMI(sstmac::sw::AMI::AMI_COMPUTE, this);
      os_->setAMI(sstmac::sw::AMI::AMI_QUERY, this);

      cores_ = params->get_int_param("node_cores");
      memory_latency_ = params->get_time_param("node_mem_latency");
      bool errorflag = false;
      memory_bandwidth_ = get_bandwidth(
          params->get_param("node_mem_bandwidth").c_str(), errorflag);
      freq_ = params->get_double_param("node_frequency");

      timestamp pre(-1);
      if (params->has_param("node_preemption"))
        {
          pre = params->get_time_param("node_preemption");
        }
      preempt_ = pre;

      SSTMAC_DEBUG << "constructed a simplenode with parameters mem_lat = "
            << memory_latency_ << ", memory_bandwidth = " << memory_bandwidth_
            << ", freq = " << freq_ << ", cores = " << cores_ << "\n";
    }

    void
    simplenode::initialize(nodeaddress::ptr my_id, eventmanager::ptr man)
    {

      node::initialize(my_id, man);

      SSTMAC_DEBUG << "starting simplenode initialization \n";

      double tcycle = 1.0 / (freq_);

      timestamp tflop(tcycle); // 10 clock cycles for a floating point
      //timestamp tmemop(memory_latency_.sec() + 1.0 / (memory_bandwidth_));
      timestamp tmemop = tflop;

      sstmac::hw::simpleprocessor_config::ptr cfg =
          sstmac::hw::simpleprocessor_config::construct(cores_, freq_, 10, 30);

      compute_sched_ = sstmac::hw::migratingcomputescheduler::construct(
          my_id->unique_id(), cores_, preempt_);
      compute_sched_->set_eventmanager(man);
      proc_model_ = sstmac::hw::simpleprocessor::construct(cfg, msg_handler_,
          compute_sched_);

      sstmac::hw::instructionprocessor_config::ptr pcfg =
          sstmac::hw::instructionprocessor_config::construct(cores_,
              INST_FLOP_LABEL, tflop, INST_MEMOP_LABEL, tmemop,
              INST_MEMOP_LOCAL_LABEL, tmemop, INST_MEMOP_RAND_LABEL, tmemop);

      proc_inst_ = sstmac::hw::instructionprocessor::construct(pcfg,
          msg_handler_, compute_sched_);

      mem_model_ = sstmac::hw::simple_memorymodel::construct(memory_bandwidth_,
          memory_latency_, msg_handler_, my_id->unique_id());
      mem_model_->set_eventmanager(man);

      if (params_ && params_->has_param("disk_type"))
        {
          if (params_->get_param("disk_type") == "vesper")
            {
              disk_ = sstmac::hw::disk_controller::construct(
                  my_id->unique_id(), params_, msg_handler_, man);
            }
        }

      if (params_ && params_->has_param("node_nppfs_server"))
        {
          long nid = params_->get_long_param("node_nppfs_server");
          nodeaddress::ptr addr = nodeaddress::convert_from_uniqueid(nid);
          nppfs_server_ = addr;

          std::string fname = params_->get_param("node_nppfs_diskfile");
          sstmac::sw::nppfs_server::ptr serv =
              sstmac::sw::nppfs_server::construct(fname, addr->equals(nodeid_));
          os_->register_ser(serv);
          serv->start();

        }

      SSTMAC_DEBUG << "simplenode initialization done, starting things \n";
      os_->start_apps();

      launcher_->start();
    }

    node::ptr
    simplenode::clone() const
    {
      node::ptr ret = construct(params_);

      return ret;
    }

    // --- sstmac::sw::AMI pure virtuals ---- //
    void
    simplenode::execute_kernel(int func, sst_message::ptr data)
    {
      bool failed = false;

      switch (func)
        {
      case sstmac::sw::AMI::AMI_COMM_SEND:
        {

          if (data->get_type() == sst_message::FILEMETA)
            {
              data->toaddr_ = nppfs_server_;

              if (!nppfs_server_)
                throw ssterror(
                    "simplenode: I don't have an address for the nppfs server");
            }

          nodeaddress::ptr nid = boost::dynamic_pointer_cast<nodeaddress>(
              data->toaddr_);
          data->fromaddr_ = nodeid_;

          SSTMAC_DEBUG << "Node(" << nodeid_->to_string() << "): sending to "
                << data->toaddr_->to_string() << "\n";

          if (data->toaddr_->equals(nodeid_))
            {
              schedule(now(), msg_handler_, data);
              // msg_handler_->handle(data);
            }
          else
            nic_->send(nid, data, msg_handler_);

          failed = nic_->has_failed() || mem_model_->has_failed();
          break;
        }

      case sstmac::sw::AMI::AMI_COMP_TIME:
        {

          proc_model_->compute(data, os_->get_current_threadid());

          failed = compute_sched_->has_failed();

          break;
        }

      case sstmac::sw::AMI::AMI_COMP_INSTR:
        {

          proc_inst_->compute(data, os_->get_current_threadid());

          failed = compute_sched_->has_failed();

          break;
        }

      case sstmac::sw::AMI::AMI_COMP_DISKACCESS:
        {
          sstmac::sw::disk_message::ptr dm = boost::dynamic_pointer_cast<
              sstmac::sw::disk_message>(data);
          disk_->request_access(data);

          failed = disk_->has_failed();
          break;
        }

      case sstmac::sw::AMI::AMI_COMP_MEMACCESS:
        {

          mem_model_->do_access(data);

          mem_model_->has_failed();
          break;
        }

      case sstmac::sw::AMI::AMI_COMP_SLEEP:
        {
          timestamp delay = now() + data->get_time();
          SSTMAC_DEBUG << "node (" << nodeid_->to_string()
                << "): sleeping until " << delay << "\n";
          schedule(delay, msg_handler_, data);
          break;
        }

      default:
        std::cout << "Kernel number: " << func << std::endl;
        throw ssterror("simplenode: cannot process kernel");

        }

    }

    bool
    simplenode::has_failed()
    {
      return nic_->has_failed() || mem_model_->has_failed()
          || compute_sched_->has_failed();
    }
    void
    simplenode::reset_failed()
    {
      nic_->reset_failed();
      mem_model_->reset_failed();
      compute_sched_->reset_failed();
    }

    ptr_type::ptr
    simplenode::execute_kernel(int func)
    {
      switch (func)
        {
      case sstmac::sw::AMI::AMI_QUERY_TOPOLOGY:
        {
          sstmac::ptr_type::ptr top = nic_->get_topology();
          return top;
          break;
        }
      case sstmac::sw::AMI::AMI_QUERY_ADDRESS:
        {
          return nodeid_;
          break;
        }

      case sstmac::sw::AMI::AMI_QUERY_MACHINE:
        {
          return sstmac::sw::machine_descriptor::construct(
              nic_->get_topology()->n(), nic_->get_topology()->name());
          break;
        }

        std::cout << "Kernel number: " << func << std::endl;
        throw ssterror("simplenode: cannot process kernel");
        }
    }

    timestamp
    simplenode::now()
    {
      return eventmodule::now();
    }

    bool
    simplenode::kernel_supported(int func)
    {
      switch (func)
        {
      case sstmac::sw::AMI::AMI_COMM_SEND:
        return true;

      case sstmac::sw::AMI::AMI_COMM_NIC_SEND_ACK:
        return true;

      case sstmac::sw::AMI::AMI_COMP_TIME:
        return true;

      case sstmac::sw::AMI::AMI_COMP_SLEEP:
        return true;

      case sstmac::sw::AMI::AMI_COMP_INSTR:
        return true;

      case sstmac::sw::AMI::AMI_COMP_MEMACCESS:
        return true;

      case sstmac::sw::AMI::AMI_QUERY_TOPOLOGY:
        return true;

      case sstmac::sw::AMI::AMI_QUERY_ADDRESS:
        return true;

      case sstmac::sw::AMI::AMI_COMP_DISKACCESS:
        return disk_;

      default:
        return false;

        }
    }

    std::list<failable::ptr>
    simplenode::get_failure_components()
    {
      std::list<failable::ptr> ret;
      if (compute_sched_)
        ret.push_back(compute_sched_);
      if (mem_model_)
        ret.push_back(mem_model_);
      if (disk_)
        ret.push_back(disk_);
      if (nic_)
        ret.push_back(nic_);
      return ret;
    }

  /*void
   simplenode::init_failure_model(double l_proc, double l_mem, double l_disk,
   double l_nic, failable::ptr ps, double p_ps_node)
   {

   SSTMAC_DEBUG << "node::init_failure_model() -- lproc = " << l_proc
   << ", lmem = " << l_mem << ", ldisk = " << l_disk << ", lnic = "
   << l_nic << "\n";

   ps->add_failable_child(compute_sched_, p_ps_node);
   compute_sched_->init_failable(l_proc, compute_sched_);

   ps->add_failable_child(mem_model_, p_ps_node);
   mem_model_->init_failable(l_mem, mem_model_);

   if (disk_)
   {
   ps->add_failable_child(disk_, p_ps_node);
   disk_->init_failable(l_disk, disk_);
   }

   ps->add_failable_child(nic_, p_ps_node);
   nic_->init_failable(l_nic, nic_);
   }*/

  }
} // end of namespace sstmac.
