/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/processor/migratingcomputescheduler.h>

namespace sstmac
{
  namespace hw
  {

    using namespace std;


    migratingcomputescheduler::migratingcomputescheduler(long nodeid, int ncores, timestamp preempt) :
        computescheduler(nodeid, ncores, preempt)
    {


      for (int i = 0; i < n_cores_; i++)
        {
          compute_slots_[i] = false;
        }
    }



    int
    migratingcomputescheduler::find_open_core(long threadid)
    {
      int core = -1;
      for (int i = 0; i < n_cores_; i++)
        {
          SSTMAC_DEBUG << "migratingcomputescheduler::request_compute checking for core "
                << i << "\n";
          if (compute_slots_.find(i) == compute_slots_.end())
            {
              std::cout << "migratingcomputescheduler:: core: " << i << ", num cores="
                  << n_cores_ << std::endl;
              throw ssterror(
                  "migratingcomputescheduler::request_compute - can't find core in compute_slots");
            }
          if (!compute_slots_[i])
            {

              return i;
            }
        }

      return -1;
    }

    void
    migratingcomputescheduler::schedule_computation(const timestamp& comp,
        const sst_message::ptr& to_deliver, eventhandler::ptr &handler, long thread_id)
    {

      use_me();

      int core = find_open_core(thread_id);

      if (core < 0) //put in suspended queue
        {
          SSTMAC_DEBUG << "[t=" << now()
                << "] -- migratingcomputescheduler suspending compute request" << "\n";
          suspendedQ.push(
              compute_job::construct(to_deliver, comp, comp, handler, thread_id));

        }
      else
        {
          SSTMAC_DEBUG << "[t=" << now()
                << "] -- migratingcomputescheduler: computing for time " << comp << "\n";

          timestamp actual = comp;
          if (preemption_.sec() > 0)
            {
              actual = timestamp(min(comp.sec(), preemption_.sec()));
            }
          compute_slots_[core] = true;
          slots_compute_[to_deliver] = compute_job::construct(to_deliver,
              actual, comp, handler, thread_id, core);
          timestamp time = now() + actual;
          eventmodule::schedule(time, ptr(this), to_deliver);

          // opsys->block(data->get_key());
        }
    }

    void
    migratingcomputescheduler::schedule_computation(compute_job::ptr jorb)
    {
      use_me();
      int core = find_open_core(jorb->threadid_);

      if (core < 0) //put in suspended queue
        {
          SSTMAC_DEBUG << "migratingcomputescheduler suspending compute request" << "\n";
          suspendedQ.push(jorb);

        }
      else
        {
          SSTMAC_DEBUG << "migratingcomputescheduler: computing for time "
                << jorb->timeleft_ << "\n";

          timestamp actual = jorb->timeleft_;
          if (preemption_.sec() > 0)
            {
              actual = timestamp(min(jorb->timeleft_.sec(), preemption_.sec()));
            }
          jorb->time_ = actual;

          compute_slots_[core] = true;
          slots_compute_[jorb->msg_] = jorb;
          timestamp time = now() + actual;
          eventmodule::schedule(time, ptr(this), jorb->msg_);

          // opsys->block(data->get_key());
        }
    }

    void
    migratingcomputescheduler::handle(const sst_message::ptr &msg)
    {
      SSTMAC_DEBUG << "migratingcomputescheduler::handle(" << ")" << "\n";

      if (msg->get_type() == sst_message::FAILURE)
        {
          failure_occurred(msg, self_);
          return;
        }

      compute_job::ptr jorb = slots_compute_[msg];
      slots_compute_.erase(msg);
      compute_slots_[jorb->core_] = false;

      if (!suspendedQ.empty())
        {
          SSTMAC_DEBUG << "migratingcomputescheduler("
                << "): executing suspended kernel " << "\n";

          schedule_computation(suspendedQ.front()->time_,
              suspendedQ.front()->msg_, suspendedQ.front()->handle_, suspendedQ.front()->threadid_);
          suspendedQ.pop();
        }

      if (jorb->time_ >= jorb->timeleft_)
        {

          SSTMAC_DEBUG << "migratingcomputescheduler("
                << "): calling handle on a finished job " << "\n";
          jorb->handle_->handle(jorb->msg_);
        }
      else
        {

          SSTMAC_DEBUG
                <<
                "migratingcomputescheduler - computation was preempted, scheduling again \n";

          jorb->timeleft_ = timestamp(
              jorb->timeleft_.sec() - jorb->time_.sec());
          schedule_computation(jorb);
        }
    }

  }
} //end of namespace sstmac
