/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009-2010 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/main/sstmac.h>

#include <iostream>
#include <iterator>
#include <string>
#include <map>
#include <list>
#include <getopt.h>
#include <stdint.h>
#include <ctype.h>
#include <errno.h>
#include <math.h>
#include <boost/lexical_cast.hpp>

#include <sstmac/backends/native/nodeid.h>
#include <sstmac/software/process/appdescr.h>

#include <sstmac/common/fileio.h>

using namespace std;
using namespace sstmac;
using namespace sstmac::sw;
using namespace sstmac::hw;
using namespace sstmac::native;

using boost::lexical_cast;

static std::ostream&
operator<<(std::ostream &os, const opts &oo)
{
  if (oo.help)
    os << "  --help \\\n";
  if (oo.color)
    os << "  --color \\\n";
  os << "  --debug=\"" << oo.debug << "\" \\\n";
  os << "  --configfile=\"" << oo.configfile << "\" \\\n";
  os << "  --mode=\"" << oo.mode << "\" \\\n";
  os << "  --stoptime=" << oo.stoptime.sec();
  return os;
}

#define PARSE_OPT_SUCCESS 0
#define PARSE_OPT_EXIT_SUCCESS 1
#define PARSE_OPT_EXIT_FAIL 2

static int
try_main(int argc, char **argv)
{
  opts oo;
  int parse_status = parse_opts(argc, argv, oo);
  if (parse_status == PARSE_OPT_SUCCESS)
    {
      if (oo.debug != "")
        logger::set_user_param(oo.debug);

      if (oo.dumpimetafile != "")
        {
          std::cout << "Running parsedumpi on metafile " << oo.dumpimetafile
              << std::endl;
          if (oo.network == "")
            {
              std::cerr
                  << "ERROR: parsedumpi needs a network (e.g. hopper, franklin) specified with the -n option"
                  << std::endl;
              return 1;
            }
          oo.configfile = "parsedumpi.ini";
        }

      if (oo.color)
        logger::set_debug_coloring(true);

      if (oo.configfile == "")
        {
          std::cerr << "ERROR: need a configuration file" << std::endl;
          return 1;
        }

      //set up the search path
      SSTFileIO::init();

      //use the config file to set up file search paths
      size_t pos = oo.configfile.find_last_of('/');
      if (pos != string::npos)
        {
          std::string dir = oo.configfile.substr(0, pos + 1);
          SSTFileIO::add_path(dir);
        }

      file_parameters::ptr params = file_parameters::construct(oo.configfile);
      if (oo.network != "")
        {
          std::string machinefile = oo.network + ".ini";
          params->append_params(machinefile);
        }

      if (oo.params)
        { // there were command-line overrides
          oo.params->combine_into(params);
        }

      if (oo.dumpimetafile != "")
        {
          params->add_param_override("launch_dumpi_metaname", oo.dumpimetafile);
          //count the number of applications to launch
          std::ifstream in;
          in.open(oo.dumpimetafile.c_str());

          if (!in.is_open()) //see if the file is included in the include path
            {
              std::cerr << "ERROR: " << oo.dumpimetafile
                  << " is not a valid dumpi meta file" << std::endl;
              return 1;
            }

          std::string line;
          while (in.good())
            {
              std::getline(in, line);
              if (line.find("=") != std::string::npos)
                { //an assignment
                  pair<string, string> keyval = sim_parameters::parse_param(
                      line);
                  if (keyval.first == "numprocs")
                    params->add_param_override("launch_app1_size",
                        keyval.second);
                }
            }
        }

      // make a dummy to make sure the converter gets built
      //  (see nodeid.h)
      nodeaddress::ptr dummy = nodeid::construct(-1);

      long timescale = params->get_optional_long_param("timestamp_resolution",
          1);
      timestamp::init_stamps(timescale);

      if (params->has_param("stat_time_interval"))
        {
          stat_time_window::set_time_interval(
              params->get_time_param("stat_time_interval"));
        }

      /** This must come after all params are processed */
      SSTAppDesc_build::init_apps(params);

      if (oo.mode == "normal" || oo.mode == "notime")
        {
          try
            {
              run(params, oo.stoptime, (oo.mode != "notime"));
            }
          catch (const std::exception &e)
            {
              std::cout << argv[0]
                  << ": caught exception while running simulation:\n"
                  << e.what() << "\n";
              return 1;
            }
          catch (...)
            {
              std::cout << argv[0]
                  << ": caught unknown exception while running simulation\n";
              return 1;
            }
        }
      else if (oo.mode == "debug")
        {
          run_debugger(params, oo);
        }
    }
  else if (parse_status == PARSE_OPT_EXIT_SUCCESS)
    {
      return 0;
    }
  else if (parse_status == PARSE_OPT_EXIT_FAIL)
    {
      return 1;
    }

  return 0;
}

int
main(int argc, char **argv)
{
  int rc;

  try
    {
      rc = try_main(argc, argv);
    }
  catch (const std::exception &e)
    {
      std::cout << argv[0]
          << ": caught exception while setting up simulation:\n" << e.what()
          << "\n";
      return 1;
    }
  catch (...)
    {
      std::cout << argv[0]
          << ": caught unknown exception while setting up simulation\n";
      return 1;
    }

  return rc;
}

int
parse_opts(int argc, char **argv, opts &oo)
{
  int dompitest = 0;
  option gopt[] =
    {
      { "help", no_argument, NULL, 'h' },
      { "debug", required_argument, NULL, 'd' },
      { "stoptime", required_argument, NULL, 't' },
      { "configfile", required_argument, NULL, 'f' },
      { "param", required_argument, NULL, 'p' },
      { "mode", required_argument, NULL, 'm' },
      { "color", no_argument, NULL, 'c' },
      { "parse", required_argument, NULL, 'u' }, //u = undumpi
          { "network", required_argument, NULL, 'n' },
          { "mpitest", no_argument, &dompitest, 1 },
          { NULL, 0, NULL, '\0' } };
  int ch;
  bool errorflag = false;
  std::list<pair<string, string> > paramlist;
  while ((ch = getopt_long(argc, argv, "hcd:f:t:p:m:n:u", gopt, NULL)) != -1)
    {
      switch (ch)
        {
      case 0:
        //this set an input flag
        break;
      case 'h':
        oo.help = 1;
        break;
      case 'd':
        oo.debug = optarg;
        break;
      case 'f':
        oo.configfile = optarg;
        break;
      case 'u':
        oo.dumpimetafile = optarg;
        break;
      case 't':
        oo.stoptime = timestamp(atof(optarg));
        break;
      case 'n':
        oo.network = optarg;
        break;
      case 'p':
        {
          if (!oo.params)
            {
              oo.params = command_parameters::construct();
            }
          pair<string, string> keyval = sim_parameters::parse_param(optarg);
          paramlist.push_back(keyval);
          oo.params->add_param(keyval.first, keyval.second);
          break;
        }
      case 'm':
        oo.mode = optarg;
        break;

      case 'c':
        oo.color = true;
        break;

      default:
        std::cerr << "Unhandled input flag.\n";
        errorflag = true;
        std::cout << "optarg: " << optarg << std::endl;
        break;
        }
      if (oo.help)
        {
          print_help(argc, argv);
          return PARSE_OPT_EXIT_SUCCESS;
        }
      if (errorflag)
        {
          std::cerr << "Error return\n";
          return PARSE_OPT_EXIT_FAIL;
        }
    }

  /** check to see if we should do an mpitestall */
  if (dompitest)
    {
      if (oo.configfile != "")
        {
          std::cerr
              << "Cannot set config file with --configfile and also request --mpitest"
              << std::endl;
          return PARSE_OPT_EXIT_FAIL;
        }
      oo.configfile = "mpitestall.ini";

      if (oo.network == "")
        {
          std::cerr
              << "For --mpitest, you must specify a network and node model with --network"
              << std::endl;
          return PARSE_OPT_EXIT_FAIL;
        }
    }

  /** Set sensible defaults for some of the paramters if not set*/
  if (oo.configfile == "")
    oo.configfile = "parameters.ini";

  std::cerr << argv[0] << " \\\n" << oo;
  if (paramlist.size())
    {
      for (std::list<pair<string, string> >::iterator i = paramlist.begin();
          i != paramlist.end(); i++)
        {
          std::cerr << " \\\n  --param \"" << i->first << "=" << i->second
              << "\"";
        }
    }
  std::cerr << "\n";

  return PARSE_OPT_SUCCESS;
}

void
run_debugger(sim_parameters::ptr params, opts oo)
{
  std::cout << "------------------------------------------ \n";
  std::cout << "--- Welcome to the SST/macro debugger ---- \n";
  std::cout << "------------------------------------------- \n";
  std::cout << "Type \"help\" for more information. \n";

  timestamp stop = oo.stoptime;
  std::string dbg_str = "timestamps | " + oo.debug;
  logger::set_user_param(dbg_str);
  bool exit = false;
  bool hasrunyet = false;
  eventmanager::ptr eventman;

  while (!exit)
    {
      std::cout << logger::red << "sstmac-debug >> " << logger::black;
      std::string command;
      std::cin >> command;

      if (command == "help" || command == "h")
        {
          std::cout << "Commands are: \n";
          std::cout << "  help (h): print this help list \n";
          std::cout << "  start (st): start the simulation \n";
          std::cout
              << "  stopat (sa) <value>: set a breakpoint at the given time \n";
          std::cout << "  exit: exit the debugger \n";
          std::cout << "  silly: print something silly \n";
          std::cout << "  debug (d) <value>: add <value> to the debug list \n";
          std::cout
              << "  debug-list (dl): see the list of things being debugged \n";
          std::cout << "  debug-clear (cl): clear the debug list \n";
          std::cout << "  now: print the current time of the simulation \n";
          std::cout << "  step|next (n): execute one event \n";
          std::cout
              << "  reset: reset the simulation so it can be started again\n";
        }
      else if (command == "start" || command == "st")
        {
          if (!hasrunyet)
            eventman = run(params, stop, false);
          else
            eventman->run();
          hasrunyet = true;
        }
      else if (command == "stopat" || command == "sa")
        {
          double time;
          std::cin >> time;
          stop = timestamp(time);
          std::cout << logger::green << "debug: should stop at " << stop
              << "\n";
          if (hasrunyet)
            {
              eventman->schedule(stop, manager::stophandler::construct(),
                  sst_message::ptr());
            }
        }
      else if (command == "exit")
        {
          exit = true;
        }
      else if (command == "silly")
        {
          std::cout << "firetruck lightbulb purple-monkey-dishwasher \n";
        }
      else if (command == "debug" || command == "d")
        {
          std::string val;
          std::cin >> val;
          if (dbg_str == "")
            {
              dbg_str = "<debug> " + val;
            }
          else
            {
              dbg_str = dbg_str + " | <debug> " + val;
            }

          logger::set_user_param(dbg_str);
        }
      else if (command == "debug-list" || command == "dl")
        {
          std::cout << "debug list: " << dbg_str << "\n";
        }
      else if (command == "debug-clear" || command == "cl")
        {
          dbg_str = "timestamps";
          logger::set_user_param(dbg_str);
        }
      else if (command == "next" || command == "step" || command == "n")
        {
          if (!hasrunyet)
            {
              std::cout << "haven't started yet, have to start first \n";
            }
          else
            {
              eventman->step();
            }
        }
      else if (command == "now")
        {
          if (!hasrunyet)
            std::cout << "0 \n";
          else
            std::cout << eventman->now() << "\n";
        }
      else if (command == "reset")
        {

          stop = oo.stoptime;
          dbg_str = "timestamps | " + oo.debug;
          hasrunyet = false;
          logger::set_user_param(dbg_str);

        }
      else
        {

          std::cout << "unknown command: " << command << "\n";
        }
    }

  std::cout << "---------------------------------------------------------- \n";
  std::cout << "--- Exiting the SST/macro debugger normally, so long! ---- \n";
  std::cout << "---------------------------------------------------------- \n";

}

void
print_help(int, char **argv)
{
  char *name = argv[0];
  char *tail = name + strlen(name);
  while (tail != name && *tail != '/')
    --tail;
  if (tail != name)
    ++tail;

  std::cerr
      << "usage: "
      << tail
      << " -f <config file> [options] \n"
      << " \n"
      << "Options are: \n"
      << "\t[(--help|-h)]                          \n"
      << "\t[(--color|-c)]                         \n"
      << "\t[(--debug|-d)             <value> ]    \n"
      << "\t[(--configfile|-f)        <value> ]    \n"
      << "\t[(--param|-p)       <key>=<value> ]    \n"
      << "\t[(--stoptime|-t)          <value> ]    \n"
      << "\t[(--mode|-m)              <value> ]    \n"
      << "\n"
      << "Configuration file is not optional. See parameters.ini for \n"
      << "a list of parameters. \n"
      << "\n"
      << "Turn ANSI console output coloring on with --color or -c, which is \n"
      << "particularly useful when debugging, but should not be used for make check \n"
      << "or redirecting to a file. \n" << "\n"
      << "--param (-p) is used to override any parameters in the \n"
      << "configuration file, for example: \n" << "\t-p network_name=buffer \n"
      << "\n" << "--stoptime (-t) stops the simulation at the given time, \n"
      << "which is useful for debugging \n" << "\n"
      << "--mode (-m) can be used to open the debugger by saying -m debug, \n"
      << "and can turn off printing the simulation at the end with -m notime \n"
      << "\n" << "Valid arguments to --debug (-d) are strings of the form \n"
      << "\"<(debug|stats)> (name1) | (name2) | ... \" \n" << "\t- examples: \n"
      << "\t\t- \"<debug> mpi\" turns on all mpi output \n"
      << "\t\t- \"<debug> app | mpicollective\" turns on all \n"
      << "\t\tskeleton and mpicollective output \n"
      << "\t\t- \"<stats> all\" turns on all stat collection \n"
      << "\t- some current useful debug tags are: \n"
      << "\t\tall sim app mpi mpiapi mpicollective mpiqueue os processor \n"
      << "\t\tnode network manager eventmanager allocation indexing \n"
      << "\t\tlaunch lib_compute lib_sleep flower \n"
      << "\t- current stat tags are: \n" << "\t\tall processor network \n";

}

