/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/software/libraries/filesystem/nppfs_server.h>
#include <sstmac/software/libraries/filesystem/file_metadata.h>
#include <sstmac/common/basicstringtokenizer.h>

#include <math.h>
#include <fstream>

namespace sstmac
{
  namespace sw
  {

    using namespace pst::BasicStringTokenizer;

    debug nppfs_server::dbg_;

    void
    nppfs_server::start()
    {
      service::start();

      lib_mem_ = lib_compute_memmove::construct("nppfs_server_lib_memmove");
      register_lib(lib_mem_);

      ptr self(this);
      waiter::ptr wait = waiter::construct();
      lib_mem_->init_handler(self);
      lib_mem_->init_waiter(wait);
    }

    service::ptr
    nppfs_server::clone()
    {
      return construct(fname_, is_meta_);
    }

    void
    nppfs_server::handle(const sst_message::ptr &msg)
    {
      /* if (msg->get_type() == sst_message::COMPUTE)
       {
       mem_access::ptr mem = boost::dynamic_pointer_cast<mem_access>(msg);
       disk_message::ptr dm = computing_[mem->uid_];
       os_->execute_kernel(AMI::AMI_COMPUTE, AMI::AMI_COMP_DISKACCESS, dm);
       }
       else */if (msg->get_type() == sst_message::FILEMETA || msg->get_type()
          == sst_message::FILEDATA)
        {
          file_metadata::ptr meta = boost::dynamic_pointer_cast<file_metadata>(
              msg);

          if (meta->op_ == file_metadata::OPEN)
            {

              SSTMAC_DEBUG << "nppfs_server["
                    << os_->get_my_addr()->to_string()
                    << "]::handle - servicing OPEN request for file "
                    << meta->fname_ << " from node "
                    << msg->fromaddr_->to_string() << "\n";
              if (files_.find(meta->fname_) == files_.end())
                {
                  throw valueerror(
                      "nppfs_server::handle - OPEN request, can't find filename "
                          + meta->fname_);
                }
              meta->set_type(sst_message::FILEDATA); // this is necessary so the node doesn't redirect it
              meta->data_ = files_[meta->fname_];
              meta->op_ = file_metadata::OPEN_RESP;
              meta->bytes_ = 8 + 8 * files_[meta->fname_]->nodes_.size();
              meta->toaddr_ = meta->fromaddr_;
              meta->set_lib_name(meta->from_lib_);
              meta->set_needs_ack(false);

              if (!lib_mem_)
                {
                  throw valueerror(
                      "nppfs_server - memory library has not been initialized, probably need to call start on the service");
                }

              // lib_mem_->compute_block_access(meta->bytes_);

              os_->execute_kernel(AMI::AMI_COMM, AMI::AMI_COMM_SEND, meta);

            }
          else if (meta->op_ == file_metadata::CREATE)
            {
              SSTMAC_DEBUG << "nppfs_server["
                    << os_->get_my_addr()->to_string()
                    << "]::handle - servicing CREATE request for file "
                    << meta->fname_ << " from node "
                    << msg->fromaddr_->to_string() << "\n";

              file_descriptor::ptr newfile = file_descriptor::construct(
                  meta->fname_);

              newfile->nodes_.push_back(meta->fromaddr_);
              newfile->total_size_ = 0;
              newfile->sizes_.push_back(0);
              newfile->sectors_.push_back(rand() % 10);

              meta->set_type(sst_message::FILEDATA); // this is necessary so the node doesn't redirect it
              meta->data_ = newfile;
              meta->op_ = file_metadata::CREATE_RESP;
              meta->bytes_ = 8 + 8 * newfile->nodes_.size();
              meta->toaddr_ = meta->fromaddr_;
              meta->set_lib_name(meta->from_lib_);
              meta->set_needs_ack(false);

              if (!lib_mem_)
                {
                  throw valueerror(
                      "nppfs_server - memory library has not been initialized, probably need to call start on the service");
                }

              // lib_mem_->compute_block_access(meta->bytes_);

              os_->execute_kernel(AMI::AMI_COMM, AMI::AMI_COMM_SEND, meta);
            }
          else if (meta->op_ == file_metadata::READ || meta->op_
              == file_metadata::WRITE)
            {
              SSTMAC_DEBUG << "nppfs_server["
                    << os_->get_my_addr()->to_string()
                    << "]::handle - servicing READ/WRITE request for file "
                    << meta->fname_ << " from node "
                    << msg->fromaddr_->to_string() << "\n";
              openfile::ptr of = boost::dynamic_pointer_cast<openfile>(
                  meta->data_);
              int sec = of->get_sector() + of->get_offset_sectors();
              int numsec = ceil(
                  (double) meta->bytes_ / (sstmac::hw::Disk::blockSize
                      * 1000000.0));

              // mem_access::ptr mem = mem_access::construct(meta->access_size_,
              //     get_lib_name());

              disk_message::DISK_ACCESS_TYPE t = (meta->op_
                  == file_metadata::READ) ? disk_message::READ
                  : disk_message::WRITE;
              disk_message::ptr dm = disk_message::construct(get_lib_name(), t,
                  meta->bytes_, sec, numsec, ACCESS_ID);

              //mem->uid_ = compute_counter_;
              // computing_[compute_counter_] = dm;
              // compute_counter_++;

              os_->execute_kernel(AMI::AMI_COMPUTE, AMI::AMI_COMP_DISKACCESS,
                  dm);

              //os_->execute_kernel(AMI::AMI_COMPUTE, AMI::AMI_COMP_MEMACCESS,
              //    mem);

              pending_[ACCESS_ID] = of;
              sources_[ACCESS_ID] = meta->fromaddr_;
              fromlibs_[ACCESS_ID] = meta->from_lib_;
              needs_resp_[ACCESS_ID] = meta->op_ == file_metadata::READ
                  || meta->write_resp_;
              ACCESS_ID++;
            }
        }
      else if (msg->get_type() == sst_message::DISK)
        { // a response from the disk model
          SSTMAC_DEBUG << "nppfs_server[" << os_->get_my_addr()->to_string()
                << "]::handle - servicing a DISK complete message \n";
          disk_message::ptr cast = boost::dynamic_pointer_cast<disk_message>(
              msg);
          long uid = cast->get_uid();
          SSTMAC_DEBUG << "nppfs_server[" << os_->get_my_addr()->to_string()
                << "]::handle - disk message " << uid << " came back \n";

          if (needs_resp_[uid])
            {
              SSTMAC_DEBUG << "nppfs_server["
                    << os_->get_my_addr()->to_string()
                    << "]::handle - sending a response back to "
                    << sources_[uid]->to_string() << " \n";
              openfile::ptr of = pending_[uid];
              file_metadata::ptr
                  meta =
                      file_metadata::construct(
                          "nppfs",
                          of->name(),
                          ((cast->get_access_type() == disk_message::READ) ? file_metadata::READ_RESP
                              : file_metadata::WRITE_RESP));
              meta->toaddr_ = sources_[uid];
              meta->bytes_ = cast->get_byte_length();
              meta->set_lib_name(fromlibs_[uid]);
              meta->set_needs_ack(false);

              os_->execute_kernel(AMI::AMI_COMM, AMI::AMI_COMM_SEND, meta);
            }

          pending_.erase(uid);
          sources_.erase(uid);
          fromlibs_.erase(uid);
          needs_resp_.erase(uid);
        }
    }

    void
    nppfs_server::read_disk_file()
    {
      std::ifstream in(fname_.c_str());
      if (in.is_open())
        {

          file_descriptor::ptr current;
          std::string line;

          while (in.good())
            {
              std::getline(in, line);

              if (line.length() > 1)
                {
                  std::deque<std::string> tok;

                  std::string space = " ";
                  tokenize(line, tok, space);

                  if (tok[0] == "file")
                    {
                      SSTMAC_DEBUG << "nppfs_server: adding file named "
                            << tok[1] << "\n";
                      current = file_descriptor::construct(tok[1]);
                      files_[tok[1]] = current;

                    }
                  else if (tok[0] == "nodes")
                    {
                      for (int i = 1; i < tok.size(); i++)
                        {
                          long uid = atol(tok[i].c_str());
                          nodeaddress::ptr addr =
                              nodeaddress::convert_from_uniqueid(uid);
                          current->nodes_.push_back(addr);
                        }
                    }
                  else if (tok[0] == "sizes")
                    {
                      for (int i = 1; i < tok.size(); i++)
                        {
                          long size = atol(tok[i].c_str());
                          current->sizes_.push_back(size);
                          current->total_size_ += size;
                        }
                    }
                  else if (tok[0] == "sectors")
                    {
                      for (int i = 1; i < tok.size(); i++)
                        {
                          long sec = atol(tok[i].c_str());
                          current->sectors_.push_back(sec);
                        }
                    }
                }
            }

        }
    }

  }
}/* namespace sstmac */
