/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_LIBRARIES_HPX_HPX_APP_H_INCLUDED
#define SSTMAC_SOFTWARE_LIBRARIES_HPX_HPX_APP_H_INCLUDED

#include <sstmac/software/process/app.h>
#include <hpx/runtime/actions/hpx_action.h>
#include <hpx/runtime/hpx_thread_manager.h>
#include <hpx/agas/agas_server.h>
#include <sstmac/software/ami/environment.h>

namespace hpx
{

  /**
   * A base type for applications that can run via mpi.
   * Intended to be safe for multiple inheritance (for applications
   * that use MPI and some other form of communication).
   */
  class hpx_app : public sstmac::sw::app
  {

  protected:

    hpx_thread_manager::ptr threadman_;
    agas_server::ptr agas_;

    hpx_app(sstmac::sw::softwareid id) :
        app(id)
    {

    }

    virtual void
    hpx_main() = 0;

    template<typename ActName>
      void
      hpx_register_action(std::string actname)
      {
        boost::intrusive_ptr<ActName> actptr = ActName::construct();
        threadman_->register_action<ActName>(actname, actptr);
      }

    void
    init()
    {

      //start the agas server if not already there
      sstmac::sw::library::ptr agas = os_->get_lib(agas_server::lib_name());
      if (agas == sstmac::sw::library::ptr()) // only do one mpiserver per app per node
        {
          agas = agas_server::construct();
          os_->register_lib(agas);
          agas_server::ptr casted = boost::dynamic_pointer_cast<agas_server>(
              agas);
          casted->start();

        }
      agas_ = boost::dynamic_pointer_cast<agas_server>(agas);

      //start the thread manager if not already there
      sstmac::sw::library::ptr tman = os_->get_lib(
          hpx_thread_manager::lib_name());
      if (tman == sstmac::sw::library::ptr()) // only do one mpiserver per app per node
        {
          tman = hpx_thread_manager::construct(agas_);
          os_->register_lib(tman);
          hpx_thread_manager::ptr casted = boost::dynamic_pointer_cast<
              hpx_thread_manager>(tman);
          casted->start();

        }
      threadman_ = boost::dynamic_pointer_cast<hpx_thread_manager>(tman);
      register_actions();
      register_components();

    }

    virtual void
    register_actions() = 0;

    virtual void
    register_components() = 0;

  public:

    typedef boost::intrusive_ptr<hpx_app> ptr;

    /// Goodbye.
    virtual
    ~hpx_app()
    {
    }

    virtual void
    skeleton_main()
    {
      init();

      if (id_.task_.id == 0)
        {
          globalid::set_num_localities(os_->get_env(id_.app_)->get_app_count());
          hpx_main();
        }
    }

    virtual std::string
    to_string() const
    {
      return "hpx_app";
    }

    //convenience wrapper
    sstmac::nodeaddress::ptr
    find_here()
    {
      return os_->get_my_addr();
    }

    static hpx_thread_manager::ptr
    get_current_thread_manager();

    static agas_server::ptr
    get_current_agas_server();

    static void
    hpx_block(sstmac::sw::key::ptr blocker);

    static void
    hpx_thread_swap();

    static void
    hpx_unblock(sstmac::sw::key::ptr unblock);

    static void
    hpx_send_msg(sstmac::sst_message::ptr msg);

    static sstmac::nodeaddress::ptr
    gbl_find_here();

    static sstmac::sw::environment::const_ptr
    get_env();

    static sstmac::sw::OS_API::ptr
    get_os();

    static void
    compute(const sstmac::timestamp &t);

    static void
    hpx_register_lib(sstmac::sw::library::ptr lib);


  };

} // end of namespace sstmac.

#endif
