/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <hpx/runtime/hpx_thread_manager.h>
#include <hpx/runtime/actions/hpx_action.h>
#include <hpx/components/hpx_component.h>

#include <hpx/parcels/remote_action.h>
#include <hpx/parcels/remote_action_result.h>
#include <hpx/lcos/remote_callback.h>

namespace hpx
{
  long hpx_thread_manager::TRANS_ID_CNT = 0;
  sstmac::debug hpx_thread_manager::dbg_;

  void
  hpx_thread_manager::handle(const sstmac::sst_message::ptr &msg)
  {

    if (msg->get_type() == sstmac::sst_message::SEND_ACK)
      {
        return;
      }
    hpx::parcels::parcel::ptr parc = boost::dynamic_pointer_cast<
        hpx::parcels::parcel>(msg);

    if (!parc)
      {
        throw sstmac::ssterror(
            "hpx_thread_manager::handle(...) - can't cast msg to a parcel");
      }

    switch (parc->type())
      {

    case hpx::parcels::parcel::REMOTE_ACTION:
      {

        hpx::parcels::remote_action::ptr rem = boost::dynamic_pointer_cast<
            hpx::parcels::remote_action>(parc);
        SSTMAC_DEBUG << "hpx_thread_manager -- invoking action " << rem->which()
              << " from remote call \n";
        hpx::lcos::remote_callback::ptr sendback =
            hpx::lcos::remote_callback::construct(rem->trans_id(),
                rem->fromaddr_);
        if (rem->is_gid())
          {
            sstmac::nodeaddress::ptr loc = agas_->resolve_locality(rem->gid());

            if (!is_local(loc))
              {
                throw sstmac::unimplementederror(
                    "remote action arrived but the gid is not here, component migration is not implemented yet");
              }
            this->start_action(rem->which(), rem->params(), sendback,
                rem->gid());
          }
        else
          {
            this->start_action(rem->which(), rem->params(), sendback,
                os_->get_my_addr());
          }
        break;

      }

    case hpx::parcels::parcel::REMOTE_ACTION_RESULT:
      {
        SSTMAC_DEBUG
          << "hpx_thread_manager -- got a result back from remote action \n";
        hpx::parcels::remote_action_result::ptr rem =
            boost::dynamic_pointer_cast<hpx::parcels::remote_action_result>(
                parc);
        if (rem_actions_.find(rem->trans_id()) == rem_actions_.end())
          {
            throw sstmac::ssterror(
                "hpx_thread_manager::handle() -- transaction id not found in saved remote action invocations");
          }
        hpx::lcos::base_lco_result::ptr lco = rem_actions_[rem->trans_id()];
        rem_actions_.erase(rem->trans_id());
        lco->set_value(rem->result(), os_);

        break;
      }

    default:
      std::cout << "parcel number: " << parc->type() << "\n";
      throw sstmac::ssterror(
          "hpx_thread_manager::handle(..) -- unknown parcel type");
      break;
      }

  }

  globalid
  hpx_thread_manager::create_component(std::string name)
  {
    SSTMAC_DEBUG << "hpx_thread_manager -- creating component \n";
    if (comp_factory_.find(name) == comp_factory_.end())
      {
        std::cout << "component name: " << name << "\n";
        throw sstmac::ssterror(
            "hpx_thread_manager::create_component() - component has not been registered yet");
      }

    globalid newid = agas_->new_gid(lva::COMPONENT);
    hpx::components::hpx_component::ptr comp = comp_factory_[name]->clone(
        newid);
    comp->set_curr_locality(os_);
    components_[newid.gid_] = comp;
    return newid;
  }

  void
  hpx_thread_manager::start_action(std::string actname,
      hpx_action_params::ptr params,
      boost::intrusive_ptr<hpx::lcos::base_lco_result> prom,
      sstmac::nodeaddress::ptr locality)
  {
    SSTMAC_DEBUG << "hpx_thread_manager -- starting action " << actname
          << " with locality " << locality->to_string() << " \n";
    if (is_local(locality))
      {
        SSTMAC_DEBUG << "hpx_thread_manager -- action is local, at "
              << locality->to_string() << " \n";
        if (actions_.find(actname) == actions_.end())
          {
            throw sstmac::ssterror(
                "hpx_thread_manager::start_action - can't find action in map,  make sure it was registered");
          }

        boost::intrusive_ptr<actions::hpx_action> actlib =
            boost::dynamic_pointer_cast<actions::hpx_action>(actions_[actname]);

        if (!actlib)
          {
            throw sstmac::ssterror(
                "hpx_thread_manager::start_action() - problem casting found action to hpx_action");
          }

        boost::intrusive_ptr<hpx_action_thread> pxth =
            hpx_action_thread::construct(selfptr_, agas_, actlib, params, prom);

        SSTMAC_DEBUG << "hpx_thread_manager -- starting action thread \n";
        os_->start_thread(pxth);
      }
    else
      {
        SSTMAC_DEBUG << "hpx_thread_manager -- action is remote, at "
              << locality->to_string() << " \n";
        long trans = TRANS_ID_CNT++;
        hpx::parcels::remote_action::ptr msg =
            hpx::parcels::remote_action::construct(actname, params, trans,
                os_->now());
        msg->toaddr_ = locality;
        msg->set_lib_name(libname_);
        rem_actions_[trans] = prom;
        os_->execute_kernel(sstmac::sw::AMI::AMI_COMM,
            sstmac::sw::AMI::AMI_COMM_SEND, msg);
      }
  }

  void
  hpx_thread_manager::start_action(std::string actname,
      hpx_action_params::ptr params,
      boost::intrusive_ptr<hpx::lcos::base_lco_result> prom, globalid gid)
  {
    SSTMAC_DEBUG << "hpx_thread_manager -- starting action " << actname
          << " with globalid " << gid.gid_ << "\n";
    sstmac::nodeaddress::ptr loc = agas_->resolve_locality(gid);

    if (is_local(loc))
      {
        SSTMAC_DEBUG << "hpx_thread_manager -- action is local, at "
              << loc->to_string() << " \n";
        if (actions_.find(actname) == actions_.end())
          {
            throw sstmac::ssterror(
                "hpx_thread_manager::start_action - can't find action in map,  make sure it was registered");
          }

        boost::intrusive_ptr<actions::hpx_action> actlib =
            boost::dynamic_pointer_cast<actions::hpx_action>(actions_[actname]);

        if (!actlib)
          {
            throw sstmac::ssterror(
                "hpx_thread_manager::start_action() - problem casting found action to hpx_action");
          }

        if (components_.find(gid.gid_) == components_.end())
          {
            throw sstmac::ssterror(
                "hpx_thread_manager::start_action() -- component not found here, something wrong with agas");
          }

        hpx::components::hpx_component::ptr comp = components_[gid.gid_];

        if (!comp)
          {
            throw sstmac::ssterror(
                "hpx_thread_manager::start_action() -- component didn't cast right");
          }

        boost::intrusive_ptr<hpx_action_thread> pxth =
            hpx_action_thread::construct(selfptr_, agas_, actlib, params, prom,
                comp);

        SSTMAC_DEBUG << "hpx_thread_manager -- starting action thread \n";
        os_->start_thread(pxth);
      }
    else
      {
        SSTMAC_DEBUG << "hpx_thread_manager -- action is remote, at "
              << loc->to_string() << " \n";
        long trans = TRANS_ID_CNT++;
        hpx::parcels::remote_action::ptr msg =
            hpx::parcels::remote_action::construct(actname, params, trans, gid,
                os_->now());
        msg->toaddr_ = loc;
        msg->set_lib_name(libname_);
        rem_actions_[trans] = prom;
        os_->execute_kernel(sstmac::sw::AMI::AMI_COMM,
            sstmac::sw::AMI::AMI_COMM_SEND, msg);
      }

  }

}
