/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_LIBRARIES_HPX_RUNTIME_HPX_THREAD_MANAGER_H_INCLUDED
#define SSTMAC_SOFTWARE_LIBRARIES_HPX_RUNTIME_HPX_THREAD_MANAGER_H_INCLUDED

#include <sstmac/software/services/service.h>
#include <sstmac/software/libraries/compute/lib_compute_time.h>
#include <hpx/runtime/actions/hpx_action_params.h>
#include <hpx/runtime/hpx_action_thread.h>
#include <sstmac/software/libraries/unblockhandler.h>
#include <sstmac/software/process/blockwaiter.h>
#include <hpx/agas/agas_server.h>
#include <map>
#include <hpx/agas/globalid.h>
#include <boost/unordered_map.hpp>
#include <hpx/runtime/actions/hpx_action.h>

namespace hpx
{



  class hpx::components::hpx_component;

  class hpx_thread_manager : public sstmac::sw::service
  {

  public:
    typedef boost::intrusive_ptr<hpx_thread_manager> ptr;

  protected:

    static sstmac::debug dbg_;

    std::map<std::string, actions::hpx_action::ptr> actions_;

    sstmac::sw::lib_compute_time::ptr libcomp_;
    sstmac::sw::key::ptr compkey_;

    agas_server::ptr agas_;

    ptr selfptr_;

    boost::unordered_map<std::string,
        boost::intrusive_ptr<hpx::components::hpx_component> > comp_factory_;
    boost::unordered_map<long,
        boost::intrusive_ptr<hpx::components::hpx_component> > components_;

    static long TRANS_ID_CNT;
    boost::unordered_map<long, hpx::lcos::base_lco_result::ptr> rem_actions_;

    hpx_thread_manager(agas_server::ptr agas) :
        agas_(agas)
    {
      libname_ = lib_name();

      if (!dbg_.is_real())
        {
          dbg_ = sstmac::debug("<debug> hpx | hpx_thread_manager",
              sstmac::logger::blue);
          sstmac::logger::register_logger(&dbg_);
        }
    }

  public:

    static ptr
    construct(agas_server::ptr agas)
    {
      return ptr(new hpx_thread_manager(agas));
    }

    virtual
    ~hpx_thread_manager()
    {
    }

    static std::string
    lib_name()
    {
      return "hpx_thread_manager";
    }

    virtual void
    start()
    {
      service::start();

      compkey_ = sstmac::sw::key::construct();
      libcomp_ = sstmac::sw::lib_compute_time::construct(
          "hpx_thread_manager-lib-comp");
      libcomp_->init_handler(
          sstmac::sw::unblockhandler::construct(os_, compkey_));
      libcomp_->init_waiter(sstmac::sw::blockwaiter::construct(os_, compkey_));
      register_lib(libcomp_);

      selfptr_ = boost::dynamic_pointer_cast<hpx_thread_manager>(self_);

      SSTMAC_DEBUG << "hpx_thread_manager started on node " << os_->get_my_addr()->to_string() << "\n";
    }

    virtual sstmac::sw::service::ptr
    clone()
    {
      return construct(agas_);
    }

    virtual void
    handle(const sstmac::sst_message::ptr &msg);

    globalid
    create_component(std::string name);

    template<typename ActName>
      void
      register_action(std::string name, boost::intrusive_ptr<ActName> actptr)
      {
        if (actions_.find(name) != actions_.end())
          {
            std::cout << "Action name: " << name << "\n";
            throw sstmac::ssterror(
                "hpx_thread_manager::register_action() -- action already registered here");
          }

        actions_[name] = actptr;

       // os_->register_lib(actptr);
      }

    template<typename CompName>
      void
      register_component(std::string name, boost::intrusive_ptr<CompName> compptr)
      {
        if (comp_factory_.find(name) != comp_factory_.end())
          {
            std::cout << "Component name: " << name << "\n";
            throw sstmac::ssterror(
                "hpx_thread_manager::register_component() -- component already registered here");
          }

        comp_factory_[name] = compptr;


        compptr->register_actions(selfptr_);
      }

    bool
    is_local(sstmac::nodeaddress::ptr loc)
    {
      return loc->equals(os_->get_my_addr());
    }

    void
    start_action(std::string actname, hpx_action_params::ptr params,
        boost::intrusive_ptr<hpx::lcos::base_lco_result> prom,
        sstmac::nodeaddress::ptr locality);

    void
    start_action(std::string actname, hpx_action_params::ptr params,
        boost::intrusive_ptr<hpx::lcos::base_lco_result> prom, globalid gid);
  };

}

#endif
