/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/software/libraries/library.h>

namespace sstmac
{
  namespace sw
  {

    library::library() :
      handler_(eventhandler::ptr()), waiter_(waiter::ptr()),
          self_initialized_(false)
    {

    }

    sst_message::ptr
    library::wait_any()
    {
      if (!waiter_)
        {
          throw libraryerror(
              "library::wait - waiter has not been initialized. If the default library"
                "constructor was called, make sure that init_os was also called");
        }
      current_any_ = wait_any::construct();
      waiter_->wait();
      return current_any_->recent_;
    }

    sst_message::ptr
    library::wait_this(const sst_message::ptr &msg)
    {
      if (!waiter_)
        {
          throw libraryerror(
              "library::wait - waiter has not been initialized. If the default library"
                "constructor was called, make sure that init_os was also called");
        }
      thises_[msg] = wait_this::construct(msg);
      waiter_->wait();
      return msg;
    }

    sst_message::ptr
    library::wait_type(sst_message::MESSAGE_TYPES t)
    {
      if (!waiter_)
        {
          throw libraryerror(
              "library::wait - waiter has not been initialized. If the default library"
                "constructor was called, make sure that init_os was also called");
        }
      current_any_ = wait_type::construct(t);
      waiter_->wait();
      return current_any_->recent_;
    }

    sst_message::ptr
    library::wait_num_type(sst_message::MESSAGE_TYPES t, long num)
    {
      if (!waiter_)
        {
          throw libraryerror(
              "library::wait - waiter has not been initialized. If the default library"
                "constructor was called, make sure that init_os was also called");
        }
      current_any_ = wait_num_type::construct(t, num);
      waiter_->wait();
      return current_any_->recent_;
    }

    void
    library::register_lib(const library::ptr &lib)
    {
      if (os_)
        os_->register_lib(lib);
      else
        throw valueerror("library::register_lib - os not initialized yet");

      otherlibs_.push_back(lib);
    }

    void
    library::init_os(OS_API::ptr os)
    {
      os_ = os;

      if (!handler_)
        {
          key::ptr k = key::construct();
          handler_ = unblockhandler::construct(os_, k);
          waiter_ = blockwaiter::construct(os_, k);
          self_initialized_ = true;
        }
    }

    void
    library::incoming_message(const sst_message::ptr &msg)
    {
      bool handle = false;
      if (thises_.find(msg) != thises_.end())
        {
          wait::ptr w = thises_[msg];
          handle = w->incoming_message(msg);
          thises_.erase(msg);
        }
      else
        {

          if (!current_any_)
            {
              std::cout << "library " << to_string()
                  << " incoming message with no current wait \n";
              throw ssterror("library::incoming message");

            }
          handle = current_any_->incoming_message(msg);
        }

      if (!handler_)
        {
          throw libraryerror(
              "library::wait - handler has not been initialized. If"
                "the default library constructor was called, make sure that init_os was also called");
        }

      if (handle)
        {
          handler_->handle(msg);

        }

    }

    sst_message::ptr
    library::get_arrived_msg()
    {
      if (!self_initialized_)
        {
          throw libraryerror(
              "library::get_arrived_msg - can't use this function unless the"
                "library base class auto-filled the handler");
        }
      unblockhandler::ptr ubh = boost::dynamic_pointer_cast<unblockhandler>(
          handler_);
      return ubh->saved_msg_;
    }

  }
} //end namespace

