/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/software/libraries/mpi/mpi_strategies/mpi_collective/engines/mpibcastengine.h>
#include <sstmac/software/libraries/mpi/mpi_queue/mpiqueue.h>
#include <sstmac/software/libraries/mpi/mpistatus.h>

#include <sstmac/software/libraries/mpi/mpi_strategies/mpi_iterator/topologyiterator.h>
#include <sstmac/common/errors.h>
#include <sstmac/common/debug.h>



namespace sstmac { namespace sw
{

  // Pointer types.
  
  
  using boost::dynamic_pointer_cast;

  // A utility type for sendrecv
  // Although this is very unlikely to happen for bcast, there is the
  // possibility of ending up with recursive calls to sendrecv (and an
  // invalid iterator) if we don't separate iteration from communications
  struct sendrecvop
  {
    bool sending;
    mpiid partner;
  };

  //
  // Start next cycle of sends and receives.
  //
  void
  mpibcastengine::sendrecv()
  {
    //if(debugger::active())
    // debug() << "mpibcastengine[" << this << "]:: sendrecv(" << now << ")\n";
    if (pending_recvs_)
      {
        throw illformederror("mpibcastengine::sendrecvs: Pending receive.");
      }
    std::vector<sendrecvop> ops;
    while (iter_->active())
      {
        if (iter_->sending())
          {
            sendrecvop op =
              { true, mpiid(iter_->dest()) };
            ops.push_back(op);
          }
        if (iter_->receiving())
          {
            sendrecvop op =
              { false, mpiid(iter_->source()) };
            ops.push_back(op);
            // We can only safely handle one receive at a time.
            // (a bcast will generally receive zero or one time,
            // followed by zero or more sends).
            iter_->next();
            break;
          }
        iter_->next();
      }
    // Do any sends/receives.
    for (size_t it = 0; it < ops.size(); ++it)
      {
        if (!ops[it].sending)
          {
            ++pending_recvs_;
            this->start_recv(count_, type_, tag_, ops[it].partner);
          }
        else
          {
            ++pending_sends_;
            this->start_send(count_, type_, tag_, ops[it].partner, content_);
          }
      }
    // If we have no pending sends or receives, we are done.
    if (!completing_ && pending_sends_ == 0 && pending_recvs_ == 0)
      {
        completing_ = true;
        mpicollective::complete(content_);
      }
  }

  //
  // Hi.
  //
  mpibcastengine::mpibcastengine(const topologyiterator::ptr& iter,
      const mpirequest::ptr& thekey, const mpiqueue::ptr& queue,
      int count, const mpitype &type, const mpiid &root, const mpitag &tag,
      const mpicomm &comm, const payload::const_ptr& content,
      eventhandler::ptr completion) :
    mpicollective(thekey, queue, comm, completion), iter_(iter), count_(count),
        type_(type), tag_(tag), rank_(comm.rank()), pending_sends_(0),
        pending_recvs_(0)
  {
    completing_ = false;
    if (comm.rank() == root)
      {
        if (content)
          content_ = content;
      }
  }

  //
  // Callback method to indicate that a send operation has completed.
  //
  void
  mpibcastengine::send_complete(int count, const mpitype &type,
      const mpitag &tag, const mpiid &dest)
  {
    if (dbg_.is_active())
      dbg_ << "mpibcastengine[" << comm_.rank().id << "]::send_complete \n";
    //          << count << ", " << type << ", " << tag << ", " << dest << ")\n";
    --pending_sends_;
    if (!pending_recvs_)
      this->sendrecv();
  }

  //
  // Callback method to indicate that a receive operation has completed.
  //
  void
  mpibcastengine::recv_complete(int count, const mpistatus_t& stat)
  {
    if (dbg_.is_active())
      dbg_ << "mpibcastengine[" << comm_.rank().id << "]::recv_complete \n";
    //           << ", " << count << stat << ")\n";
    content_ = stat->content();
    // this->start_immediate_copy(count, stat->recv_type(),
    //                            stat->tag(), stat->source());
    --pending_recvs_;
    this->sendrecv();
  }

  //
  // Completion of a copy event.
  //
  void
  mpibcastengine::copy_complete(int count, const mpitype &type,
      const mpitag &tag, const mpiid &source)
  {
    //  if(debugger::active())
    //    debug() << "mpibcastengine[" << this << "]::copy_complete(" << now
    //           << ", " << count << ", " << type << ", " << tag << ", "
    //           << source << ")\n";
    --pending_recvs_;
    this->sendrecv();
  }

  //
  // Hello.
  //
  mpibcastengine::ptr
  mpibcastengine::construct(const topologyiterator::ptr& iter,
      const mpirequest::ptr& thekey, const mpiqueue::ptr& queue,
      int count, const mpitype &type, const mpiid &root, const mpitag &tag,
      const mpicomm &comm, const payload::const_ptr& content,
      eventhandler::ptr completion)
  {
    mpibcastengine::ptr retval(new mpibcastengine(iter, thekey, queue,
        count, type, root, tag, comm, content, completion));

    return retval;
  }

  //
  // Goodbye.
  //
  mpibcastengine::~mpibcastengine() throw ()
  {
  }

  //
  // Start this kernel.
  //
  void
  mpibcastengine::start()
  {
    //if(debugger::active())
    //  debug() << "mpibcastengine[" << this << "]::start(" << start_time
    //          << ", " << my_node << ")\n";
    mpicollective::start();
    this->sendrecv();
  }

} } // end of namespace sstmac

