/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/software/libraries/mpi/mpi_strategies/mpi_collective/engines/mpidirectalltoall.h>
#include <sstmac/software/libraries/mpi/mpi_strategies/mpi_collective/mpicollpayload.h>
#include <sstmac/common/errors.h>
#include <sstmac/common/debug.h>

namespace sstmac { namespace sw
{

  // Pointer types.
  
  using boost::dynamic_pointer_cast;

  //
  // Hello.
  //
  mpidirectalltoall::mpidirectalltoall(const mpirequest::ptr& thekey,
      const mpiqueue::ptr& queue, const std::vector<int> &sendcnt,
      const mpitype &sendtype, const std::vector<int> &recvcnt,
      const mpitype &recvtype, const mpitag &tag, const mpicomm &comm,
      const std::vector<payload::const_ptr >& content,
      eventhandler::ptr completion) :
    mpicollective(thekey, queue, comm, completion), sendcnt_(sendcnt),
        sendtype_(sendtype), recvcnt_(recvcnt), recvtype_(recvtype), tag_(tag),
        comm_(comm), max_concurrent_ops_(5)
  {
    if (sendcnt_.size() != size_t(comm_.size().id) || recvcnt_.size()
        != size_t(comm_.size().id))
      throw valueerror("mpidirectalltoall(...):  Length of send- and/or recv- "
        "arrays is not the same as communicator size.");
    // Clone the content if it is non-null.
    //magic_->add_collective_payload(typeid(*this), tag_, comm_,
    //                               (content ? content->clone() : content),
    //                               true);
    send_content_ = content;

    for (int i = 0; i < send_content_.size(); i++)
      {
        recv_content_.push_back(payload::const_ptr ());
      }

    recv_content_[comm.rank().id] = send_content_[comm.rank().id];
  }

  //
  // Callback method to indicate that a send operation has completed.
  //
  void
  mpidirectalltoall::send_complete(int count, const mpitype &type,
      const mpitag &tag, const mpiid &dest)
  {
    if (dbg_.is_active())
      dbg_ << "mpidirectalltoall[" << comm_.rank().id << "]::send_complete( "
          << dest.id << ", " << count << ", " << type.id << ", " << tag.id
          << ", " << dest.id << ") psend = " << pending_sends_ << ", precv = "
          << pending_recvs_ << "\n";

    --pending_sends_;

    if (!pendsend_.empty())
      {
        pendsend snd = pendsend_.front();
        pendsend_.pop_front();
        pending_sends_++;
        this->start_send(snd.cnt, sendtype_, tag_, snd.dest,
            send_content_[snd.dest.id]);
      }
    else
      {

        if ((pending_sends_ == 0) && (pending_recvs_ == 0) && pendrecv_.empty() && pendsend_.empty())
          {
            // if (!magic_->have_all_collective_payload(typeid(*this), tag_, comm_))
            //  throw illformederror(
            //      "mpidirectalltoall::send_complete:  Transfers "
            //        "done but not all payload has been gathered.");
            //payload::const_ptr content = magic_->get_collective_payload(
            //    typeid(*this), tag_, comm_);
            mpicollective::complete(mpicollpayload::construct(recv_content_));
          }
      }
  }

  //
  // Callback method to indicate that a receive operation has completed.
  //
  void
  mpidirectalltoall::recv_complete(int count,
      const mpistatus_t& stat)
  {
    if (dbg_.is_active())
      dbg_ << "mpidirectalltoall[" << comm_.rank().id << "]::recv_complete( "
          << stat->source().id << ", " << count << ", " << stat << ") psend = "
          << pending_sends_ << ", precv = " << pending_recvs_ << "\n";
    //this->start_immediate_copy(count, stat->recv_type(),
    //                           stat->tag(), stat->source());

    payload::const_ptr other = dynamic_pointer_cast<const payload> (
        stat->content());
    int index = stat->source().id;
    recv_content_[index] = other;

    --pending_recvs_;

    /** DEPRECATED JJW 01/15/2012 Do not chain recvs. This generates a huge call
    if (!pendrecv_.empty())
      {
            stack that breaks the program. 
        pendrecv rcv = pendrecv_.front();
        pendrecv_.pop_front();
        pending_recvs_++;
        this->start_recv(rcv.cnt, recvtype_, tag_, rcv.source);
      }
    else
      {

      }
     */
    if ((pending_sends_ == 0) && (pending_recvs_ == 0) && pendrecv_.empty() && pendsend_.empty())
      {
        mpicollective::complete(mpicollpayload::construct(recv_content_));
      }
  }

  //
  // Callback to signal copy completion (only on clients).
  //
  void
  mpidirectalltoall::copy_complete(int count, const mpitype &type,
      const mpitag &tag, const mpiid &source)
  {
    if (dbg_.is_active())
      dbg_ << "mpidirectalltoall[" << comm_.rank().id << "]::copy_complete( "
          << ", " << count << ", " << type.id << ", " << tag.id << ", "
          << source.id << ") -- pending recvs: " << pending_recvs_ << "\n";
    if (!pendrecv_.empty())
      {
        pendrecv rcv = pendrecv_.back();
        pendrecv_.pop_back();
        this->start_recv(rcv.cnt, recvtype_, tag_, rcv.source);
      }
    else
      {
        --pending_recvs_;
        if ((pending_sends_ == 0) && (pending_recvs_ == 0))
          {
            // if(! magic_->have_all_collective_payload(typeid(*this), tag_, comm_))
            //  throw illformederror("mpidirectalltoall::copy_complete:  Transfers "
            //                       "done but not all payload has been gathered.");
            // payload::const_ptr content =
            //  magic_->get_collective_payload(typeid(*this), tag_, comm_);
            // mpicollective::complete();
          }
      }
  }

  //
  // Hello.
  //
  mpidirectalltoall::ptr
  mpidirectalltoall::construct(const mpirequest::ptr& thekey,
      const mpiqueue::ptr& queue, const std::vector<int> &sendcnt,
      const mpitype &sendtype, const std::vector<int> &recvcnt,
      const mpitype &recvtype, const mpitag &tag, const mpicomm &comm,
      const std::vector<payload::const_ptr >& content,
      eventhandler::ptr completion)
  {
    mpidirectalltoall::ptr retval(new mpidirectalltoall(thekey, queue,
        sendcnt, sendtype, recvcnt, recvtype, tag, comm, content, completion));

    return retval;
  }

  //
  // Goodbye.
  //
  mpidirectalltoall::~mpidirectalltoall() throw ()
  {
  }

  //
  // Set the maximum number of concrurrent sends.
  //
  void
  mpidirectalltoall::set_max_concurrent_ops(int maxval)
  {
    max_concurrent_ops_ = maxval;
  }

  // Utility routine.
  template<typename T>
    T
    min(T a, T b)
    {
      return (a < b ? a : b);
    }

  //
  // Start this kernel.
  //
  void
  mpidirectalltoall::start()
  {
    if (dbg_.is_active())
      dbg_ << "mpidirectalltoall[" << comm_.rank().id << "]::start( )\n";
    mpicollective::start();
    // Prepare to fire off all my sends and receives.
    mpiid rank = comm_.rank();
    mpiid size = comm_.size();
    // Catch the case where comm_.size() was 1.
    if (size.id <= 1)
      {
        // if(! magic_->have_all_collective_payload(typeid(*this), tag_, comm_))
        //   throw illformederror("mpidirectalltoall::start:  Transfers "
        //                       "done but not all payload has been gathered.");

        // payload::const_ptr content =
        //   magic_->get_collective_payload(typeid(*this), tag_, comm_);

        mpicollective::complete(mpicollpayload::construct(send_content_));
      }
    else
      {
        pending_sends_ = 0;
        pending_recvs_ = 0;
        int numactive = (max_concurrent_ops_ > 0 ? min(max_concurrent_ops_,
            size.id - 1) : size.id - 1);
        // Stagger the sends so not everybody hammers node 0 at the same time.
        for (int relrank = 1; relrank < size.id; ++relrank)
          {
            mpiid peer((size.id + rank.id - relrank) % size.id);
            if (peer != rank)
              {
                if (relrank <= numactive)
                  {
                    pending_sends_++;
                    this->start_send(sendcnt_.at(peer.id), sendtype_, tag_,
                        peer, send_content_[peer.id]);
                  }
                else
                  pendsend_.push_back(pendsend(sendcnt_.at(peer.id), peer));
              }
          }
        // Stagger the receives in the opposite order from the sends
        // to avoid deadlock.
        for (int relrank = 1; relrank <= size.id; ++relrank)
          {
            mpiid peer((size.id + rank.id + relrank) % size.id);
            if (peer != rank)
              {
                if (relrank <= numactive)
                  {
                    pending_recvs_++;
                    this->start_recv(sendcnt_.at(peer.id), sendtype_, tag_,
                        peer);
                  }
                else
                  pendrecv_.push_back(pendrecv(recvcnt_.at(peer.id), peer));
              }
          }

        /** DEPRECATED JJW 01/15/2012 Start all recvs at the beginning 
        if (!pendrecv_.empty() && pending_recvs_ == 0) //this is possible, believe it or not
          {
            pendrecv rcv = pendrecv_.front();
            pendrecv_.pop_front();
            pending_recvs_++;
            this->start_recv(rcv.cnt, recvtype_, tag_, rcv.source);
          }
        */

        /** Start all the recvs at one to avoid generating a HUGE call stack from chained recvs
        */
        while (!pendrecv_.empty())
        {
            pendrecv rcv = pendrecv_.front();
            pendrecv_.pop_front();
            pending_recvs_++;
            this->start_recv(rcv.cnt, recvtype_, tag_, rcv.source);
        }

      }
  }

} } // end of namespace sstmac
