/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/software/libraries/mpi/mpi_strategies/mpi_collective/engines/mpihalvingreducescatter.h>

#include <sstmac/common/errors.h>
#include <sstmac/common/debug.h>

namespace sstmac { namespace sw {

  // Pointer types.

  using boost::dynamic_pointer_cast;

  //
  // A private method to perform data exchanges.
  //
  void mpihalvingreducescatter::sendrecv() {
    SSTMAC_DEBUG << "mpihalvingreducescatter[" << comm_.rank().id << "]::sendrecv, i have " << pending_sends_ << " sends left and " << pending_recvs_ << " recvs left \n";
    if(pending_recvs_) // We can't send anything until our receive is done.
      return;
    if(! peers_.empty()) {
        commstep curr = peers_.front();
        peers_.pop_front();
        // Incrementing sends and receives separately - just so we don't
        // end up entering sendrecv too early.
        if(curr.sendbytes > 0) ++pending_sends_;
        if(curr.recvbytes > 0) ++pending_recvs_;
        if(curr.sendbytes > 0)
          this->start_send(curr.sendbytes, type_, tag_, curr.peer);
        if(curr.recvbytes > 0)
          this->start_recv(curr.recvbytes, type_, tag_, curr.peer);
    }
    else if(pending_sends_ == 0 && pending_recvs_ == 0) {
        // We're done.

        mpicollective::complete(content_);
    }
  }

  //
  // Hello.
  //
  mpihalvingreducescatter::
  mpihalvingreducescatter(const mpirequest_t& thekey,
      const mpiqueue::ptr& queue,
      const std::vector<int> &recvcnts,
      const mpitype &type, const mpiop &op,
      const mpitag &tag, const mpicomm &comm,
      const payload::const_ptr& content,
      eventhandler::ptr completion) :
      mpicollective(thekey, queue, comm, completion),
      recvcnts_(recvcnts), type_(type), op_(op), tag_(tag), comm_(comm),
      pending_sends_(0), pending_recvs_(0)
  {
    // Clone the content if it is non-null
    std::vector<payload::const_ptr > vals;
    for (int i = 0; i < comm.size().id; i++)
      {
        if (i == comm.rank().id)
          vals.push_back(content);
        else
          vals.push_back(payload::const_ptr ());
      }

    content_ = mpicollpayload::construct(vals);


    if(recvcnts_.size() != size_t(comm.size().id))
      throw valueerror("mpihalvingreducescatter(...): Number of peers "
          "differs from the number of receive counts.");
    std::vector<int>::const_iterator it, end = recvcnts.end();
    for(it = recvcnts.begin(); it != end; ++it)
      if(*it < 0)
        throw valueerror("mpihalvingreducescatter(...):  "
            "Found negative receive count.");
  }

  //
  // Callback method to indicate that a send operation has completed.
  //
  void mpihalvingreducescatter::send_complete(
      int count,
      const mpitype &type,
      const mpitag &tag,
      const mpiid &dest)
  {
    SSTMAC_DEBUG<< "mpihalvingreducescatter[" << comm_.rank().id << "]::send_complete(" << ", "
        << count << ", " << type.to_string() << ", " << tag.to_string() << ", " << dest.to_string() << ")\n";
    --pending_sends_;
    this->sendrecv();
  }

  //
  // Callback method to indicate that a receive operation has completed.
  //
  void mpihalvingreducescatter::recv_complete(
      int count,
      const mpistatus_t& stat)
  {
    SSTMAC_DEBUG << "mpihalvingreducescatter[" << comm_.rank().id << "]::recv_complete(" << ", "
        << count << ", " << stat <<")\n";

   // this->start_funneled_copy(count, stat->recv_type(),
    //    stat->tag(), stat->source());

    //payload::const_ptr load = stat->content();

   // if (reduce_this_recv_)
    //  {
       // if (load)
      //    {
       //     content_ = combine_content(content_, load, op_);
       //   }
     // }

    --pending_recvs_;
       this->sendrecv();

  }

  //
  // Callback to signal copy completion.
  //
  void mpihalvingreducescatter::copy_complete(
      int count,
      const mpitype &type,
      const mpitag &tag,
      const mpiid &source)
  {
    SSTMAC_DEBUG << "mpihalvingreducescatter[" << comm_.rank().id << "]::copy_complete(" << ", "
        << count << ", " << type.to_string() << ", " << tag.to_string() << ", " << source.to_string() << ")\n";
    --pending_recvs_;
    this->sendrecv();
  }

  //
  // Hello.
  //
  mpihalvingreducescatter::ptr
  mpihalvingreducescatter::construct(const mpirequest_t& thekey,
      const mpiqueue::ptr& queue,
      const std::vector<int> &cnts,
      const mpitype &type, const mpiop &op,
      const mpitag &tag, const mpicomm &comm,
      const payload::const_ptr& content,
      eventhandler::ptr completion)
  {
    mpihalvingreducescatter::ptr retval
    (new mpihalvingreducescatter(thekey, queue, cnts, type,
        op, tag, comm, content, completion));

    return retval;
  }

  //
  // Goodbye.
  //
  mpihalvingreducescatter::~mpihalvingreducescatter() throw() {
  }

  //
  // Utility routine to find communication partner given a pivot.
  //
  inline int partner(int rank, int pivot) {
    if(rank >= (pivot<<1))
      throw valueerror("mpihalvingreducescatter::start::partner: "
          "rank out of valid range.");
    return (rank < pivot ? rank + pivot : rank - pivot);      
  }
  //
  // Utility routine to find the power-of-two rank for a given index.
  //
  inline int to_pof2(const mpiid &rank, int extras) {
    if(rank.id >= 2*extras) return rank.id-extras;
    else return (rank.id%2 ? -1 : rank.id/2);
  }
  //
  // Utility routine to convert a power-of-two rank to an MPI rank.
  //
  inline mpiid from_pof2(int rank, int extras) {
    if(rank < 0)
      throw valueerror("mpihalfringreducescatter::start::from_pof2: "
          "Negative rank.");
    return mpiid(rank > extras ? rank+extras : rank*2);
  }
  //
  // Utility routine to find the greatest power of two not greater than value
  //
  inline int msb(int value) {
    if(value < 0)
      throw valueerror("mpihalvingreducescatter::start::msb: Negative value");
    int rv = 1;
    while(rv <= value)
      rv <<= 1;
    return (rv>>1);
  }

  //
  // Get busy child.
  //
  void mpihalvingreducescatter::start()
  {
    SSTMAC_DEBUG << "mpihalvingreducescatter[" << comm_.rank().id << "]::start()\n";
    mpicollective::start();
    //
    // Figure out our entire pattern of sends and receives.
    const int size = msb(comm_.size().id);
    const int extras = comm_.size().id - size;
    const int rank = to_pof2(comm_.rank(), extras);
    // Set up mappings of how much we are sending/receiving.
    std::vector<int> sizes;  sizes.reserve(size);
    int total_elements = 0;
    for(size_t it = 0; it < recvcnts_.size(); ++it) {
        total_elements += recvcnts_.at(it);
        if(to_pof2(mpiid(it), extras) < 0)
          sizes.back() += recvcnts_[it];
        else
          sizes.push_back(recvcnts_[it]);
    }
    //
    // First, check whether we need to collapse this node.
    if(rank < extras) {
        if(rank < 0) {
            // This node needs to pre-send all it's data down.
            peers_.push_back(commstep(mpiid(comm_.rank().id-1), total_elements, 0));
        }
        else {
            peers_.push_back(commstep(mpiid(comm_.rank().id+1), 0, total_elements));
        }
    }
    //
    // Next, set up all the normal (power-of-two) data exchanges.
    if(rank >= 0) {
        int pivot = size;
        while(pivot > 1) {
            const int cap = pivot;
            pivot >>= 1;
            // Figure out where we're sending
            const int relrank = rank % cap;
            const int shift = rank - relrank;
            const int pal = partner(relrank, pivot) + shift;
            // Figure out what we're sending
            int lowframe[]  = { shift, pivot+shift };
            int highframe[] = { pivot+shift, cap+shift };
            int *recvframe = (relrank < pivot ? &lowframe[0]  : &highframe[0]);
            int *sendframe = (relrank < pivot ? &highframe[0] : &lowframe[0]);
            int sendamt = 0, recvamt = 0;
            for(int it = recvframe[0]; it < recvframe[1]; ++it)
              sendamt += sizes.at(it);
            for(int it = sendframe[0]; it < sendframe[1]; ++it)
              recvamt += sizes.at(it);
            // and add it to the queue of send/receive pairs.
            peers_.push_back(commstep(from_pof2(pal,extras), sendamt, recvamt));
        }
    }
    //
    // Finally, check whether we need to send back any data.
    if(rank < extras) {
        if(rank < 0) {
            // We need to get our data back from below.
            peers_.push_back(commstep(mpiid(comm_.rank().id-1),
                0, recvcnts_.at(comm_.rank().id)));
        }
        else {
            peers_.push_back(commstep(mpiid(comm_.rank().id+1),
                recvcnts_.at(comm_.rank().id+1), 0));
        }
    }
    // and now we can go.
    this->sendrecv();
  }


} } // end of namespace sstmac
