/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/software/libraries/mpi/mpi_strategies/mpi_collective/engines/mpiscanengine.h>
#include <sstmac/software/libraries/mpi/mpi_strategies/mpi_collective/mpicollpayload.h>
#include <sstmac/common/eventhandler.h>

#include <sstmac/software/libraries/mpi/mpi_strategies/mpi_iterator/topologyiterator.h>
#include <sstmac/common/errors.h>
#include <sstmac/common/debug.h>

namespace sstmac { namespace sw {
/*
  // Pointer types.
  
  
  using boost::dynamic_pointer_cast;

  // A utility type for sendrecv
  // Although this is very unlikely to happen for bcast, there is the
  // possibility of ending up with recursive calls to sendrecv (and an
  // invalid iterator) if we don't separate iteration from communications
  struct sendrecvop {
    bool sending;
    mpiid partner;
  };

  //
  // Start next cycle of sends and receives.
  //
  void mpiscanengine::sendrecv(const timestamp &now) {
    if(debugger::active())
      debug() << "mpiscanengine[" << this << "]:: sendrecv(" << now << ")\n";
    if(pending_recvs_) {
      throw illformederror("mpiscanengine::sendrecvs: Pending receive.");
    }
    std::vector<sendrecvop> ops;
    while(iter_->active()) {
      if(iter_->sending() && iter_->dest() > comm_.rank().id) {
        sendrecvop op = { true, mpiid(iter_->dest()) };
        ops.push_back(op);
      }
      if(iter_->receiving() && iter_->source() < comm_.rank().id) {
        sendrecvop op = { false, mpiid(iter_->source()) };
        ops.push_back(op);
        // We can only safely handle one receive at a time.
        // (a scan will generally receive zero or one time,
        // followed by zero or more sends).
        iter_->next();
        break;
      }
      iter_->next();
    }
    // Do any sends/receives.
    for(size_t it = 0; it < ops.size(); ++it) {
      if(! ops[it].sending) {
        ++pending_recvs_;
        this->start_recv(count_, type_, tag_, ops[it].partner);
      }
      else {
        ++pending_sends_;
        this->start_send(count_, type_, tag_, ops[it].partner, content_);
      }
    }
    // If we have no pending sends or receives, we are done.
    if(pending_sends_ == 0 && pending_recvs_ == 0) {
      mpicollective::complete(now, content_);
    }
  }

  //
  // Hi.
  //
  mpiscanengine::
  mpiscanengine(const topologyiterator::ptr& iter,
                const boost::shared_ptr<processrequest>& thekey,
                const mpiqueue::ptr& queue,
                int count, const mpitype &type, 
                const mpiop &op,
                const mpitag &tag, const mpicomm &comm,
                const payload::const_ptr& content,
                eventhandler::ptr completion) :
    mpicollective(thekey, queue, comm, completion), iter_(iter),
    count_(count), type_(type), op_(op), tag_(tag), comm_(comm),
    pending_sends_(0), pending_recvs_(0)
  {
    content_ = mpicollpayload::construct(comm.rank().id + 1);
    // Clone the content if it is non-null.
    content_->set(comm.rank(), (content ? content->clone() : content));
  }

  //
  // Callback method to indicate that a send operation has completed.
  //
  void mpiscanengine::send_complete(
                                     int count,
                                     const mpitype &type,
                                     const mpitag &tag,
                                     const mpiid &dest)
  {
    if(debugger::active())
      debug() << "mpiscanengine[" << this << "]::send_complete(" << now << ", "
              << count << ", " << type << ", " << tag << ", " << dest << ")\n";
    --pending_sends_;
    if(! pending_recvs_)
      this->sendrecv(now);

  }

  //  
  // Callback method to indicate that a receive operation has completed.
  //
  void mpiscanengine::recv_complete(
                                    int count,
                                    const mpistatus::const_ptr& stat)
  {
    if(debugger::active())
      debug() << "mpibcastengine[" << this << "]::recv_complete(" << now << ", "
              << count << ", " << stat << ")\n";
    // Right now, we are simulating the reduce as being simply a copy,
    // but we can change this any which way we want.
    shared_ptr<const mpicollpayload> other =
      dynamic_pointer_cast<const mpicollpayload>(stat->content());
    mpiid sender = stat->source();
    if(! other)
      throw valueerror("mpiscanengine::recv_complete:  Not valid payload.");
    const mpicollpayload::content_t &data = other->value();
    if(data.size() <= sender.id)
      throw illformederror("mpiscanengine::recv_complete:  Received data "
                           "with incorrect lenght.");
    for(mpiid nbor = mpiid(0); nbor <= sender; ++nbor)
      content_->set(nbor, other->get(nbor));
    this->start_funneled_copy(count, stat->recv_type(),
                              stat->tag(), stat->source());
  }

  //
  // Completion of a copy event.
  //
  void mpiscanengine::copy_complete( int count,
                                    const mpitype &type, const mpitag &tag,
                                    const mpiid &source)
  {
    if(debugger::active())
      debug() << "mpiscanengine[" << this << "]::copy_complete( " << now << ", "
              << count << ", " << type << ", " << tag << ", " << source
              << ")\n";
    --pending_recvs_;
    if(! pending_recvs_)
      this->sendrecv(now);
  }

  //
  // Make a new engine.
  //
  shared_ptr<mpiscanengine>
  mpiscanengine::construct(const topologyiterator::ptr& iter,
                           const boost::shared_ptr<processrequest>& thekey,
                           const mpiqueue::ptr& queue,
                           int count, const mpitype &type, 
                           const mpiop &op,
                           const mpitag &tag, const mpicomm &comm,
                           const payload::const_ptr& content,
                           eventhandler::ptr completion)
  {
    shared_ptr<mpiscanengine> retval
      (new mpiscanengine(iter, thekey, queue, count, type,
                         op, tag, comm, content, completion));
    retval->set_self(retval);
    return retval;
  }

  //
  // Goodbye.
  //
  mpiscanengine::~mpiscanengine() throw() {
  }

  void mpiscanengine::start(const timestamp &now,
                            const boost::shared_ptr<nodeinterface>& my_node)
  {
    if(debugger::active())
      debug() << "mpiscanengine::start(" << now << ", " << my_node << ")\n";
    mpicollective::start(now, my_node);
    this->sendrecv(now);
  }
*/

} } // end of namespace sstmac

