/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/software/libraries/mpi/mpiserver.h>

#include <sstmac/software/libraries/mpi/mpimessage.h>
#include <sstmac/common/messages/sst_message.h>
#include <sstmac/software/libraries/mpi/mpi_queue/mpiqueue.h>

#include <sstmac/common/errors.h>

namespace sstmac
{
  namespace sw
  {

    using boost::dynamic_pointer_cast;

    using namespace std;

    debug mpiserver::dbg_;

    void
    mpiserver::handle(const sst_message::ptr &msg)
    {
      SSTMAC_DEBUG << "mpiserver[" << os_->get_my_addr()->to_string()
            << "]: handle message " << msg->to_string() << " \n";
      if (msg->get_type() == sst_message::COMPUTE)
        {

          mem_access::ptr mem = boost::dynamic_pointer_cast<mem_access>(msg);

          if (!mem)
            {
              throw ssterror(
                  "mpiserver::handle -- could not cast compute message to memory access");
            }

          bool handled = false;
          if (computing_recv_.size() > 0)
            {
              mpimessage::ptr recvfront = computing_recv_.front();

              long tidrecv = get_tid_from_mess(recvfront, false);

              if (mem->dstthread_ == tidrecv)
                {

                  computing_recv_.pop_front();
                  handle_recv(recvfront);
                  handled = true;
                }
            }

          if (computing_send_.size() > 0 && !handled)
            {

              mpimessage::ptr sendfront = computing_send_.front();
              long tidsend = get_tid_from_mess(sendfront, true);
              if (mem->dstthread_ == tidsend)
                {

                  computing_send_.pop_front();
                  handle_send(sendfront);
                }

            }

          computing_ = false;
          compute();
        }
      else
        {
          mpimessage::ptr mpi = boost::dynamic_pointer_cast<mpimessage>(msg);
          pending_recv_.push_back(mpi);
          compute();
        }
    }

    long
    mpiserver::get_tid_from_mess(mpimessage::ptr mess, bool is_send)
    {
      mpiid theid;

      if (is_send || mess->get_type() == sst_message::SEND_ACK)
        {
          theid = mess->source();
        }
      else
        {
          theid = mess->dest();
        }
      if (sids_.find(theid.id) == sids_.end())
        {
          return -1;
        }
      softwareid sid = sids_[theid.id];
      SSTMAC_DEBUG << "which is software id " << sid.to_string() << "\n";
      app::ptr a = fullos_->get_app_ptr(sid);
      if (!a)
        {
          return -1;
        }
      return a->get_id();
    }

    void
    mpiserver::compute()
    {
      // if (!computing_)
      //  {
      SSTMAC_DEBUG << "mpiserver[" << os_->get_my_addr()->to_string()
            << "]: computing " << " \n";

      if (pending_recv_.size() > 0) // do recv's first

        {

          //first, get enough time on the processor
          size_t numcomputing = pending_recv_.size(); //save the number we are computing for, in case some arrive while we're computing


          SSTMAC_DEBUG << "mpiserver[" << os_->get_my_addr()->to_string()
                << "]: " << numcomputing << " recv's \n";

          //long bytes = 0;


          list<mpimessage::ptr>::iterator it = pending_recv_.begin();
          for (size_t i = 0; i < numcomputing; i++)
            {
              SSTMAC_DEBUG << "receiving at task " << (*it)->dest().id << "\n";
              long size = (*it)->get_byte_length();
              if ((*it)->get_type() == sst_message::SEND_ACK)
                {
                  size = 32;
                }

              lib_mem_->compute_block_access(size, -1,
                  get_tid_from_mess(*it, false));
              it++;
              computing_ = true;
            }

          /* list<mpimessage::ptr>::iterator it = pending_recv_.begin();
           for (size_t i = 0; i < numcomputing; i++)
           {


           mpimessage::ptr mess = *it;
           bytes += mess->get_byte_length();
           it++;

           }*/

          //lib_mem_->compute_block_access(bytes);

          for (size_t i = 0; i < numcomputing; i++)
            {
              mpimessage::ptr mess = pending_recv_.front();
              pending_recv_.pop_front();
              computing_recv_.push_back(mess);
            }
          // computing_ = true;
        }

      if (pending_send_.size() > 0) // now do sends

        {

          //first, get enough time on the processor
          size_t numcomputing = pending_send_.size(); //save the number we are computing for, in case some arrive while we're computing

          SSTMAC_DEBUG << "mpiserver[" << os_->get_my_addr()->to_string()
                << "]: " << numcomputing << " send's \n";

          // long bytes = 0;

          list<mpimessage::ptr>::iterator it = pending_send_.begin();
          for (size_t i = 0; i < numcomputing; i++)
            {
              SSTMAC_DEBUG << "sending from task " << (*it)->source().id
                    << "\n";

              lib_mem_->compute_block_access((*it)->get_byte_length(), -1,
                  get_tid_from_mess(*it, true));
              it++;
              computing_ = true;
            }

          /*for (size_t i = 0; i < numcomputing; i++)
           {
           mpimessage::ptr mess = *it;
           bytes += mess->get_byte_length();
           it++;

           }*/

          // lib_mem_->compute_block_access(bytes);

          for (size_t i = 0; i < numcomputing; i++)
            {
              mpimessage::ptr mess = pending_send_.front();
              pending_send_.pop_front();
              computing_send_.push_back(mess);
            }
          // computing_ = true;
        }

      /* if (bytes >= 0)
       {
       computing_ = true;
       lib_mem_->compute_block_access(bytes);
       }*/

      //  }
    }

    //
    // Define a task for this server.  Each task can only be defined once
    //
    void
    mpiserver::add_task(const softwareid &sid, const mpiqueue::ptr& peer,
        int rank)
    {
      if (peer == 0)
        {
          throw nullerror("mpiserver::add_task:  null peer pointer.");
        }
      queue_t::iterator it = queue_.find(sid);
      if (it != queue_.end())
        {
          throw illformederror(
              "mpiserver::add_task:  Attempt to redefine peer.");
        }
      queue_[sid] = peer;

      sids_[rank] = sid;
    }

    void
    mpiserver::unregister(const softwareid &sid)
    {
      queue_t::iterator it = queue_.find(sid);
      if (it == queue_.end())
        {
          throw illformederror(
              "mpiserver::unregister -- can't find sid in queue list.");
        }
      queue_.erase(sid);

    }

    //
    // Test whether the given task is defined on this node.
    //
    bool
    mpiserver::has_task(const softwareid &tid) const
    {
      return (queue_.find(tid) != queue_.end());
    }

    //
    // Send data.
    //
    void
    mpiserver::send(const mpimessage::ptr& payload)
    {
      pending_send_.push_back(payload);
      compute();
    }

    void
    mpiserver::handle_send(mpimessage::ptr &payload)
    {
      SSTMAC_DEBUG << "mpiserver[" << os_->get_my_addr()->to_string()
            << "]: sending...\n";
      taskid recver = payload->dest_task();
      softwareid sid(payload->app(), recver);
      queue_t::iterator it = queue_.find(sid);
      if (it != queue_.end())
        {
          // Transfer between processes on local node.
          SSTMAC_DEBUG << "mpiserver: copy to local task \n";
          pending_recv_.push_back(payload);
          //it->second->incoming_message(payload);

          //push a nic ack to the source also
          mpimessage::ptr nicack = boost::dynamic_pointer_cast<mpimessage>(
              payload->clone(sst_message::SEND_ACK));

          taskid sender = payload->source_task();
          softwareid sid2(payload->app(), sender);
          queue_t::iterator it2 = queue_.find(sid2);
          if (it2 != queue_.end())
            {
              it2->second->incoming_message(nicack);
            }

          compute();
        }
      else
        {
          // Network transfer.
          SSTMAC_DEBUG << "mpiserver: executing os kernel to send remote \n";

          library::os_->execute_kernel(AMI::AMI_COMM, AMI::AMI_COMM_SEND,
              payload);
        }
    }

    //
    // Complete a receive.
    void
    mpiserver::handle_recv(mpimessage::ptr &mess)
    {
      SSTMAC_DEBUG << "mpiserver[" << os_->get_my_addr()->to_string()
            << "]: handling recv - " << mess->to_string() << "\n";
      if (mess == 0)
        {
          throw illformederror("mpiserver::incoming_message:  Message is not "
            "of type mpimessage.");
        }

      queue_t::iterator it;
      if (mess->get_type() == sst_message::SEND_ACK)
        {
          taskid sender = mess->source_task();
          softwareid sid(mess->app(), sender);
          SSTMAC_DEBUG << "mpiserver[" << os_->get_my_addr()->to_string()
                << "]: checking for sender \n";
          it = queue_.find(sid);
        }
      else
        {
          taskid recver = mess->dest_task();
          softwareid sid(mess->app(), recver);
          SSTMAC_DEBUG << "mpiserver[" << os_->get_my_addr()->to_string()
                << "]: checking for receiver \n";
          it = queue_.find(sid);
        }

      if (it == queue_.end())
        {

          std::cerr << "mpiserver[" << os_->get_my_addr()->to_string()
              << "]::incoming_message:  " << "For app " << mess->app()
              << ", failed to find receiver " << mess->dest_task()
              << " or sender " << mess->source_task()
              << " in the task queues.  " << "My tasks are: [ ";
          for (queue_t::iterator a = queue_.begin(); a != queue_.end(); ++a)
            {
              std::cerr << a->first << ", ";
            }
          std::cerr << "]\n";
          throw illformederror(
              "mpiserver::incoming_message:  Message for task "
                "not registered with this server.");

        }
      else
        {
          SSTMAC_DEBUG << "mpiserver[" << os_->get_my_addr()->to_string()
                << "]: delivering to queue " << it->first.to_string() << "\n";
          it->second->incoming_message(mess);
        }
    }

  }
} // end of namespace sstmac
