/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009-2011 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <mpi/sstmac_mpi.h>

#include <sstmac/software/libraries/mpi/mpistatus.h>
#include <sstmac/software/libraries/mpi/mpirequest.h>
#include <sstmac/software/libraries/mpi/mpi_types/mpitype.h>
#include <sstmac/software/libraries/mpi/mpi_types/mpiop.h>
#include <sstmac/software/libraries/mpi/mpi_types/mpiid.h>
#include <sstmac/software/libraries/mpi/mpi_types/mpitag.h>
#include <sstmac/software/libraries/mpi/mpi_comm/mpigroup.h>
//#include <sstmac/software/libraries/mpi/mpiapi.h>

#include <boost/intrusive_ptr.hpp>
#include <sstmac/software/process/thread.h>
#include <sstmac/software/libraries/mpi/mpiapp.h>
#include <sstmac/software/libraries/mpi/mpiapi.h>
#include <iostream>

#include <sstmac/common/messages/payload.h>
#include <sstmac/common/messages/valuepayload.h>
#include <sstmac/common/messages/vectorpayload.h>

#include <boost/unordered_map.hpp>

// global symbol isn't found unless this code is in sstmac namespace???
// tried "using namespace sstmac;", symbol not found

using namespace sstmac::sw;

namespace sstmac
{

  bool has_been_initialized = false;

  static bool payloads_ = true;

  static boost::unordered_map<long long, mpicomm> comm_map;
  static std::map<mpicomm, long long> map_comm;
  static long long comm_counter;

  static boost::unordered_map<long long, mpigroup> group_map;
  static long long group_counter;

  static boost::unordered_map<long long, mpirequest_t> request_map;
  static long long req_counter;

  static debug dbg_;

  extern "C" void
  MPI_Disable_Payloads()
  {
    payloads_ = false;
  }
  extern "C" void
  MPI_Enable_Payloads()
  {
    payloads_ = true;
  }

  /*
   * ---------------------------------------------------------------------------------
   * ------------------ Validity checkers -----------------------
   * ---------------------------------------------------------------------------------
   */

  void
  check_comm(MPI_Comm c, std::string here)
  {
    if (comm_map.find(c) == comm_map.end())
      {
        throw ssterror(
            "sstmac_mpi::check_comm() - could not find MPI_Comm in comm_map: "
                + here);
      }
  }

  void
  check_request(MPI_Request c, std::string here)
  {
    if (request_map.find(c) == request_map.end())
      {
        throw ssterror(
            "sstmac_mpi::check_request() - could not find MPI_Request in request_map: "
                + here);
      }
  }

  void
  check_group(MPI_Group c, std::string here)
  {
    if (group_map.find(c) == group_map.end())
      {
        throw ssterror(
            "sstmac_mpi::check_group() - could not find MPI_Group in group_map: "
                + here);
      }
  }

  /*
   * ---------------------------------------------------------------------------------
   * ------------------ Utility functions to convert to sstmac -----------------------
   * ---------------------------------------------------------------------------------
   */
  mpiapi::ptr
  get_current_mpi()
  {
    thread::ptr t = OS_API::get_current_thread();
    mpiapp::ptr app;
    try
      {
        app = boost::dynamic_pointer_cast<mpiapp>(t);
      }
    catch (std::exception& e)
      {
        std::cerr << e.what() << std::endl;
        return mpiapi::ptr();
      }

    if (!app)
      {
        throw sstmac::ssterror(
            "sstcpp::get_current_mpi - app didn't cast, make sure app is an mpiapp");
      }

    return app->mpi();
  }

  mpitype
  get_mpi_type(int t)
  {
    switch (t)
      {
    case MPI_CHAR:
      return mpitype::mpi_char;
    case MPI_BYTE:
      return mpitype::mpi_byte;
    case MPI_DOUBLE:
      return mpitype::mpi_double;
    case MPI_INT:
      return mpitype::mpi_int;
    case MPI_LONG:
      return mpitype::mpi_long;

    default:
      throw sstmac::ssterror("sstmac_mpi::get_mpi_type() - unknown type");
      }
  }

  int
  get_mpi_datatype(mpitype t)
  {
    if (t == mpitype::mpi_char)
      return MPI_CHAR;
    else if (t == mpitype::mpi_byte)
      return MPI_BYTE;
    else if (t == mpitype::mpi_int)
      return MPI_INT;
    else if (t == mpitype::mpi_double)
      return MPI_DOUBLE;
    else if (t == mpitype::mpi_long)
      return MPI_LONG;
    else
      {
        throw sstmac::ssterror("sstmac_mpi::get_mpi_type() - unknown type");
      }
  }

  mpitag
  get_mpi_tag(int t)
  {
    if (t == MPI_ANY_TAG)
      {
        return mpitag::any_tag;
      }
    else
      {
        return mpitag(t);
      }
  }

  mpiop
  get_mpi_op(int o)
  {
    switch (o)
      {
    case MPI_MIN:
      return mpiop::min;
    case MPI_MAX:
      return mpiop::max;
    case MPI_SUM:
      return mpiop::sum;
    default:
      throw sstmac::ssterror("sstmac_mpi::get_mpi_op() - unknown operation");
      }
  }

  // ---------- These functions were added to support SST/macro payloads,
  // ----------   should be called before and after mpi calls that use real payloads

  template<typename Type>
    payload::const_ptr
    construct_payload(const void* buf, int count)
    {
      Type* casted = (Type*) buf;
      if (count > 1)
        {
          return vector1payload<Type>::construct(casted, count);
        }
      else
        {
          return valuepayload<Type>::construct(*casted);
        }
    }

  payload::const_ptr
  get_mpi_payload(const void* buf, int count, MPI_Datatype type)
  {
    if (buf == NULL || !payloads_)
      {
        return payload::null();
      }
    else
      {
        switch (type)
          {
        case MPI_CHAR:
        case MPI_BYTE:
          {
            return construct_payload<char> (buf, count);
          }

        case MPI_INT:
          {
            return construct_payload<int> (buf, count);
          }

        case MPI_DOUBLE:
          return construct_payload<double> (buf, count);

        case MPI_LONG:
          return construct_payload<long> (buf, count);

        default:
          std::cout << "type: " << type << "\n";
          throw sstmac::ssterror("sstmac_mpi::get_mpi_payload() - unknown type");
          }
      }
  }

  std::vector<payload::const_ptr>
  get_mpi_vector_payload(const void* buf, int count, MPI_Datatype type)
  {
    std::vector<payload::const_ptr> ret;
    if (buf == NULL || !payloads_)
      {
        return ret;
      }
    else
      {

        switch (type)
          {
        case MPI_CHAR:
        case MPI_BYTE:
          {
            char* casted = (char*) buf;
            for (int i = 0; i < count; i++)
              {
                ret.push_back(construct_payload<char> (casted + i, count));
              }
            break;
          }

        case MPI_INT:
          {
            int* casted = (int*) buf;
            for (int i = 0; i < count; i++)
              {
                ret.push_back(construct_payload<int> (casted + i, count));
              }
            break;
          }

        case MPI_DOUBLE:
          {
            double* casted = (double*) buf;
            for (int i = 0; i < count; i++)
              {
                ret.push_back(construct_payload<double> (casted + i, count));
              }
            break;
          }

        case MPI_LONG:
          {
            long* casted = (long*) buf;
            for (int i = 0; i < count; i++)
              {
                ret.push_back(construct_payload<long> (casted + i, count));
              }
            break;
          }

        default:
          std::cout << "type: " << type << "\n";
          throw sstmac::ssterror(
              "sstmac_mpi::get_mpi_vector_payload() - unknown type");
          }

        return ret;
      }
  }

  template<typename Type>
    void
    set_value_payload(void* buf, int count,
        std::vector<payload::const_ptr> load)
    {

      Type* casted = (Type*) buf;

      for (int i = 0; i < load.size(); i++)
        {
          if (load[i] != payload::null())
            {
              typename valuepayload<Type>::const_ptr loadcast =
                  boost::dynamic_pointer_cast<const valuepayload<Type> >(
                      load[i]);
              casted[i] = loadcast->data();
            }
        }
    }

  template<typename Type>
    void
    set_value_payload(void* buf, int count, payload::const_ptr load)
    {
      Type* casted = (Type*) buf;
      if (count > 1)
        {
          typename vector1payload<Type>::const_ptr loadcast =
              boost::dynamic_pointer_cast<const vector1payload<Type> >(load);
          for (int i = 0; i < count; i++)
            {
              std::cout << "setting value: " << loadcast->data()[i] << "\n";
              *casted = loadcast->data()[i];
            }
        }
      else
        {

          typename valuepayload<Type>::const_ptr loadcast =
              boost::dynamic_pointer_cast<const valuepayload<Type> >(load);
          *casted = loadcast->data();

        }
    }
  void
  set_mpi_payload(void* buf, int count, MPI_Datatype type,
      payload::const_ptr load)
  {
    if (buf == NULL || load == payload::null() || !payloads_)
      {
        return;
      }
    else
      {
        switch (type)
          {
        case MPI_CHAR:
        case MPI_BYTE:
          {
            set_value_payload<char> (buf, count, load);
            break;
          }

        case MPI_INT:
          {
            set_value_payload<int> (buf, count, load);
            break;
          }

        case MPI_DOUBLE:
          {
            set_value_payload<double> (buf, count, load);
            break;
          }

        case MPI_LONG:
          {
            set_value_payload<long> (buf, count, load);
            break;
          }

        default:
          throw sstmac::ssterror("sstmac_mpi::set_mpi_payload() - unknown type");
          }
      }
  }

  void
  set_mpi_payload(void* buf, int count, MPI_Datatype type,
      std::vector<payload::const_ptr> load)
  {
    if (buf == NULL || !payloads_)
      {
        return;
      }
    else
      {
        if (load.size() != count)
          {
            throw ssterror(
                "sstmac_mpi::set_mpi_payload(vector) - count does not match vector size");
          }
        switch (type)
          {
        case MPI_CHAR:
        case MPI_BYTE:
          {
            set_value_payload<char> (buf, count, load);
            break;
          }

        case MPI_INT:
          set_value_payload<int> (buf, count, load);
          break;

        case MPI_DOUBLE:
          set_value_payload<double> (buf, count, load);
          break;

        case MPI_LONG:
          set_value_payload<long> (buf, count, load);
          break;

        default:
          throw sstmac::ssterror("sstmac_mpi::set_mpi_payload() - unknown type");
          }
      }
  }

  void
  set_mpi_vector_payload(void* buf, int count, MPI_Datatype type,
      std::vector<payload::const_ptr> load)
  {
    if (buf == NULL || !payloads_)
      {
        return;
      }
    else
      {
        if (load.size() != count)
          {
            throw ssterror(
                "sstmac_mpi::set_mpi_payload(vector) - count does not match vector size");
          }
        switch (type)
          {
        case MPI_CHAR:
        case MPI_BYTE:
          {
            char* casted = (char*) buf;
            for (int i = 0; i < load.size(); i++)
              {
                set_value_payload<char> (casted + i, count, load[i]);
              }
            break;
          }

        case MPI_INT:
          {
            int* casted = (int*) buf;
            for (int i = 0; i < load.size(); i++)
              {
                set_value_payload<int> (casted + i, count, load[i]);
              }
            break;
          }

        case MPI_LONG:
          {
            long* casted = (long*) buf;
            for (int i = 0; i < load.size(); i++)
              {
                set_value_payload<long> (casted + i, count, load[i]);
              }
            break;
          }

        case MPI_DOUBLE:
          {
            double* casted = (double*) buf;
            for (int i = 0; i < load.size(); i++)
              {
                set_value_payload<double> (casted + i, count, load[i]);
              }
            break;
          }

        default:
          throw sstmac::ssterror("sstmac_mpi::set_mpi_payload() - unknown type");
          }
      }
  }

  /*
   * ---------------------------------------------------------------------------------
   * ------------------ The real MPI functions ---------------------------------------
   * ---------------------------------------------------------------------------------
   */
  extern "C" int
  MPI_Init(int *argc, char ***argv)
  {
    mpiapi::ptr app = get_current_mpi();

    if (!dbg_.is_real())
      {
        dbg_ = sstmac::debug("<debug> sstmac_mpi", sstmac::debug::magenta);
        sstmac::logger::register_logger(&dbg_);
      }
    SSTMAC_DEBUG << "MPI_Init \n";
    app->init();

    if (!has_been_initialized)
      {
        has_been_initialized = true;
        req_counter = 0;
        group_counter = 0;
        comm_counter = 0;
      }
    mpicomm world = app->comm_world();
    comm_map[comm_counter] = world;
    map_comm[world] = comm_counter;
    comm_counter++;

    return MPI_SUCCESS; //yes, error handling would be nice
  }

  extern "C" int
  MPI_Finalize()
  {
    SSTMAC_DEBUG << "MPI_Finalize \n";
    mpiapi::ptr app = get_current_mpi();
    app->finalize();
    return MPI_SUCCESS; //yes, error handling would be nice
  }

  extern "C" int
  MPI_Wait(MPI_Request *request, struct MPI_Status *status)
  {
    static std::string here = "MPI_Wait";
    SSTMAC_DEBUG << here << "\n";
    check_request(*request, here);
    mpiapi::ptr app = get_current_mpi();
    mpirequest_t req = request_map[*request];
    const_mpistatus_t stat;
    app->wait(req, stat);
    status->MPI_SOURCE = stat->source().id;
    status->MPI_TAG = stat->tag().id;
    status->MPI_ERROR = 0;
    return MPI_SUCCESS; //yes, error handling would be nice
  }

  extern "C" int
  MPI_Waitany(int count, MPI_Request array_of_requests[], int *index,
      struct MPI_Status *status)
  {
    static std::string here = "MPI_Waitany";
    SSTMAC_DEBUG << here << "\n";
    mpiapi::ptr mpi = get_current_mpi();
    const_mpistatus_t stat;
    std::vector<mpirequest_t> reqs;
    for (int i = 0; i < count; i++)
      {
        check_request(array_of_requests[i], here);
        reqs.push_back(request_map[array_of_requests[i]]);
      }
    int ind;
    mpi->waitany(reqs, ind, stat);
    *index = ind;
    status->MPI_SOURCE = stat->source().id;
    status->MPI_TAG = stat->tag().id;
    status->MPI_ERROR = 0;
    return MPI_SUCCESS; //yes, error handling would be nice
  }
  extern "C" int
  MPI_Waitsome(int incount, MPI_Request array_of_requests[], int *outcount,
      int array_of_indices[], struct MPI_Status array_of_statuses[])
  {
    static std::string here = "MPI_Waitsome";
    SSTMAC_DEBUG << here << "\n";
    mpiapi::ptr mpi = get_current_mpi();
    const_mpistatus_t stat;
    std::vector<mpirequest_t> reqs;
    for (int i = 0; i < incount; i++)
      {
        check_request(array_of_requests[i], here);
        reqs.push_back(request_map[array_of_requests[i]]);
      }

    std::vector<int> ind;
    std::vector<const_mpistatus_t> stats;
    mpi->waitsome(reqs, ind, stats);
    int tempind[ind.size()]; // = (int*) malloc(ind.size() * sizeof(int));
    MPI_Status tmpstat[stats.size()]; // = (Status*) malloc(stats.size() * sizeof(Status));

    *outcount = stats.size();
    for (int i = 0; i < ind.size(); i++)
      {
        tempind[i] = ind[i];
        tmpstat[i].MPI_SOURCE = stats[i]->source().id;
        tmpstat[i].MPI_TAG = stats[i]->tag().id;
        tmpstat[i].MPI_ERROR = 0;
      }

    array_of_statuses = tmpstat;
    array_of_indices = tempind;

    return MPI_SUCCESS; //yes, error handling would be nice
  }

  extern "C" int
  MPI_Get_count(struct MPI_Status *status, MPI_Datatype datatype, int *count)
  {
    static std::string here = "MPI_Get_count";
    SSTMAC_DEBUG << here << "\n";
    throw sstmac::ssterror("MPI_Get_count is unimplemented");
    return MPI_SUCCESS; //yes, error handling would be nice
  }

  extern "C" int
  MPI_Test(MPI_Request *request, int *flag, struct MPI_Status *status)
  {
    static std::string here = "MPI_Test";
    SSTMAC_DEBUG << here << "\n";
    check_request(*request, here);
    mpiapi::ptr mpi = get_current_mpi();
    bool f;
    mpirequest_t req = request_map[*request];
    const_mpistatus_t stat;
    mpi->test(req, f, stat);
    status->MPI_SOURCE = stat->source().id;
    status->MPI_TAG = stat->tag().id;
    status->MPI_ERROR = 0;
    *flag = f;
    return MPI_SUCCESS; //yes, error handling would be nice
  }

  extern "C" int
  MPI_Cancel(MPI_Request *request)
  {
    static std::string here = "MPI_Cancel";
    SSTMAC_DEBUG << here << "\n";
    check_request(*request, here);
    mpiapi::ptr mpi = get_current_mpi();
    mpirequest_t req = request_map[*request];
    mpi->cancel(req);
    request_map.erase(*request);
    return MPI_SUCCESS; //yes, error handling would be nice
  }

  extern "C" int
  MPI_Request_free(MPI_Request *request)
  {
    static std::string here = "MPI_Request_free";
    SSTMAC_DEBUG << here << "\n";
    check_request(*request, here);
    request_map.erase(*request);
    return MPI_SUCCESS;
  }

  extern "C" int
  MPI_Comm_dup(MPI_Comm comm, MPI_Comm *newcomm)
  {
    static std::string here = "MPI_Comm_dup";
    SSTMAC_DEBUG << here << "\n";
    check_comm(comm, here);
    mpiapi::ptr mpi = get_current_mpi();
    mpicomm ret;
    mpi->comm_dup(comm_map[comm], ret);
    *newcomm = comm_counter;
    comm_map[comm_counter] = ret;
    map_comm[ret] = comm_counter;
    comm_counter++;
    return MPI_SUCCESS;
  }

  extern "C" int
  MPI_Comm_rank(MPI_Comm comm, int *rank)
  {
    static std::string here = "MPI_Comm_rank";
    SSTMAC_DEBUG << here << "\n";
    check_comm(comm, here);
    *rank = comm_map[comm].rank().id;
    return MPI_SUCCESS;
  }

  extern "C" int
  MPI_Comm_split(MPI_Comm comm, int color, int key, MPI_Comm *newcomm)
  {
    static std::string here = "MPI_Comm_split";
    check_comm(comm, here);
    mpiapi::ptr mpi = get_current_mpi();
    SSTMAC_DEBUG << "[rank=" << mpi->comm_rank(comm_map[comm]).id << "] "
          << here << "(" << comm << ", " << color << ", " << key << ") \n";

    mpicomm ret;
    mpi->comm_split(comm_map[comm], color, key, ret);
    *newcomm = comm_counter;
    comm_map[comm_counter] = ret;
    map_comm[ret] = comm_counter;
    comm_counter++;
    return MPI_SUCCESS;
  }

  extern "C" int
  MPI_Comm_size(MPI_Comm comm, int *size)
  {
    static std::string here = "MPI_Comm_size";
    SSTMAC_DEBUG << here << "\n";
    check_comm(comm, here);
    *size = comm_map[comm].size().id;
    return MPI_SUCCESS;
  }

  extern "C" int
  MPI_Comm_free(MPI_Comm *comm)
  {
    static std::string here = "MPI_Comm_free";
    SSTMAC_DEBUG << here << "\n";
    check_comm(*comm, here);
    mpiapi::ptr mpi = get_current_mpi();
    mpi->comm_free(comm_map[*comm]);
    map_comm.erase(comm_map[*comm]);
    comm_map.erase(*comm);
    return MPI_SUCCESS;
  }

  extern "C" int
  MPI_Comm_create(MPI_Comm comm, MPI_Group group, MPI_Comm *newcomm)
  {
    static std::string here = "MPI_Comm_create";
    SSTMAC_DEBUG << here << "\n";
    check_comm(comm, here);
    mpiapi::ptr mpi = get_current_mpi();
    mpicomm ret;
    mpi->comm_create(comm_map[comm], group_map[group].members_, ret);
    *newcomm = comm_counter;
    comm_map[comm_counter] = ret;
    map_comm[ret] = comm_counter;
    comm_counter++;
    return MPI_SUCCESS;
  }

  extern "C" int
  MPI_Comm_group(MPI_Comm comm, MPI_Group *group)
  {
    static std::string here = "MPI_Comm_group";
    SSTMAC_DEBUG << here << "\n";
    check_comm(comm, here);
    mpiapi::ptr mpi = get_current_mpi();
    mpicomm c = comm_map[comm];
    mpigroup ret;
    for (long i = 0; i < c.size().id; i++)
      {
        ret.members_.push_back(mpiid(i));
      }
    group_map[group_counter] = ret;
    *group = group_counter;
    group_counter++;
    return MPI_SUCCESS;
  }

  extern "C" int
  MPI_Group_incl(MPI_Group group, int n, int *ranks, MPI_Group *newgroup)
  {
    static std::string here = "MPI_Group_incl";
    SSTMAC_DEBUG << here << "\n";
    check_group(group, here);
    mpigroup gr = group_map[group];
    mpigroup ret;
    for (int i = 0; i < n; i++)
      {
        ret.members_.push_back(gr.members_[ranks[i]]);
      }
    group_map[group_counter] = ret;
    *newgroup = group_counter;
    group_counter++;
    return MPI_SUCCESS;
  }
  extern "C" int
  MPI_Group_free(MPI_Group *group)
  {
    static std::string here = "MPI_Group_free";
    SSTMAC_DEBUG << here << "\n";
    check_group(*group, here);
    group_map.erase(*group);
    return MPI_SUCCESS;
  }

  extern "C" int
  MPI_Send(void *buf, int count, MPI_Datatype datatype, int dest, int tag,
      MPI_Comm comm)
  {
    static std::string here = "MPI_Send";
    SSTMAC_DEBUG << here << "-" << (long) buf << ", " << count << ", "
          << datatype << ", " << dest << ", " << tag << ", " << comm << "\n";
    check_comm(comm, here);
    mpiapi::ptr mpi = get_current_mpi();

    sstmac::payload::const_ptr load = get_mpi_payload(buf, count, datatype);

    mpi->send(count, get_mpi_type(datatype), mpiid(dest), mpitag(tag),
        comm_map[comm], load);

    return MPI_SUCCESS;
  }

  extern "C" int
  MPI_SendRecv(void *sendbuf, int sendcount, MPI_Datatype sendtype, int dest,
      int sendtag, void *recvbuf, int recvcount, MPI_Datatype recvtype,
      int source, int recvtag, MPI_Comm comm, struct MPI_Status* stat)
  {
    static std::string here = "MPI_SendRecv";
    SSTMAC_DEBUG << here << "\n";
    check_comm(comm, here);
    MPI_Request sendreq;

    MPI_Isend(sendbuf, sendcount, sendtype, dest, sendtag, comm, &sendreq);
    MPI_Recv(recvbuf, recvcount, recvtype, source, recvtag, comm, stat);
    MPI_Status st;
    MPI_Wait(&sendreq, &st);

    return MPI_SUCCESS;
  }

  extern "C" int
  MPI_Isend(const void *buf, int count, MPI_Datatype datatype, int dest,
      int tag, MPI_Comm comm, MPI_Request *request)
  {
    static std::string here = "MPI_Isend";
    SSTMAC_DEBUG << here << "\n";
    check_comm(comm, here);
    mpiapi::ptr mpi = get_current_mpi();
    mpirequest_t req;
    bool hasload = (buf != NULL);
    sstmac::payload::const_ptr load = get_mpi_payload(buf, count, datatype);

    mpi->isend(count, get_mpi_type(datatype), mpiid(dest), mpitag(tag),
        comm_map[comm], req, load);

    request_map[req_counter] = req;
    *request = req_counter;
    req_counter++;

    return MPI_SUCCESS;
  }

  extern "C" int
  MPI_Recv(void *buf, int count, MPI_Datatype datatype, int source, int tag,
      MPI_Comm comm, struct MPI_Status *status)
  {
    static std::string here = "MPI_Recv";
    SSTMAC_DEBUG << here << "-" << (long) buf << ", " << count << ", "
          << datatype << ", " << source << ", " << tag << ", " << comm << "\n";
    check_comm(comm, here);
    mpiapi::ptr mpi = get_current_mpi();
    const_mpistatus_t stat;
    mpi->recv(count, get_mpi_type(datatype), mpiid(source), mpitag(tag),
        comm_map[comm], stat);
    set_mpi_payload(buf, count, datatype, stat->content());
    status->MPI_SOURCE = stat->source().id;
    status->MPI_TAG = stat->tag().id;
    status->MPI_ERROR = 0;

    return MPI_SUCCESS;
  }

  extern "C" int
  MPI_Irecv(void *buf, int count, MPI_Datatype datatype, int source, int tag,
      MPI_Comm comm, MPI_Request *request)
  {
    static std::string here = "MPI_Irecv";
    SSTMAC_DEBUG << here << "\n";
    check_comm(comm, here);

    if (buf != NULL)
      {
        throw ssterror(
            "MPI_Irecv: payload buffer does not work with Irecv, you must pass NULL and call MPI_Recover_payload(request)");
      }
    mpiapi::ptr mpi = get_current_mpi();
    mpirequest_t req;
    mpi->irecv(count, get_mpi_type(datatype), mpiid(source), mpitag(tag),
        comm_map[comm], req);
    request_map[req_counter] = req;
    *request = req_counter;
    req_counter++;

    return MPI_SUCCESS;
  }

  int
  MPI_Recover_Payload(void *buf, MPI_Request *request)
  {
    static std::string here = "MPI_Recover_Payload";
    SSTMAC_DEBUG << here << "\n";
    check_request(*request, here);
    mpirequest_t req = request_map[*request];
    if (!req->is_complete())
      {
        return 1;
      }
    else
      {
        const_mpistatus_t stat = req->get_status();
        set_mpi_payload(buf, stat->get_count(),
            get_mpi_datatype(stat->send_type()), stat->content());
        return 0;
      }
  }

  extern "C" int
  MPI_Allreduce(const void *sendbuf, void *recvbuf, int count,
      MPI_Datatype datatype, MPI_Op op, MPI_Comm comm)
  {
    static std::string here = "MPI_Allreduce";
    SSTMAC_DEBUG << here << "\n";
    check_comm(comm, here);
    mpiapi::ptr mpi = get_current_mpi();
    sstmac::payload::const_ptr load = get_mpi_payload(sendbuf, count, datatype);
    SSTMAC_DEBUG << "rank " << mpi->comm_world().rank().id << ": using comm "
          << comm << ", which has size " << comm_map[comm].size().id << "\n";
    if (load)
      SSTMAC_DEBUG << "rank " << mpi->comm_world().rank().id << ": reducing "
            << load->to_string() << "\n";
    sstmac::payload::const_ptr res;
    mpi->allreduce(count, get_mpi_type(datatype), get_mpi_op(op),
        comm_map[comm], load, res);
    if (res)
      SSTMAC_DEBUG << "rank " << mpi->comm_world().rank().id
            << ": ended up with " << res->to_string() << "\n";
    set_mpi_payload(recvbuf, count, datatype, res);
    return MPI_SUCCESS;
  }

  extern "C" int
  MPI_Reduce(const void *sendbuf, void *recvbuf, int count,
      MPI_Datatype datatype, MPI_Op op, int root, MPI_Comm comm)
  {
    static std::string here = "MPI_Reduce";
    SSTMAC_DEBUG << here << "\n";
    check_comm(comm, here);
    mpiapi::ptr mpi = get_current_mpi();
    sstmac::payload::const_ptr load = get_mpi_payload(sendbuf, count, datatype);
    sstmac::payload::const_ptr res;
    mpi->reduce(count, get_mpi_type(datatype), get_mpi_op(op), mpiid(root),
        comm_map[comm], load, res);
    set_mpi_payload(recvbuf, count, datatype, res);
    return MPI_SUCCESS;
  }

  extern "C" int
  MPI_Barrier(MPI_Comm comm)
  {
    static std::string here = "MPI_Barrier";
    SSTMAC_DEBUG << here << "\n";
    check_comm(comm, here);
    mpiapi::ptr mpi = get_current_mpi();
    mpi->barrier(comm_map[comm]);
    return MPI_SUCCESS;
  }

  extern "C" int
  MPI_Bcast(void *buffer, int count, MPI_Datatype datatype, int root,
      MPI_Comm comm)
  {
    static std::string here = "MPI_Bcast";
    SSTMAC_DEBUG << here << "\n";
    check_comm(comm, here);
    mpiapi::ptr mpi = get_current_mpi();

    sstmac::payload::const_ptr load = get_mpi_payload(buffer, count, datatype);

    mpi->bcast(count, get_mpi_type(datatype), mpiid(root), comm_map[comm], load);

    set_mpi_payload(buffer, count, datatype, load);

    return MPI_SUCCESS;
  }

  extern "C" int
  MPI_Gather(void *sendbuf, int sendcnt, MPI_Datatype sendtype, void *recvbuf,
      int recvcnt, MPI_Datatype recvtype, int root, MPI_Comm comm)
  {
    static std::string here = "MPI_Gather";
    SSTMAC_DEBUG << here << "\n";
    check_comm(comm, here);
    sstmac::sw::mpiapi::ptr mpi = get_current_mpi();
    sstmac::payload::const_ptr load = get_mpi_payload(sendbuf, sendcnt,
        sendtype);
    std::vector<sstmac::payload::const_ptr> cc;
    mpi->gather(sendcnt, get_mpi_type(sendtype), recvcnt,
        get_mpi_type(recvtype), sstmac::sw::mpiid(root), comm_map[comm], load,
        cc);
    set_mpi_payload(recvbuf, recvcnt, recvtype, cc);
    return MPI_SUCCESS;
  }

  extern "C" int
  MPI_Allgather(void *sendbuf, int sendcount, MPI_Datatype sendtype,
      void *recvbuf, int recvcount, MPI_Datatype recvtype, MPI_Comm comm)
  {
    static std::string here = "MPI_Allgather";
    SSTMAC_DEBUG << here << "\n";
    check_comm(comm, here);
    sstmac::sw::mpiapi::ptr mpi = get_current_mpi();
    sstmac::payload::const_ptr load = get_mpi_payload(sendbuf, sendcount,
        sendtype);
    std::vector<sstmac::payload::const_ptr> cc;
    mpi->allgather(sendcount, get_mpi_type(sendtype), recvcount,
        get_mpi_type(recvtype), comm_map[comm], load, cc);
    set_mpi_payload(recvbuf, recvcount, recvtype, cc);
    return MPI_SUCCESS;
  }

  extern "C" int
  MPI_Scatter(void *sendbuf, int sendcnt, MPI_Datatype sendtype, void *recvbuf,
      int recvcnt, MPI_Datatype recvtype, int root, MPI_Comm comm)
  {
    static std::string here = "MPI_Scatter";
    SSTMAC_DEBUG << here << "\n";
    check_comm(comm, here);
    sstmac::sw::mpiapi::ptr mpi = get_current_mpi();
    sstmac::payload::const_ptr load;
    std::vector<sstmac::payload::const_ptr> cc = get_mpi_vector_payload(
        sendbuf, sendcnt, sendtype);
    mpi->scatter(sendcnt, get_mpi_type(sendtype), recvcnt,
        get_mpi_type(recvtype), sstmac::sw::mpiid(root), comm_map[comm], cc,
        load);
    set_mpi_payload(recvbuf, recvcnt, recvtype, load);
    return MPI_SUCCESS;
  }

  extern "C" int
  MPI_Alltoall(void *sendbuf, int sendcount, MPI_Datatype sendtype,
      void *recvbuf, int recvcount, MPI_Datatype recvtype, MPI_Comm comm)
  {
    static std::string here = "MPI_Alltoall";
    SSTMAC_DEBUG << here << "\n";
    check_comm(comm, here);
    sstmac::sw::mpiapi::ptr mpi = get_current_mpi();
    std::vector<sstmac::payload::const_ptr> load;
    std::vector<sstmac::payload::const_ptr> cc = get_mpi_vector_payload(
        sendbuf, sendcount, sendtype);
    mpi->alltoall(sendcount, get_mpi_type(sendtype), recvcount,
        get_mpi_type(recvtype), comm_map[comm], cc, load);
    set_mpi_vector_payload(recvbuf, recvcount, recvtype, load);
    return MPI_SUCCESS;
  }

  extern "C" MPI_Comm
  MPI_Comm_World()
  {
    static std::string here = "MPI_Comm_World";
    SSTMAC_DEBUG << here << "\n";
    sstmac::sw::mpiapi::ptr mpi = get_current_mpi();
    mpicomm world = mpi->comm_world();
    return map_comm[world];
  }

  extern "C" double
  MPI_Wtime()
  {
    mpiapi::ptr app = get_current_mpi();

    return app->wtime().sec(); //yes, error handling would be nice
  }

}

