/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_LIBRARIES_WAIT_H_INCLUDED
#define SSTMAC_SOFTWARE_LIBRARIES_WAIT_H_INCLUDED

#include <sstmac/common/ptr_type.h>

namespace sstmac
{
  namespace sw
  {
    class wait : public ptr_type
    {
    public:
      enum WAIT_TYPE
      {
        ANY, THIS, TYPE, NUM_TYPE
      };

      WAIT_TYPE type_;

    protected:
      wait(WAIT_TYPE t) :
        type_(t)
      {

      }
    public:
      sst_message::ptr recent_;
      typedef boost::intrusive_ptr<wait> ptr;


      virtual
      ~wait()
      {
      }

      virtual std::string
      to_string() const
      {
        return "wait";
      }

      virtual bool
      incoming_message(const sst_message::ptr &msg) = 0;

    };

    class wait_any : public wait
    {

    protected:
      wait_any() :
        wait(ANY)
      {

      }
    public:

      typedef boost::intrusive_ptr<wait_any> ptr;
      static ptr
      construct()
      {
        return ptr(new wait_any());
      }

      virtual
      ~wait_any()
      {
      }

      virtual bool
      incoming_message(const sst_message::ptr &msg)
      {
        recent_ = msg;
        return true;
      }
    };

    class wait_this : public wait
    {

    protected:
      sst_message::ptr saved_;

      wait_this(const sst_message::ptr &msg) :
        wait(THIS), saved_(msg)
      {

      }
    public:

      typedef boost::intrusive_ptr<wait_this> ptr;
      static ptr
      construct(const sst_message::ptr &msg)
      {
        return ptr(new wait_this(msg));
      }

      virtual
      ~wait_this()
      {
      }

      virtual bool
      incoming_message(const sst_message::ptr &msg)
      {
        recent_ = msg;
        return (msg == saved_);
      }
    };

    class wait_type : public wait
    {

    protected:
      sst_message::MESSAGE_TYPES type_;

      wait_type(sst_message::MESSAGE_TYPES t) :
        wait(TYPE), type_(t)
      {

      }
    public:

      typedef boost::intrusive_ptr<wait_type> ptr;
      static ptr
      construct(sst_message::MESSAGE_TYPES t)
      {
        return ptr(new wait_type(t));
      }

      virtual
      ~wait_type()
      {
      }

      virtual bool
      incoming_message(const sst_message::ptr &msg)
      {
        recent_ = msg;
        if (msg->get_type() == type_)
          {
            return true;
          }

      }
    };

    class wait_num_type : public wait
    {

    protected:
      sst_message::MESSAGE_TYPES type_;
      long num_;
      long seen_;

      wait_num_type(sst_message::MESSAGE_TYPES t, long num) :
        wait(NUM_TYPE), type_(t), num_(num), seen_(0)
      {

      }
    public:

      typedef boost::intrusive_ptr<wait_num_type> ptr;
      static ptr
      construct(sst_message::MESSAGE_TYPES t, long num)
      {
        return ptr(new wait_num_type(t, num));
      }

      virtual
      ~wait_num_type()
      {
      }

      virtual bool
      incoming_message(const sst_message::ptr &msg)
      {
        recent_ = msg;
        if (msg->get_type() == type_)
          {
            seen_++;
          }
        return (seen_ == num_);
      }
    };

  }
}
#endif
