/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_PROCESS_THREAD_H_INCLUDED
#define SSTMAC_SOFTWARE_PROCESS_THREAD_H_INCLUDED

#include <sstmac/software/process/softwareid.h>
//#include <sstmac/software/threading/context.h>
#include <sstmac/common/nocopy.h>
#include <sstmac/common/debug.h>
#include <sstmac/common/nodeaddress.h>
#include <sstmac/common/ptr_type.h>

#include <unistd.h>
#include <list>
#include <utility>

namespace sstmac { namespace sw
{

  // operatingsystem and key need to include this header to get thread_data_t
  // Forward declare these rather than including them to avoid circular
  // dependency issues.
  class threading_interface;
  class key;
  class OS_API;
  class library;

  // We need to know the thread associated with each threading_interface
  // so current thread pointer can be maintained
  // (used for threadstack_ in operatingsystem and block_ in key )
  class thread;
  typedef std::pair< threading_interface*, boost::intrusive_ptr<thread> >
    thread_data_t;
  
  /**
   * Base class for tasks that can be cooperative scheduled using threadcontext.
   */


  struct threadinfo {
    boost::intrusive_ptr<thread> thethread;
  };

  class thread : virtual public nocopy, public virtual ptr_type
  {
  public:
    typedef boost::intrusive_ptr<thread> ptr;

    friend class operatingsystem;

    /// Help resolve deadlock situations.
    enum state
    {
      PENDING, ACTIVE, SUSPENDED, BLOCKED, DONE
    };

    virtual std::string to_string() const{
      return "thread";
    }

  protected:

    /// Monitor state for deadlock detection.
    state state_;

    /// Each thread can only run under one OS/scheduler.
    boost::intrusive_ptr<OS_API> os_;

  private:

    bool isInit;
    /// This object.
    // boost::weak_ptr<thread> self_;
    /// The stack given to this thread.
    void *stack_;
    /// The stacksize.
    size_t stacksize_;

    static long THREAD_ID_CNT;
    long threadid_;


    /// The magic.
    //threadcontext_t *context_;
    threading_interface* context_;

    /// This key gets used when other threads ask to wait (join) this thread.
    boost::intrusive_ptr<key> key_;

    std::list< boost::intrusive_ptr<library> > pending_libs_;
      
      std::list<boost::intrusive_ptr<library> > libs_;

  private:

    void
    destroy();

    /// Run routine that defines the initial context for this task.
    /// This routine calls the virtual thread::run method.
    static void
    run_routine(void* threadptr);

  protected:
    /// Initialize private data.
    thread();

    void register_lib(const boost::intrusive_ptr<library> &lib);
    void unregister_lib(const boost::intrusive_ptr<library> &lib);
      void unregister_all_libs();

    timestamp now();

    nodeaddress::ptr get_physical_address();



  public:
    virtual
    ~thread();

    /// Get current thread state.
    state
    get_state() const
    {
      return state_;
    }

    bool
    is_started()
    {
      return (state_ != PENDING);
    }

    long
    get_id() const {
      return threadid_;
    }

    virtual void
    kill()
    {

      state_ = DONE;
      destroy();
    }

    /// method that gets called by the operating system.
    void
    init_thread(threading_interface* tocopy,
                void *stack, int stacksize,
                boost::intrusive_ptr<OS_API> os,
                threading_interface *yield_to);

    /// Derived types need to override this method.
    virtual void
    run() = 0;

    /// A convenience request to start a new thread.
    /// The current thread has to be initialized for this to work.
    void
    start_thread(const thread::ptr& thr);

    void
    join();

    boost::intrusive_ptr<key> get_join_key();

    /// Test whether the current task has been initialized (activated)
    /// by a scheduler.
    bool initialized() const;

    /// Access the context for this thread (handle with care).
    /* inline threadcontext_t*
     context()
     {
     return &context_;
     }*/

  };

} } // end of namespace sstmac
#endif
