/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <math.h>

#include <sstmac/software/services/launch/instantlaunch.h>
#include <sstmac/software/ami/environment.h>
//#include <sstmac/simulator/nodeinterface.h>
//#include <sstmac/backends/native/components/node/node.h>

//#include <sstmac/backends/native/launch/instantlaunch-completion.h>
#include <sstmac/software/process/taskid.h>

#include <sstmac/common/errors.h>
#include <sstmac/common/debug.h>

#include <sstmac/software/services/launch/launch_message.h>

#include <sstmac/software/services/launch/machine_descriptor.h>
#include <sstmac/software/skeletons/skeletonfactory.h>

#include <sstmac/software/libraries/checkpoint/restartable_app.h>

#ifdef SSTMAC_HAVE_BOOST_1_37
#include <boost/unordered_map.hpp>
#else
#include <map>
#endif

namespace sstmac
{
  namespace sw
  {

#define RESTART_TIME 10

    std::multimap<nodeaddress::ptr, taskid, nodeaddress::nodeaddresscomp>
        instantlaunch::taskmap_;
    std::map<appid, environment::const_ptr> instantlaunch::envmap_;
    std::map<int, nodeaddress::ptr> instantlaunch::nodemap_;

    // Pointer types.

    //
    // Protected constructor.
    //
    instantlaunch::instantlaunch(std::list<launchinfo::ptr> &tolaunch,
        allocationstrategy::ptr &a, indexstrategy::ptr &ind) :
      tolaunch_(tolaunch), alloc_(a), indexing_(ind), block_(key::construct())
    {

    }

    service::ptr
    instantlaunch::clone()
    {
      return construct(tolaunch_, alloc_, indexing_);
    }

    //
    // Public constructor
    //
    instantlaunch::ptr
    instantlaunch::construct(std::list<launchinfo::ptr> &tolaunch,
        allocationstrategy::ptr &a, indexstrategy::ptr &ind)
    {
      instantlaunch::ptr retval(new instantlaunch(tolaunch, a, ind));

      return retval;
    }

    void
    instantlaunch::handle(const sst_message::ptr &msg)
    {

      if (msg->get_type() == sst_message::SEND_ACK)
        {
          return;
        }

      launch_message::ptr lmsg = boost::dynamic_pointer_cast<launch_message>(
          msg);

      if (!lmsg)
        {
          std::cout << "msg: " << msg->to_string() << "\n";
          std::cout << "from: " << msg->fromaddr_->to_string() << "\n";
          std::cout << "to: " << msg->toaddr_->to_string() << "\n";
          throw valueerror(
              "instantlaunch[" + os_->get_my_addr()->to_string()
                  + "]: wanted a launch_message, but got something else");
        }

      launchinfo::ptr linfo = lmsg->get_info();

      if (lmsg->is_restart_)
        {
          fullos_->restarting(true);
          restart_app(linfo);
          launch_message::ptr sleep = launch_message::construct(linfo);
          sleep->is_restart_node_ = true;
          sleep->set_time(timestamp(RESTART_TIME / 2));

          os_->execute_kernel(AMI::AMI_COMPUTE, AMI::AMI_COMP_SLEEP, sleep);
          return;
        }

      if (lmsg->is_restart_node_)
        {
          fullos_->restarting(false);
          return;
        }

      if (lmsg->is_restart_wait_)
        {

          fullos_->cleanup();

          //restart with checkpoint state
          std::list<app::ptr>::iterator it, end = lmsg->newapps_.end();
          for (it = lmsg->newapps_.begin(); it != end; it++)
            {
              SSTMAC_DEBUG << "restarting app " << (*it)->get_sid().to_string()
                    << " on node " << os_->get_my_addr()->to_string() << "\n";
              fullos_->add_application(*it, linfo->params_);
            }

          states_[linfo->aid_] = LAUNCHED;
          fullos_->start_apps();

          return;
        }

      current_ = lmsg;

      if (envmap_.find(linfo->aid_) == envmap_.end())
        {
          SSTMAC_DEBUG
                << "instantlaunch: initializing allocation for starting at "
                << linfo->start_ << "\n";

          //this app has not been started anywhere yet, need to allocate, index, and build environment
          environment::ptr env = environment::construct();

          taskmap_.clear();

          SSTMAC_DEBUG << "instantlaunch: requesting allocation for "
                << linfo->num_tasks_ << " tasks \n";

          long nnode;
          if (linfo->num_tasks_ < 0)
            {
              nnode = -1;
            }
          else
            {
              nnode = linfo->extra_nodes_ + ceil(double(linfo->num_tasks_) / linfo->ntask_per_node_);
            }
          std::set<nodeaddress::ptr, nodeaddress::nodeaddresscomp> nodes;
          alloc_->allocate(nnode, nodes);

          SSTMAC_DEBUG << "instantlaunch: allocator returned " << nodes.size()
                << " nodes \n";

          std::vector<nodeaddress::ptr> indexed;
          indexing_->allocate(nodes, linfo->ntask_per_node_, indexed,
              linfo->num_tasks_);

          long i = 0;
          std::vector<nodeaddress::ptr>::iterator it, end = indexed.end();
          for (it = indexed.begin(); it != end && i < linfo->num_tasks_; it++)
            {

              sstmac::sw::taskid id(i);

              SSTMAC_DEBUG << "taskid: " << id << "\n";

              SSTMAC_DEBUG << "adding task " << id.to_string() << " to node "
                    << (*it)->to_string() << "\n";

              env->add_item(id, *it);

              taskmap_.insert(std::pair<nodeaddress::ptr, taskid>(*it, id));
              nodemap_[id.id] = *it;

              i++;
            }

          if(i < nodes.size()){
              SSTMAC_DEBUG << "instantlaunch: there were " << (nodes.size() - i) << " extra nodes allocated \n";
              std::set<nodeaddress::ptr, nodeaddress::nodeaddresscomp>::iterator iter = nodes.begin();
              for(long j = 0; j < i; j++){
                  iter++;
              }


              for(; iter != nodes.end(); iter++){
                  env->add_extra(*iter);
              }
          }

          envmap_[linfo->aid_] = env;

        }

      fullos_->set_environment(envmap_[linfo->aid_], linfo->aid_);

      nodeaddress::ptr myaddr = os_->get_my_addr();
      SSTMAC_DEBUG << "instantlaunch(" << myaddr->to_string()
            << "): considering launching a task...";

      if (taskmap_.find(myaddr) != taskmap_.end())
        {

          std::multimap<nodeaddress::ptr, taskid>::iterator first =
              taskmap_.lower_bound(myaddr), last = taskmap_.upper_bound(myaddr);
          do
            {
              SSTMAC_DEBUG << "launching task " << first->second << "\n";
              softwareid sid(linfo->aid_, first->second);

              app::ptr theapp = linfo->apptype_->clone(sid);
              theapp->init_libs(linfo->apptype_->parameters());
              restartable_app::ptr rest = boost::dynamic_pointer_cast<
                  restartable_app>(theapp);
              if (rest)
                {
                  rest->init_restartable(fullos_, sid, linfo->params_);
                }
              fullos_->add_application(theapp, linfo->params_);
              first++;
            }
          while (first != last);
          fullos_->start_apps();
          states_[linfo->aid_] = LAUNCHED;
        }
      else
        {
          SSTMAC_DEBUG << "nah, don't feel like it \n";
          states_[linfo->aid_] = NOT_RUNNING_HERE;
        }

    }

    void
    instantlaunch::start()
    {
      service::start();

      SSTMAC_DEBUG << "instantlaunch: starting...\n";

      if (!os_)
        {
          throw valueerror("instantlaunch: OS hasn't been registered yet");
        }
      if (!os_->kernel_supported(AMI::AMI_COMPUTE, AMI::AMI_COMP_SLEEP))
        {
          throw unimplementederror(
              "instantlaunch: needs the AMI::AMI_COMP_SLEEP kernel to be able to launch");
        }

      std::list<launchinfo::ptr>::iterator it, end = tolaunch_.end();
      for (it = tolaunch_.begin(); it != end; it++)
        {
          launch_message::ptr launchmsg = launch_message::construct(*it);
          SSTMAC_DEBUG << "app " << (*it)->apptype_->to_string()
                << " is going to launch " << (*it)->num_tasks_ << " at "
                << (*it)->start_ << "\n";

          os_->execute_kernel(AMI::AMI_COMPUTE, AMI::AMI_COMP_SLEEP, launchmsg);
        }

      SSTMAC_DEBUG << "instantlaunch: launches scheduled\n";
      return;
    }

    /**
     *  This function is called when the local node has failed.
     */
    void
    instantlaunch::apps_failed()
    {
      SSTMAC_DEBUG << "instantlaunch: apps_failed() on "
            << os_->get_my_addr()->to_string()
            << " - restarting all apps running here \n";
      std::map<appid, APP_STATE>::iterator iter, end = states_.end();

      fullos_->restarting(true);

      for (iter = states_.begin(); iter != end; iter++)
        {
          if (iter->second == LAUNCHED)
            { //this has been launched

              std::list<app::ptr> osapps = fullos_->get_app_ptrs(iter->first);

              SSTMAC_DEBUG << "instantlaunch: apps_failed() - considering app "
                    << iter->first.id << ", it has " << osapps.size()
                    << " instances running here \n";

              if (osapps.size() > 0)
                { //check to see if any are still running here

                  //get the info pointer, for ease
                  launchinfo::ptr linfo;

                  std::list<launchinfo::ptr>::iterator it, end =
                      tolaunch_.end();
                  for (it = tolaunch_.begin(); it != end; it++)
                    {
                      if ((*it)->aid_ == iter->first)
                        {
                          linfo = *it;
                          break;
                        }
                    }

                  //first, tell all the other nodes (have to get a unique set first)
#ifdef SSTMAC_HAVE_BOOST_1_37
                  typedef boost::unordered_map<nodeaddress::ptr, int> maptype;
#else
                  typedef std::map<nodeaddress::ptr, int> maptype;
#endif
                  maptype nodes;
                  for (size_t i = 0; i < linfo->num_tasks_; i++)
                    {
                      sstmac::sw::taskid tid(i);
                      nodes[nodemap_[tid.id]] = 1;
                    }

                  //don't send it to myself
                  nodes.erase(os_->get_my_addr());

                  maptype::iterator nodeiter, nodeend = nodes.end();
                  for (nodeiter = nodes.begin(); nodeiter != nodeend; nodeiter++)
                    {

                      launch_message::ptr restartmsg =
                          launch_message::construct(linfo);
                      restartmsg->toaddr_ = nodeiter->first;
                      restartmsg->is_restart_ = true;
                      os_->execute_kernel(AMI::AMI_COMM, AMI::AMI_COMM_SEND,
                          restartmsg);
                    }

                  //now restart my own
                  restart_app(linfo);

                }
            }
        }
      launchinfo::ptr linfo = *tolaunch_.begin();
      launch_message::ptr sleep = launch_message::construct(linfo);
      sleep->is_restart_node_ = true;
      sleep->set_time(timestamp(RESTART_TIME / 2));

      os_->execute_kernel(AMI::AMI_COMPUTE, AMI::AMI_COMP_SLEEP, sleep);

    }

    void
    instantlaunch::restart_app(launchinfo::ptr linfo)
    {
      SSTMAC_DEBUG << "instantlaunch[" << os_->get_my_addr()->to_string()
            << "]::restart_app() - restarting app " << linfo->aid_.id << "\n";
      //save the checkpoint states, if possible

      std::list<app::ptr> theapps = fullos_->get_app_ptrs(linfo->aid_);
      std::list<app::ptr> newapps;
      std::list<app::ptr>::iterator it, end = theapps.end();
      for (it = theapps.begin(); it != end; it++)
        {
          restartable_app::ptr rest = boost::dynamic_pointer_cast<
              restartable_app>(*it);
          if (rest)
            {
              SSTMAC_DEBUG << "instantlaunch::restart_app() - app "
                    << (*it)->get_sid().to_string() << " is checkpointable \n";
              lib_checkpoint::ptr check = rest->get_check_lib();
              check->unregister_all_libs();
              fullos_->unregister_lib(check);
              check->restart();
              app::ptr app2 = linfo->apptype_->clone((*it)->get_sid());
              restartable_app::ptr rest2 = boost::dynamic_pointer_cast<
                  restartable_app>(app2);
              rest2->init_restartable(check);
              newapps.push_back(app2);

            }
          else
            {
              SSTMAC_DEBUG << "instantlaunch::restart_app() - app "
                    << (*it)->get_sid().to_string()
                    << " not restartable with checkpoint \n";
              newapps.push_back(linfo->apptype_->clone((*it)->get_sid()));
            }

          fullos_->kill_app(*it);
        }

      launch_message::ptr sleep = launch_message::construct(linfo);
      sleep->is_restart_wait_ = true;
      sleep->newapps_ = newapps;
      sleep->set_time(timestamp(RESTART_TIME));

      os_->execute_kernel(AMI::AMI_COMPUTE, AMI::AMI_COMP_SLEEP, sleep);

    }

  }
} // end of namespace sstmac
