/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/software/skeletons/test/sstmac_mpi_testall.h>
#include <sstmac/software/libraries/mpi/sstmac_mpi.h>
#include <sstmac/common/messages/valuepayload.h>

#include <math.h>

namespace sstmac
{
  namespace sw
  {
    using namespace std;


    int sstmac_mpi_testall::errors_ = 0;

    void
    sstmac_mpi_testall::consume_params(const sim_parameters::ptr &params)
    {
      if (params->has_param("testMPI_stop_at_errors"))
        stop_at_errors_ = params->get_int_param("testMPI_stop_at_errors");
      else
        stop_at_errors_ = false;
    }

    void
    sstmac_mpi_testall::skeleton_main()
    {
      //std::cout << "starting appMPI_pingpong" << endl;

      MPI_Init(NULL, NULL);

      int rank, size;
      MPI_Comm_rank(MPI_COMM_WORLD, &rank);
      MPI_Comm_size(MPI_COMM_WORLD, &size);
      //dbg << "App " << rank << " initialized" << "\n";

      if (size < 8)
        {
          std::cout << "rank: " << rank << std::endl;
          std::cout << "size found: " << size << std::endl;
          throw rangeerror("sstmac_mpi_testall needs at least 8 nodes to test on");
        }

      //----------first, lets do a ping pong with payload
      test_sendrecv();

      // ---- test nonblocking recv
      test_asynch();

      // ---- alright, let's try a broadcast
      test_bcast();

      // --- try a reduce
      test_reduce();

      // ---- sync up with barrier
      test_barrier();

      // ---- test allreduce
      test_allreduce();

      // ------ test scatter
      test_scatter();

      // ------ test gather
      test_gather();

      //------ test communicator functions
      test_comms();

      // ------ test wait functions
      test_wait();

      // ------ test allgather
      test_allgather();

      // ------ test alltoall
      test_alltoall();

      // ------ test probing
      test_probe();

      // ------- test persistent
      test_persistent();

      // ------ test reduce_scatter
      // test_reducescatter();

      // ----- finalize
      MPI_Barrier(MPI_COMM_WORLD);

      if (rank == 0 && errors_ == 0)
        {
          SSTMAC_DEBUG << "- Finished testing! test successful \n";
        }
      else if (rank == 0 && errors_ != 0)
        {
          SSTMAC_DEBUG << "- Finished testing! " << errors_
                << " ERRORS - test not successful \n";
        }

      MPI_Finalize();

    }

    void
    sstmac_mpi_testall::test_sendrecv()
    {
      int rank, size;
      MPI_Comm_rank(MPI_COMM_WORLD, &rank);
      MPI_Comm_size(MPI_COMM_WORLD, &size);

      SSTMAC_DEBUG << "testall[" << rank << "]-- Testing payload send/recv \n";

      int count = 1; //arbitrary

      int tag(0);
      bool participant = true;
      if ((size % 2) && (rank + 1 >= size))
        {
          // This is the odd-node-out -- communicating with no-one.
          participant = false;
        }
      if (participant)
        {
          int buddy(rank ^ 1); // 0<=>1, 2<=>3, etc.

          if ((rank) & 1)
            { // even values of half-cycle plus rank.

              //dbg_ << "testall[" << rank << "]: sending payload \n";

              int pay = rank * 1000;
              MPI_Send(&pay, count, MPI_INT, buddy, tag, MPI_COMM_WORLD);

            }
          else
            {
              // dbg_ << "testall[" << rank << "]: receiving payload \n";
              MPI_Status stat;
              int recvdata;
              MPI_Recv(&recvdata, count, MPI_INT, buddy, tag, MPI_COMM_WORLD,
                  &stat);


              if (recvdata != (buddy * 1000))
                {
                  errors_++;
                  SSTMAC_DEBUG << "ERROR at rank " << rank
                        << ": SEND-RECV: received value " << recvdata
                        << " and we should have got " << buddy * 1000 << "\n";

                  if (stop_at_errors_)
                    throw ssterror("an error occurred in the application");
                }

            }

        }
    }

    void
    sstmac_mpi_testall::test_barrier()
    {
      int rank, size;
      MPI_Comm_rank(MPI_COMM_WORLD, &rank);
      MPI_Comm_size(MPI_COMM_WORLD, &size);

      SSTMAC_DEBUG << "testall[" << rank << "] -- Testing barrier " << " \n";

      if (rank % 2 == 0)
        {
          timestamp t(2e-3);
          compute(t);
        }

      MPI_Barrier(MPI_COMM_WORLD);

      SSTMAC_DEBUG << "testall[" << rank << "] -- Through barrier " << " \n";
    }

    void
    sstmac_mpi_testall::test_reduce()
    {
      /*int rank, size;
       MPI_Comm_rank(MPI_COMM_WORLD, &rank);
       MPI_Comm_size(MPI_COMM_WORLD, &size);
       int count = 100; //arbitrary

       SSTMAC_DEBUG << "testall[" << rank << "] -- Testing reduce " << " \n";

       payload::const_ptr result;

       payload::const_ptr load = valuepayload<int>::construct((1 << rank));
       mpi()->reduce(count, MPI_DOUBLE, mpiop::sum, int(0), MPI_COMM_WORLD, load,
       result);

       if (rank == 0)
       {
       const valuepayload<int>::const_ptr recvdata = boost::dynamic_pointer_cast<
       const valuepayload<int> >(result);

       if (!recvdata)
       {
       errors_++;
       SSTMAC_DEBUG << "ERROR at rank " << rank
       << ": REDUCE: NULL payload recv'd \n";
       if (stop_at_errors_)
       throw ssterror("an error occurred in the application");
       }
       else
       {
       int expected = (int) pow(2.0, MPI_COMM_WORLD.size().id) - 1;
       if (recvdata->data() != expected)
       {
       errors_++;
       SSTMAC_DEBUG << "ERROR at rank " << rank << ": REDUCE: received value "
       << recvdata->data() << " and we should have got " << expected << "\n";
       if (stop_at_errors_)
       throw ssterror("an error occurred in the application");
       }

       }
       }*/
    }

    void
    sstmac_mpi_testall::test_asynch()
    {
      int rank, size;
      MPI_Comm_rank(MPI_COMM_WORLD, &rank);
      MPI_Comm_size(MPI_COMM_WORLD, &size);
      int count = 100; //arbitrary

      MPI_Status stat;
      if (rank == 0)
        {
          SSTMAC_DEBUG << "testall[" << rank
                << "] -- Testing nonblocking recv/wait \n";
          int r(1);
          int t(1);
          MPI_Send(NULL, count, MPI_INT, r, t, MPI_COMM_WORLD);

          SSTMAC_DEBUG << "testall[" << rank
                << "]---- This should come SECOND -- \n\n\n";
        }
      else if (rank == 1)
        {
          MPI_Request req;
          int r(0);
          int t(1);
          MPI_Irecv(NULL, count, MPI_INT, r, t, MPI_COMM_WORLD, &req);
          SSTMAC_DEBUG << "testall[" << rank
                << "]---- This should come FIRST -- \n";
          MPI_Wait(&req, &stat);

          SSTMAC_DEBUG << "testall[" << rank
                << "]---- This should come THIRD -- \n";

        }
    }

    void
    sstmac_mpi_testall::test_allreduce()
    {
      int rank, size;
      MPI_Comm_rank(MPI_COMM_WORLD, &rank);
      MPI_Comm_size(MPI_COMM_WORLD, &size);
      int count = 1; //arbitrary

      SSTMAC_DEBUG << "testall[" << rank << "] -- Testing allreduce " << " \n";

      int pay = 1 << rank;;
      int allrecvdata;
      MPI_Allreduce(&pay, &allrecvdata, count, MPI_INT, MPI_SUM, MPI_COMM_WORLD);


      int expected = (int) pow(2.0, size) - 1;
      if (allrecvdata != (expected))
        {
          errors_++;
          SSTMAC_DEBUG << "ERROR at rank " << rank
                << ": ALLREDUCE: received value " << allrecvdata
                << " and we should have got " << expected << "\n";
          if (stop_at_errors_)
            throw ssterror("an error occurred in the application");
        }

    }

    void
    sstmac_mpi_testall::test_bcast()
    {
      int rank, size;
      MPI_Comm_rank(MPI_COMM_WORLD, &rank);
      MPI_Comm_size(MPI_COMM_WORLD, &size);
      int count = 1; //arbitrary

      SSTMAC_DEBUG << "testall[" << rank << "] -- Testing payload broadcast \n";

      payload::const_ptr load;

      int root(0);
      int pay;
      if (rank == root)
        {
          pay = 1234;
        }

      MPI_Bcast(&pay, count, MPI_DOUBLE, root, MPI_COMM_WORLD);
      int recvdata = pay;

      if (recvdata != (1234))
        {
          errors_++;
          SSTMAC_DEBUG << "ERROR at rank " << rank << ": BCAST: received value "
              << recvdata << " and we should have got " << 1234 << "\n";
        if (stop_at_errors_)
          throw ssterror("an error occurred in the application");
      }

  }

  void
  sstmac_mpi_testall::test_scatter()
  {
    /*  int rank, size;
     MPI_Comm_rank(MPI_COMM_WORLD, &rank);
     MPI_Comm_size(MPI_COMM_WORLD, &size);
     int count = 100; //arbitrary

     SSTMAC_DEBUG << "testall[" << rank << "] -- Testing scatter " << " \n";

     std::vector<payload::const_ptr> vals;
     std::vector<int> vcounts;

     if (rank == 0)
     {
     for (int i = 0; i < size; i++)
     {
     int val = 1 << i;
     vals.push_back(valuepayload<int>::construct(val));
     vcounts.push_back(count);
     }
     }

     payload::const_ptr result;

     mpi()->scatter(count, MPI_DOUBLE, count, MPI_DOUBLE, int(0),
     MPI_COMM_WORLD, vals, result);

     const valuepayload<int>::const_ptr scatterdata = boost::dynamic_pointer_cast<
     const valuepayload<int> >(result);

     if (!scatterdata)
     {
     errors_++;
     SSTMAC_DEBUG << "ERROR at rank " << rank
     << ": NULL payload recv'd for SCATTER \n";
     if (stop_at_errors_)
     throw ssterror("an error occurred in the application");
     }
     else
     {
     int expected = (int) (1 << rank);
     if (scatterdata->data() != (expected))
     {
     errors_++;
     SSTMAC_DEBUG << "ERROR at rank " << rank << ": SCATTER: received value "
     << scatterdata->data() << " and we should have got " << expected << "\n";
     if (stop_at_errors_)
     throw ssterror("an error occurred in the application");
     }
     }

     SSTMAC_DEBUG << "testall[" << rank << "] -- Testing scatterv " << " \n";
     mpi()->scatterv(vcounts, MPI_DOUBLE, count, MPI_DOUBLE,
     int(0), MPI_COMM_WORLD, vals, result);

     const valuepayload<int>::const_ptr scatterdata2 = boost::dynamic_pointer_cast<
     const valuepayload<int> >(result);

     if (!scatterdata2)
     {
     errors_++;
     SSTMAC_DEBUG << "ERROR at rank " << rank
     << ": NULL payload recv'd for SCATTERV \n";
     if (stop_at_errors_)
     throw ssterror("an error occurred in the application");
     }
     else
     {
     int expected = (int) (1 << rank);
     if (scatterdata2->data() != (expected))
     {
     errors_++;
     SSTMAC_DEBUG << "ERROR at rank " << rank << ": SCATTERV: received value "
     << scatterdata2->data() << " and we should have got " << expected << "\n";
     if (stop_at_errors_)
     throw ssterror("an error occurred in the application");
     }
     }*/

  }

  void
  sstmac_mpi_testall::test_gather()
  {
    /*  int rank, size;
     MPI_Comm_rank(MPI_COMM_WORLD, &rank);
     MPI_Comm_size(MPI_COMM_WORLD, &size);
     int count = 100; //arbitrary

     SSTMAC_DEBUG << "testall[" << rank << "] -- Testing gather " << " \n";

     std::vector<payload::const_ptr> vals;

     payload::const_ptr load = valuepayload<int>::construct(1 << rank);

     mpi()->gather(count, MPI_DOUBLE, count, MPI_DOUBLE, int(0),
     MPI_COMM_WORLD, load, vals);

     if (rank == 0)
     {
     std::vector<payload::const_ptr>::iterator it, end = vals.end();

     if (vals.size() != MPI_COMM_WORLD.size().id)
     {
     SSTMAC_DEBUG << "ERROR at rank " << rank << ": GATHER: result vector size ("
     << vals.size() << ") does not match MPI_COMM_WORLD comm size (" << MPI_COMM_WORLD.size().id
     << ") \n";
     if (stop_at_errors_)
     throw ssterror("an error occurred in the application");
     }

     int i = 0;
     for (it = vals.begin(); it != end; it++)
     {
     if (*it)
     {
     int expected = (int) (1 << i);
     const valuepayload<int>::const_ptr scatterdata = boost::dynamic_pointer_cast<
     const valuepayload<int> >(*it);

     if (scatterdata->data() != (expected))
     {
     errors_++;
     SSTMAC_DEBUG << "ERROR at rank " << rank << ": GATHER: received value "
     << scatterdata->data() << " and we should have got " << expected << "\n";
     if (stop_at_errors_)
     throw ssterror("an error occurred in the application");
     }
     }
     else
     {
     errors_++;
     SSTMAC_DEBUG << "ERROR at rank " << rank
     << ": NULL payload recv'd for GATHER \n";
     if (stop_at_errors_)
     throw ssterror("an error occurred in the application");
     }

     i++;

     }

     }

     SSTMAC_DEBUG << "testall[" << rank << "] -- Testing gatherv " << " \n";

     std::vector<payload::const_ptr> vals2;

     payload::const_ptr load2 = valuepayload<int>::construct(1 << rank);
     std::vector<int> vcounts;

     for (int i = 0; i < size; i++)
     {
     vcounts.push_back(count);
     }

     mpi()->gatherv(count, MPI_DOUBLE, vcounts, MPI_DOUBLE,
     int(0), MPI_COMM_WORLD, load2, vals2);

     if (rank == 0)
     {
     std::vector<payload::const_ptr>::iterator it, end = vals2.end();

     if (vals2.size() != MPI_COMM_WORLD.size().id)
     {
     SSTMAC_DEBUG << "ERROR at rank " << rank << ": GATHERV: result vector size ("
     << vals.size() << ") does not match MPI_COMM_WORLD comm size (" << MPI_COMM_WORLD.size().id
     << ") \n";
     }

     int i = 0;
     for (it = vals2.begin(); it != end; it++)
     {
     if (*it)
     {
     int expected = (int) (1 << i);
     const valuepayload<int>::const_ptr scatterdata = boost::dynamic_pointer_cast<
     const valuepayload<int> >(*it);

     if (scatterdata->data() != (expected))
     {
     errors_++;
     SSTMAC_DEBUG << "ERROR at rank " << rank << ": GATHERV: received value "
     << scatterdata->data() << " and we should have got " << expected << "\n";
     if (stop_at_errors_)
     throw ssterror("an error occurred in the application");
     }
     }
     else
     {
     errors_++;
     SSTMAC_DEBUG << "ERROR at rank " << rank
     << ": NULL payload recv'd for GATHERV \n";
     if (stop_at_errors_)
     throw ssterror("an error occurred in the application");
     }

     i++;

     }

     }*/
  }

  void
  sstmac_mpi_testall::test_allgather()
  {
    /*  int rank, size;
     MPI_Comm_rank(MPI_COMM_WORLD, &rank);
     MPI_Comm_size(MPI_COMM_WORLD, &size);
     int count = 100; //arbitrary

     SSTMAC_DEBUG << "testall[" << rank << "] -- Testing allgather " << " \n";

     std::vector<payload::const_ptr> vals;

     payload::const_ptr load = valuepayload<int>::construct(1 << rank);

     mpi()->allgather(count, MPI_DOUBLE, count, MPI_DOUBLE, MPI_COMM_WORLD,
     load, vals);

     std::vector<payload::const_ptr>::iterator it, end = vals.end();

     if (vals.size() != MPI_COMM_WORLD.size().id)
     {
     errors_++;
     SSTMAC_DEBUG << "ERROR at rank " << rank
     << ": ALLGATHER: result vector size (" << vals.size()
     << ") does not match MPI_COMM_WORLD comm size (" << MPI_COMM_WORLD.size().id << ") \n";
     if (stop_at_errors_)
     throw ssterror("an error occurred in the application");
     }

     int i = 0;
     for (it = vals.begin(); it != end; it++)
     {
     if (*it)
     {
     int expected = (int) (1 << i);
     const valuepayload<int>::const_ptr scatterdata = boost::dynamic_pointer_cast<
     const valuepayload<int> >(*it);

     if (scatterdata->data() != (expected))
     {
     errors_++;
     SSTMAC_DEBUG << "ERROR at rank " << rank << ": ALLGATHER: received value "
     << scatterdata->data() << " and we should have got " << expected << "\n";
     if (stop_at_errors_)
     throw ssterror("an error occurred in the application");
     }
     }
     else
     {
     errors_++;
     SSTMAC_DEBUG << "ERROR at rank " << rank
     << ": NULL payload recv'd for ALLGATHER \n";
     if (stop_at_errors_)
     throw ssterror("an error occurred in the application");
     }

     i++;

     }

     SSTMAC_DEBUG << "testall[" << rank << "] -- Testing allgatherv " << " \n";

     std::vector<payload::const_ptr> vals2;

     payload::const_ptr load2 = valuepayload<int>::construct(1 << rank);
     std::vector<int> vcounts;

     for (int i = 0; i < size; i++)
     {
     vcounts.push_back(count);
     }

     mpi()->allgatherv(count, MPI_DOUBLE, vcounts, MPI_DOUBLE,
     MPI_COMM_WORLD, load2, vals2);

     std::vector<payload::const_ptr>::iterator it2, end2 = vals2.end();

     if (vals2.size() != MPI_COMM_WORLD.size().id)
     {
     errors_++;
     SSTMAC_DEBUG << "ERROR at rank " << rank
     << ": ALLGATHERV: result vector size (" << vals.size()
     << ") does not match MPI_COMM_WORLD comm size (" << MPI_COMM_WORLD.size().id << ") \n";
     if (stop_at_errors_)
     throw ssterror("an error occurred in the application");
     }

     i = 0;
     for (it2 = vals2.begin(); it2 != end2; it2++)
     {
     if (*it2)
     {
     int expected = (int) (1 << i);
     const valuepayload<int>::const_ptr scatterdata = boost::dynamic_pointer_cast<
     const valuepayload<int> >(*it2);

     if (scatterdata->data() != (expected))
     {
     errors_++;
     SSTMAC_DEBUG << "ERROR at rank " << rank << ": ALLGATHERV: received value "
     << scatterdata->data() << " and we should have got " << expected << "\n";
     if (stop_at_errors_)
     throw ssterror("an error occurred in the application");
     }
     }
     else
     {
     errors_++;
     SSTMAC_DEBUG << "ERROR at rank " << rank
     << ": NULL payload recv'd for ALLGATHERV \n";
     if (stop_at_errors_)
     throw ssterror("an error occurred in the application");
     }

     i++;

     }*/

  }

  void
  sstmac_mpi_testall::test_alltoall()
  {
    /*  int rank, size;
     MPI_Comm_rank(MPI_COMM_WORLD, &rank);
     MPI_Comm_size(MPI_COMM_WORLD, &size);
     int count = 100; //arbitrary

     SSTMAC_DEBUG << "testall[" << rank << "] -- Testing alltoall " << " \n";

     std::vector<payload::const_ptr> vals;

     for (int i = 0; i < size; i++)
     {
     int val = 1 << i;
     vals.push_back(valuepayload<int>::construct(val));
     }

     std::vector<payload::const_ptr> result;

     mpi()->alltoall(count, MPI_DOUBLE, count, MPI_DOUBLE, MPI_COMM_WORLD,
     vals, result);

     std::vector<payload::const_ptr>::iterator it, end = result.end();

     if (result.size() != MPI_COMM_WORLD.size().id)
     {
     errors_++;
     SSTMAC_DEBUG << "ERROR at rank " << rank
     << ": ALLTOALL: result vector size (" << result.size()
     << ") does not match MPI_COMM_WORLD comm size (" << MPI_COMM_WORLD.size().id << ") \n";
     if (stop_at_errors_)
     throw ssterror("an error occurred in the application");
     }

     int i = 0;
     for (it = result.begin(); it != end; it++)
     {
     if (*it)
     {
     int expected = (int) (1 << rank);
     const valuepayload<int>::const_ptr scatterdata = boost::dynamic_pointer_cast<
     const valuepayload<int> >(*it);

     if (scatterdata->data() != (expected))
     {
     errors_++;
     SSTMAC_DEBUG << "ERROR at rank " << rank << ": ALLTOALL: received value "
     << scatterdata->data() << " and we should have got " << expected << "\n";
     if (stop_at_errors_)
     throw ssterror("an error occurred in the application");
     }
     }
     else
     {
     errors_++;
     SSTMAC_DEBUG << "ERROR at rank " << rank
     << ": NULL payload recv'd for ALLTOALL for rank " << i << " \n";
     if (stop_at_errors_)
     throw ssterror("an error occurred in the application");
     }

     i++;

     }*/

  }

  void
  sstmac_mpi_testall::test_comms()
  {
    /*int rank, size;
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &size);

    SSTMAC_DEBUG << "testall[" << rank << "] -- Testing comm create \n";

    MPI_Group gr_ev;
    MPI_Group gr_odd;
    for (int i = 0; i < size; i++)
      {
        if (i % 2 == 0)
          gr_ev.members_.push_back(mpiid(i));
        else
          gr_odd.members_.push_back(mpiid(i));
      }

    MPI_Comm evens;

    MPI_Comm_create(MPI_COMM_WORLD, gr_ev, &evens);

    if (rank % 2 == 0)
      {
        if (evens == MPI_Comm::comm_null)
          {
            SSTMAC_DEBUG
                  <<
                  "ERROR at rank "
                  << rank
                  << ": create_comm returned null comm when it should be a real thing \n";
            errors_++;
            if (stop_at_errors_)
              throw ssterror("an error occurred in the application");
          }
        else
          {

            if (evens.rank().id != rank / 2)
              {
                errors_++;
                SSTMAC_DEBUG << "testall[" << rank
                      << "] -- my new comm rank is " << evens.rank().id
                      << " and it should be " << (rank / 2) << "\n";
                if (stop_at_errors_)
                  throw ssterror("an error occurred in the application");
              }

            if (evens.id().id != 1)
              {
                SSTMAC_DEBUG << "ERROR at rank " << rank
                      << ": create produced id " << evens.id().id
                      << " and it should have been " << 1 << " \n";
                errors_++;
                if (stop_at_errors_)
                  throw ssterror("an error occurred in the application");
              }

          }

        SSTMAC_DEBUG << "testall[" << rank << "] -- Testing comm dup \n";

        MPI_Comm duped;
        MPI_Comm_dup(evens, &duped);
        mpicommid shouldbe(
            (rank % 2 == 0) ? evens.id().id + 1 : evens.id().id + 2);

        if (duped.id().id != shouldbe.id)
          {
            SSTMAC_DEBUG << "ERROR at rank " << rank
                  << ": create_dup produced id " << duped.id().id
                  << " and it should have been " << shouldbe.id << " \n";
            errors_++;
            if (stop_at_errors_)
              throw ssterror("an error occurred in the application");
          }

        if (duped.size() != evens.size())
          {
            SSTMAC_DEBUG << "ERROR at rank " << rank
                  << ": create_dup produced comm with size " << duped.size().id
                  << " and it should have been " << evens.size().id << " \n";
            errors_++;
            if (stop_at_errors_)
              throw ssterror("an error occurred in the application");
          }

        if (duped.rank() != evens.rank())
          {
            SSTMAC_DEBUG << "ERROR at rank " << rank
                  << ": create_dup produced comm where my rank is "
                  << duped.rank().id << " and it should have been "
                  << evens.rank().id << " \n";
            errors_++;
            if (stop_at_errors_)
              throw ssterror("an error occurred in the application");
          }

        SSTMAC_DEBUG << "testall[" << rank << "] -- Testing comm split \n";

        MPI_Comm mynewsplit;
        int newkey = duped.rank().id;
        MPI_Comm_split(duped, ((rank == 2) ? 0 : 1), newkey, &mynewsplit);
        int sizecheck = (rank == 2) ? 1 : ((duped.size().id) - 1);

        if (mynewsplit == MPI_Comm::comm_null)
          {
            SSTMAC_DEBUG << "ERROR at rank " << rank
                  << ": create_split returned null comm \n";
            errors_++;
          }
        else
          {

            if (mynewsplit.size().id != sizecheck)
              {
                SSTMAC_DEBUG << "ERROR at rank " << rank
                      << ": create_split produced comm with size "
                      << mynewsplit.size().id << " and it should have been "
                      << sizecheck << " \n";
                errors_++;
                if (stop_at_errors_)
                  throw ssterror("an error occurred in the application");
              }

          }

      }
    else
      {
        if (evens != MPI_Comm::comm_null)
          {
            SSTMAC_DEBUG
                  <<
                  "ERROR at rank "
                  << rank
                  << ": create_comm returned a real comm when it should have returned the null comm \n";
            errors_++;
            if (stop_at_errors_)
              throw ssterror("an error occurred in the application");
          }

      }*/

  }

  void
  sstmac_mpi_testall::test_wait()
  {
    int rank, size;
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &size);

    int count = 100;
    int tag(698);

    int sender(5);

    SSTMAC_DEBUG << "testall[" << rank << "] -- Testing wait any \n";

    if (rank == 0)
      {

        //MPI_Request* reqs = (MPI_Request*)malloc((size-1) * sizeof(MPI_Request));
        MPI_Request reqs[size - 1];

        for (int i = 1; i < size; i++)
          {
            MPI_Request req;
            MPI_Irecv(NULL, count, MPI_DOUBLE, int(i), tag, MPI_COMM_WORLD,
                &req);
            reqs[i - 1] = req;

          }

        MPI_Status stat;
        int index;
        MPI_Waitany(size - 1, reqs, &index, &stat);

        if (index != sender - 1)
          { //subtract 1 because I didn't push a mpirequest in the vector for myself
            SSTMAC_DEBUG << "ERROR at rank " << rank
                  << ": waitany unblocked with index " << index
                  << " and it should have been sent from " << sender << "\n";
            errors_++;
            if (stop_at_errors_)
              throw ssterror("an error occurred in the application");
          }

        //free(reqs);

      }
    else if (rank == sender)
      {
        MPI_Send(NULL, count, MPI_DOUBLE, int(0), tag, MPI_COMM_WORLD);
      }

    MPI_Barrier(MPI_COMM_WORLD);

    SSTMAC_DEBUG << "testall[" << rank << "] -- Testing wait some \n";

    int tag2(699);

    if (rank == 0)
      {
        sleep(timestamp(1)); //lag me, so the others have a chance to send

        //MPI_Request* reqs = (MPI_Request*)malloc((size-1) * sizeof(MPI_Request));
        MPI_Request reqs[size - 1];

        for (int i = 1; i < size; i++)
          {
            MPI_Request req;
            MPI_Irecv(NULL, count, MPI_DOUBLE, int(i), tag2, MPI_COMM_WORLD,
                &req);
            reqs[i - 1] = (req);
          }

        MPI_Status* stat;
        int* index;
        int outcount;
        MPI_Waitsome(size - 1, reqs, &outcount, index, stat);

        if (outcount != (size / 2) - 1)
          { //subtract 1 for myself
            SSTMAC_DEBUG << "ERROR at rank " << rank
                  << ": waitsome unblocked with number of indeces " << outcount
                  << " and it should have been " << ((size / 2) - 1) << "\n";
            errors_++;
            if (stop_at_errors_)
              throw ssterror("an error occurred in the application");
          }

      }
    else if (rank % 2 == 0)
      {
        MPI_Send(NULL, count, MPI_DOUBLE, int(0), tag2, MPI_COMM_WORLD);
      }

    MPI_Barrier(MPI_COMM_WORLD);

  }

  void
  sstmac_mpi_testall::test_reducescatter()
  {
    /* int rank, size;
     MPI_Comm_rank(MPI_COMM_WORLD, &rank);
     MPI_Comm_size(MPI_COMM_WORLD, &size);

     SSTMAC_DEBUG << "testall[" << rank << "] -- Testing reduce_scatter \n";

     std::vector<int> recvcounts;

     for (int i = 0; i < size; i++)
     {
     recvcounts.push_back(10 * rank);
     }

     mpi()->reduce_scatter(recvcounts, MPI_DOUBLE, mpiop::sum, MPI_COMM_WORLD);

     SSTMAC_DEBUG << "testall[" << rank << "]: reduce_scatter: recv counts: ";
     for (int i = 0; i < recvcounts.size(); i++)
     {
     SSTMAC_DEBUG << recvcounts[i] << ", ";
     }

     SSTMAC_DEBUG << "\n";*/

  }

  void
  sstmac_mpi_testall::test_probe()
  {
    /*int rank, size;
     MPI_Comm_rank(MPI_COMM_WORLD, &rank);
     MPI_Comm_size(MPI_COMM_WORLD, &size);

     int count = 100;
     int tag(713);
     int tag2(714);

     mpi()->barrier(MPI_COMM_WORLD);

     int sender(5);

     SSTMAC_DEBUG << "testall[" << rank << "] -- Testing probe \n";

     if (rank == 0)
     {

     std::vector<MPI_Request> reqs;

     for (int i = 1; i < size; i++)
     {
     MPI_Request req;
     mpi()->irecv(count, MPI_DOUBLE, int(i), tag, MPI_COMM_WORLD, req);
     reqs.push_back(req);
     }

     MPI_Status stat;
     int index;
     bool flag;
     std::vector<int> indices;
     std::vector<MPI_Status> statuses;

     mpi()->test(reqs.at(sender.id), flag, stat);

     if (flag)
     {
     SSTMAC_DEBUG << "ERROR at rank " << rank
     << ": test returned completed, and it probably should not have \n";
     errors_++;
     if (stop_at_errors_)
     throw ssterror("an error occurred in the application");
     }

     mpi()->testsome(reqs, indices, statuses);

     if (indices.size() > 0)
     {
     SSTMAC_DEBUG << "ERROR at rank " << rank
     << ": testsome returned completed, and it probably should not have \n";
     errors_++;
     if (stop_at_errors_)
     throw ssterror("an error occurred in the application");
     }

     mpi()->testany(reqs, index, flag, stat);

     if (flag)
     {
     SSTMAC_DEBUG << "ERROR at rank " << rank
     << ": testany returned completed, and it probably should not have \n";
     errors_++;
     if (stop_at_errors_)
     throw ssterror("an error occurred in the application");
     }

     mpi()->iprobe(sender, tag, MPI_COMM_WORLD, flag, stat);

     if (flag)
     {
     SSTMAC_DEBUG << "ERROR at rank " << rank
     << ": iprobe returned completed, and it probably should not have \n";
     errors_++;
     if (stop_at_errors_)
     throw ssterror("an error occurred in the application");
     }

     // -------  here is the blocking probe ----------- //
     mpi()->probe(sender, tag, MPI_COMM_WORLD, stat);

     if (!stat)
     {
     SSTMAC_DEBUG << "ERROR at rank " << rank
     << ": blocking probe did not return a valid mpistatus \n";
     errors_++;
     if (stop_at_errors_)
     throw ssterror("an error occurred in the application");
     }

     mpi()->test(reqs.at(sender.id - 1), flag, stat);

     if (!flag)
     {
     SSTMAC_DEBUG << "ERROR at rank " << rank
     << ": test returned NOT completed, and it should not have \n";
     errors_++;
     if (stop_at_errors_)
     throw ssterror("an error occurred in the application");
     }

     mpi()->testsome(reqs, indices, statuses);

     if (indices.size() != 1)
     {
     SSTMAC_DEBUG << "ERROR at rank " << rank
     << ": testsome returned indices with size " << indices.size()
     << ", and it should have been 1 \n";
     errors_++;
     if (stop_at_errors_)
     throw ssterror("an error occurred in the application");
     }

     mpi()->testany(reqs, index, flag, stat);

     if (index != (sender.id - 1))
     {
     SSTMAC_DEBUG << "ERROR at rank " << rank << ": testany returned index "
     << index << ", and it should not have been " << (sender.id - 1) << " \n";
     errors_++;
     if (stop_at_errors_)
     throw ssterror("an error occurred in the application");
     }

     sleep(timestamp(1));

     mpi()->iprobe(sender, tag2, MPI_COMM_WORLD, flag, stat);

     if (!flag)
     {
     SSTMAC_DEBUG << "ERROR at rank " << rank
     << ": iprobe returned NOT completed, and it should not have \n";
     errors_++;
     if (stop_at_errors_)
     throw ssterror("an error occurred in the application");
     }

     mpi()->recv(count, MPI_DOUBLE, sender, tag2, MPI_COMM_WORLD, stat);

     }
     else if (rank == sender)
     {
     compute(timestamp(2));
     mpi()->send(count, MPI_DOUBLE, int(0), tag, MPI_COMM_WORLD);

     mpi()->send(count, MPI_DOUBLE, int(0), tag2, MPI_COMM_WORLD);
     }

     mpi()->barrier(MPI_COMM_WORLD);*/

  }

  void
  sstmac_mpi_testall::test_persistent()
  {
    /* int rank, size;
     MPI_Comm_rank(MPI_COMM_WORLD, &rank);
     MPI_Comm_size(MPI_COMM_WORLD, &size);

     int count = 100;
     int tag(715);

     SSTMAC_DEBUG << "testall[" << rank << "] -- Testing persistent actions \n";

     int sender(6);
     int recver(1);

     MPI_Request asend;
     MPI_Request arecv;

     if (rank == sender)
     {
     mpi()->send_init(count, MPI_DOUBLE, recver, tag, MPI_COMM_WORLD, asend);
     }

     if (rank == recver)
     {
     mpi()->recv_init(count, MPI_DOUBLE, sender, tag, MPI_COMM_WORLD, arecv);
     }

     SSTMAC_DEBUG << "testall[" << rank
     << "] -- Persistent: Haven't started them yet \n";

     if (rank == sender)
     {
     mpi()->start(asend);
     }

     if (rank == recver)
     {
     mpi()->start(arecv);
     }

     SSTMAC_DEBUG << "testall[" << rank
     << "] -- Persistent: OK, I started them \n";

     if (rank == sender)
     {
     MPI_Status stat;
     mpi()->wait(asend, stat);
     }

     if (rank == recver)
     {
     MPI_Status stat;
     mpi()->wait(arecv, stat);
     }

     SSTMAC_DEBUG << "testall[" << rank << "] -- Persistent: I'm through \n";
     */

  }

}
} //end of namespace sstmac
