/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/software/skeletons/undumpi/dumpimeta.h>
#include <sstmac/common/debug.h>
#include <sstmac/common/errors.h>
#include <cstring>
#include <errno.h>
#include <sstmac/common/basicstringtokenizer.h>

namespace sstmac { namespace sw {

  debug dumpimeta::dbg_;
        
  dumpimeta::dumpimeta(const std::string &filename)
  {
        numprocscheck_ = false;
        fileprefixcheck_ = false;

    metafile_ = filename;
                
    openfile();
    init_callbacks();
    parsemeta();

    if(!dbg_.is_real()){
      dbg_ = debug("<debug> dumpimeta");
      logger::register_logger(&dbg_);
    }
                
    SSTMAC_DEBUG << "dumpimeta filename = " << metafile_ << "\n";

    if(!numprocscheck_ && !fileprefixcheck_) {
            throw valueerror("The input dumpi metafile does not appear valid.\n");
    }
  }
        
  dumpimeta::ptr
  dumpimeta::construct(const std::string &fileroot) {
    return  dumpimeta::ptr(new dumpimeta(fileroot));
  }
        
  dumpimeta::~dumpimeta() throw()
  {

  }
        
  void dumpimeta::parsemeta()
  {

    std::deque<std::string> token;
    const std::string split_on = "=";
    std::string line;
    std::getline(infile_, line);

    while(infile_.good()) {
      token.clear();
      pst::BasicStringTokenizer::tokenize(line, token, split_on);
      if(token.size() != 2) {
          SSTMAC_DEBUG << "Skipping input line \"" << line << "\"\n";
      }
      else {
        CallMap_t::iterator it = callback_.find(token[0]);
        if(it != callback_.end())
                (this->*(it->second))(token);
        //else
        //std::cerr << "No callback found for " << token[0] << "\n";
      }
      std::getline(infile_, line);
    }
    infile_.close();
  }
        
  void dumpimeta::parse_numprocs(const std::deque<std::string> &token) 
  {
    numprocs_ = atoi(token[1].c_str());
    numprocscheck_ = true;

    //std::cout << "numprocs = " << numprocs_ << std::endl;
  }

  int dumpimeta::getnumprocs()
  {
    return numprocs_;
  }
        
  void dumpimeta::parse_fileprefix(const std::deque<std::string> &token) 
  {
                
    fileprefix_ = token[1].c_str();
    //fileprefix_ += '-';
    //std::cout << "fileprefix = " << fileprefix_ << std::endl;
                
    size_t found;
    found = fileprefix_.find_first_of("/");
                
    if(found == 0)
      dirplusfileprefix_ = fileprefix_;
    else
    {
      found = metafile_.find_last_of("/");
      dirplusfileprefix_ = metafile_.substr(0, found+1) + fileprefix_;
    }

    fileprefixcheck_ = true;
    //std::cout << "dirplusfileprefix = " << dirplusfileprefix_ << std::endl;
  }
        
  void  dumpimeta::init_callbacks()
  {
    callback_["numprocs"] = &dumpimeta::parse_numprocs;
    callback_["fileprefix"] = &dumpimeta::parse_fileprefix;
  }

  bool dumpimeta::openfile()
  {
    infile_.open(metafile_.c_str());

    if(!infile_)
      {
        throw ioerror(std::string("dumpimeta::openfile:  Failed to open ") +
                      metafile_ + ": " + strerror(errno));
      }
        
    return true;
  }
} } //end sstmac namespace
